Package: tigger
Type: Package
Version: 1.1.3
Date: 2026-04-13
Authors@R: c(person("Daniel", "Gadala-Maria", role=c("aut"), 
                    email="daniel.gadala-maria@yale.edu"),
             person("Susanna", "Marquez", role=c("aut", "cre"), 
                    email="susanna.marquez@yale.edu"),                    
             person("Moriah", "Cohen", role=c("aut"), 
                    email="moriah.cohen@biu.ac.il"),
             person("Ayelet", "Peres", role=c("aut"), 
                    email="ayelet.peres@yale.edu"),
             person("Jason", "Vander Heiden", role=c("aut"), 
                    email="jason.vanderheiden@gmail.com"),
             person("Gur", "Yaari", role=c("aut"), 
                    email="gur.yaari@yale.edu"),
             person("Steven", "Kleinstein", role=c("aut", "cph"), 
                    email="steven.kleinstein@yale.edu"))
Title: Infers Novel Immunoglobulin Alleles from Sequencing Data
Description: Infers the V genotype of an individual from immunoglobulin (Ig)
    repertoire sequencing data (AIRR-Seq, Rep-Seq). Includes detection of 
    any novel alleles. This information is then used to correct existing V 
    allele calls from among the sample sequences.
    Citations:
    Gadala-Maria, et al (2015) <doi:10.1073/pnas.1417683112>,
    Gadala-Maria, et al (2019) <doi:10.3389/fimmu.2019.00129>.
License: AGPL-3
URL: http://tigger.readthedocs.io
BugReports: https://github.com/immcantation/tigger/issues
LazyData: true
BuildVignettes: true
VignetteBuilder: knitr
Encoding: UTF-8
Depends: R (>= 4.0), ggplot2 (>= 3.4.0)
Imports: alakazam (>= 1.3.0), dplyr (>= 1.0.0), doParallel, foreach,
        graphics, gridExtra, gtools, iterators, lazyeval, parallel,
        rlang, stats, stringi, tidyr (>= 1.1.0), utils
Suggests: knitr, rmarkdown, testthat
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-16 09:36:22 UTC; susanna
Author: Daniel Gadala-Maria [aut],
  Susanna Marquez [aut, cre],
  Moriah Cohen [aut],
  Ayelet Peres [aut],
  Jason Vander Heiden [aut],
  Gur Yaari [aut],
  Steven Kleinstein [aut, cph]
Maintainer: Susanna Marquez <susanna.marquez@yale.edu>
Repository: CRAN
Date/Publication: 2026-04-17 05:10:02 UTC
Built: R 4.5.2; ; 2026-04-17 09:45:28 UTC; unix

Package: tigger
Type: Package
Version: 1.1.3
Date: 2026-04-13
Authors@R: c(person("Daniel", "Gadala-Maria", role=c("aut"), 
                    email="daniel.gadala-maria@yale.edu"),
             person("Susanna", "Marquez", role=c("aut", "cre"), 
                    email="susanna.marquez@yale.edu"),                    
             person("Moriah", "Cohen", role=c("aut"), 
                    email="moriah.cohen@biu.ac.il"),
             person("Ayelet", "Peres", role=c("aut"), 
                    email="ayelet.peres@yale.edu"),
             person("Jason", "Vander Heiden", role=c("aut"), 
                    email="jason.vanderheiden@gmail.com"),
             person("Gur", "Yaari", role=c("aut"), 
                    email="gur.yaari@yale.edu"),
             person("Steven", "Kleinstein", role=c("aut", "cph"), 
                    email="steven.kleinstein@yale.edu"))
Title: Infers Novel Immunoglobulin Alleles from Sequencing Data
Description: Infers the V genotype of an individual from immunoglobulin (Ig)
    repertoire sequencing data (AIRR-Seq, Rep-Seq). Includes detection of 
    any novel alleles. This information is then used to correct existing V 
    allele calls from among the sample sequences.
    Citations:
    Gadala-Maria, et al (2015) <doi:10.1073/pnas.1417683112>,
    Gadala-Maria, et al (2019) <doi:10.3389/fimmu.2019.00129>.
License: AGPL-3
URL: http://tigger.readthedocs.io
BugReports: https://github.com/immcantation/tigger/issues
LazyData: true
BuildVignettes: true
VignetteBuilder: knitr
Encoding: UTF-8
Depends: R (>= 4.0), ggplot2 (>= 3.4.0)
Imports: alakazam (>= 1.3.0), dplyr (>= 1.0.0), doParallel, foreach,
        graphics, gridExtra, gtools, iterators, lazyeval, parallel,
        rlang, stats, stringi, tidyr (>= 1.1.0), utils
Suggests: knitr, rmarkdown, testthat
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-16 09:36:22 UTC; susanna
Author: Daniel Gadala-Maria [aut],
  Susanna Marquez [aut, cre],
  Moriah Cohen [aut],
  Ayelet Peres [aut],
  Jason Vander Heiden [aut],
  Gur Yaari [aut],
  Steven Kleinstein [aut, cph]
Maintainer: Susanna Marquez <susanna.marquez@yale.edu>
Repository: CRAN
Date/Publication: 2026-04-17 05:10:02 UTC
Built: R 4.6.0; ; 2026-04-25 13:48:09 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cleanSeqs)
export(findNovelAlleles)
export(findUnmutatedCalls)
export(generateEvidence)
export(genotypeFasta)
export(getMutCount)
export(getMutatedPositions)
export(getPopularMutationCount)
export(inferGenotype)
export(inferGenotypeBayesian)
export(insertPolymorphisms)
export(plotGenotype)
export(plotNovel)
export(readIgFasta)
export(reassignAlleles)
export(selectNovel)
export(sortAlleles)
export(subsampleDb)
export(updateAlleleNames)
export(writeFasta)
import(ggplot2)
importFrom(alakazam,DNA_COLORS)
importFrom(alakazam,checkColumns)
importFrom(alakazam,getAllele)
importFrom(alakazam,getFamily)
importFrom(alakazam,getGene)
importFrom(alakazam,translateDNA)
importFrom(doParallel,registerDoParallel)
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,as_data_frame)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,combine)
importFrom(dplyr,data_frame)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,do)
importFrom(dplyr,filter)
importFrom(dplyr,glimpse)
importFrom(dplyr,group_by)
importFrom(dplyr,group_indices)
importFrom(dplyr,inner_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,summarise)
importFrom(dplyr,transmute)
importFrom(dplyr,ungroup)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(foreach,registerDoSEQ)
importFrom(graphics,plot)
importFrom(gridExtra,arrangeGrob)
importFrom(gtools,ddirichlet)
importFrom(iterators,icount)
importFrom(lazyeval,interp)
importFrom(parallel,clusterEvalQ)
importFrom(parallel,clusterExport)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(stats,confint)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,ecdf)
importFrom(stats,lm)
importFrom(stats,mad)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stringi,"stri_sub<-")
importFrom(stringi,stri_detect_fixed)
importFrom(stringi,stri_length)
importFrom(stringi,stri_replace_all_regex)
importFrom(stringi,stri_sub)
importFrom(stringi,stri_trans_toupper)
importFrom(tidyr,gather)
importFrom(tidyr,spread)
importFrom(tidyr,unnest)
importFrom(utils,citation)

[![](https://www.r-pkg.org/badges/version/tigger)](https://cran.r-project.org/package=tigger)
[![](http://cranlogs.r-pkg.org/badges/grand-total/tigger)](https://www.r-pkg.org/pkg/tigger)
[![](https://cranlogs.r-pkg.org/badges/tigger)](https://www.r-pkg.org/pkg/tigger)
[![](https://img.shields.io/static/v1?label=AIRR-C%20sw-tools%20v1&message=compliant&color=008AFF&labelColor=000000&style=plastic)](https://docs.airr-community.org/en/stable/swtools/airr_swtools_standard.html)
[![](https://img.shields.io/badge/github-repo-blue?logo=github&link=https://github.com/immcantation/tigger)](https://github.com/immcantation/tigger)

TIgGER
-------------------------------------------------------------------------------

High-throughput sequencing of B cell immunoglobulin receptors is providing unprecedented insight into adaptive immunity. A key step in analyzing these data involves assignment of the germline V, D and J gene segment alleles that comprise each immunoglobulin sequence by matching them against a database of known V(D)J alleles. However, this process will fail for sequences that utilize previously undetected alleles, whose frequency in the population is unclear.

TIgGER is a computational method that significantly improves V(D)J allele assignments by first determining the complete set of gene segments carried by an individual (including novel alleles) from V(D)J-rearrange sequences. TIgGER can then infer a subject's genotype from these sequences, and use this genotype to correct the initial V(D)J allele assignments.

The application of TIgGER continues to identify a surprisingly high frequency of novel alleles in humans, highlighting the critical need for this approach. (TIgGER, however, can and has been used with data from other species.)

Core Abilities
-------------------------------------------------------------------------------

* Detecting novel alleles
* Inferring a subject's genotype
* Correcting preliminary allele calls

Required Input
-------------------------------------------------------------------------------

* A table of sequences from a single individual, with columns containing the following:
    * V(D)J-rearranged nucleotide sequence (in IMGT-gapped format)
    * Preliminary V allele calls
    * Preliminary J allele calls
    * Length of the junction region
* Germline Ig sequences in IMGT-gapped fasta format (e.g., as those downloaded from [IMGT/GENE-DB](https://www.imgt.org/genedb/)

The former can be created through the use of [IMGT/HighV-QUEST](https://www.imgt.org) and [Change-O](http://changeo.readthedocs.io).

Contact
-------------------------------------------------------------------------------

If you need help or have any questions, please contact the [Immcantation Group](mailto:immcantation@googlegroups.com).

If you have discovered a bug or have a feature request, you can open an issue using the [issue tracker](https://github.com/immcantation/tigger/issues).

To receive alerts about Immcantation releases, news, events, and tutorials, join the [Immcantation News](https://groups.google.com/g/immcantation-news) Google Group. [Membership settings](https://groups.google.com/g/immcantation-news/membership) can be adjusted to change the frequency of email updates.

## ----eval=TRUE, message=FALSE, warning=FALSE----------------------------------
# Load packages required for this example
library(tigger)
library(dplyr)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Detect novel alleles
novel <- findNovelAlleles(AIRRDb, SampleGermlineIGHV, nproc=1)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Extract and view the rows that contain successful novel allele calls
novel_rows <- selectNovel(novel)

## ----eval=TRUE, warning=FALSE, fig.width=6, fig.height=8----------------------
# Plot evidence of the first (and only) novel allele from the example data
novel_row <- which(!is.na(novel$polymorphism_call))[1]
plotNovel(AIRRDb, novel[novel_row, ])

## ----eval=TRUE, warning=FALSE, fig.width=4, fig.height=3----------------------
# Infer the individual's genotype, using only unmutated sequences and checking
# for the use of the novel alleles inferred in the earlier step.
geno <- inferGenotype(AIRRDb, germline_db=SampleGermlineIGHV, novel=novel,
                      find_unmutated=TRUE)
# Save the genotype sequences to a vector
genotype_db <- genotypeFasta(geno, SampleGermlineIGHV, novel)
# Visualize the genotype and sequence counts
print(geno)
# Make a colorful visualization. Bars indicate presence, not proportion.
plotGenotype(geno, text_size = 10)

## ----eval=TRUE, warning=FALSE, fig.width=4, fig.height=3----------------------
# Infer the individual's genotype, using the bayesian method
geno_bayesian <- inferGenotypeBayesian(AIRRDb, germline_db=SampleGermlineIGHV, 
                                       novel=novel, find_unmutated=TRUE)
# Visualize the genotype and sequence counts
print(geno_bayesian)
# Make a colorful visualization. Bars indicate presence, not proportion.
plotGenotype(geno_bayesian, text_size=10)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Use the personalized genotype to determine corrected allele assignments
# Updated genotype will be placed in the v_call_genotyped column
sample_db <- reassignAlleles(AIRRDb, genotype_db)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Find the set of alleles in the original calls that were not in the genotype
not_in_genotype <- sample_db$v_call %>%
    strsplit(",") %>%
    unlist() %>%
    unique() %>%
    setdiff(names(genotype_db))

# Determine the fraction of calls that were ambiguous before/after correction
# and the fraction that contained original calls to non-genotype alleles. Note
# that by design, only genotype alleles are allowed in "after" calls.
data.frame(Ambiguous=c(mean(grepl(",", sample_db$v_call)),
                       mean(grepl(",", sample_db$v_call_genotyped))),
           NotInGenotype=c(mean(sample_db$v_call %in% not_in_genotype),
                           mean(sample_db$v_call_genotyped %in% not_in_genotype)),
           row.names=c("Before", "After")) %>% 
    t() %>% round(3)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
evidence <- generateEvidence(sample_db, novel, geno, genotype_db, SampleGermlineIGHV, fields = NULL)

evidence %>%
  select(gene, allele, polymorphism_call, sequences, unmutated_frequency)

# The TIgGER Trifecta -----------------------------------------------------

#' Find novel alleles from repertoire sequencing data
#'
#' \code{findNovelAlleles} analyzes mutation patterns in sequences thought to
#' align to each germline allele in order to determine which positions
#' might be polymorphic.
#'
#' The TIgGER allele-finding algorithm, briefly, works as follows:
#' Mutations are determined through comparison to the provided germline.
#' Mutation frequency at each *position* is determined as a function of
#' *sequence-wide* mutation counts. Polymorphic positions exhibit a high
#' mutation frequency despite sequence-wide mutation count. False positive of
#' potential novel alleles resulting from clonally-related sequences are guarded
#' against by ensuring that sequences perfectly matching the potential novel
#' allele utilize a wide range of combinations of J gene and junction length.
#'
#' @param    data             \code{data.frame} containing repertoire data. See details.
#' @param    germline_db      vector of named nucleotide germline sequences
#'                            matching the V calls in \code{data}. These should be
#'                            the gapped reference germlines used to make the V calls.
#' @param    v_call           name of the column in \code{data} with V allele calls.
#'                            Default is \code{v_call}.
#' @param    j_call           name of the column in \code{data} with J allele calls.
#'                            Default is \code{j_call}.
#' @param    seq              name of the column in \code{data} with the
#'                            aligned, IMGT-numbered, V(D)J nucleotide sequence.
#'                            Default is \code{sequence_alignment}.
#' @param    junction         Junction region nucleotide sequence, which includes
#'                            the CDR3 and the two flanking conserved codons. Default
#'                            is \code{junction}.
#' @param    junction_length  Number of junction nucleotides in the junction sequence.
#'                            Default is \code{junction_length}.
#' @param    germline_min     the minimum number of sequences that must have a
#'                            particular germline allele call for the allele to
#'                            be analyzed.
#' @param    auto_mutrange    if \code{TRUE}, the algorithm will attempt to
#'                            determine the appropriate mutation range
#'                            automatically using the mutation count of the most
#'                            common sequence assigned to each allele analyzed.
#' @param    mut_range        range of mutations that samples may carry and
#'                            be considered by the algorithm.
#' @param    pos_range        range of IMGT-numbered positions that should be
#'                            considered by the algorithm.
#' @param    pos_range_max    Name of the column in \code{data} with the ending
#'                            positions of the V alignment in the germline
#'                            (usually \code{v_germline_end}). The end of the alignment will
#'                            be used to limit the range of positions to be
#'                            considered to count mutations. With \code{NULL} all
#'                            positions in the IMGT V region will be considered. In
#'                            this case, in sequences where the V was trimmed
#'                            on the 3', mutated nucleotides could include nucleotides
#'                            from the CDR3.
#' @param    alpha            alpha value used for determining whether the
#'                            fit y-intercept is greater than the \code{y_intercept}
#'                            threshold.
#' @param    y_intercept      y-intercept threshold above which positions should be
#'                            considered potentially polymorphic.
#' @param    j_max            maximum fraction of sequences perfectly aligning
#'                            to a potential novel allele that are allowed to
#'                            utilize to a particular combination of junction
#'                            length and J gene. The closer to 1, the less strict
#'                            the filter for junction length and J gene diversity
#'                            will be.
#' @param    min_seqs         minimum number of total sequences (within the
#'                            desired mutational range and nucleotide range)
#'                            required for the samples to be considered.
#' @param    min_frac         minimum fraction of sequences that must have
#'                            usable nucleotides in a given position for that
#'                            position to considered.
#' @param    nproc            number of processors to use.
#'
#' @return
#' A \code{data.frame} with a row for each known allele analyzed.
#' Besides metadata on the the parameters used in the search, each row will have
#' either a note as to where the polymorphism-finding algorithm exited or a
#' nucleotide sequence for the predicted novel allele, along with columns providing
#' additional evidence.
#'
#' The output contains the following columns:
#' \itemize{
#'   \item \code{germline_call}: The input (uncorrected) V call.
#'   \item \code{note}: Comments regarding the inference.
#'   \item \code{polymorphism_call}: The novel allele call.
#'   \item \code{nt_substitutions}: Mutations identified in the novel allele, relative
#'         to the reference germline (\code{germline_call})
#'   \item \code{novel_imgt}: The novel allele sequence.
#'   \item \code{novel_imgt_count}:  The number of times the sequence \code{novel_imgt}
#'         is found in the input data. Considers the subsequence of \code{novel_imgt}
#'         in the \code{pos_range}.
#'   \item \code{novel_imgt_unique_j}: Number of distinct J calls associated to \code{novel_imgt}
#'         in the input data. Considers the subsequence of \code{novel_imgt} in the \code{pos_range}.
#'   \item \code{novel_imgt_unique_cdr3}: Number of distinct CDR3 sequences associated
#'         with \code{novel_imgt} in the input data. Considers the subsequence of \code{novel_imgt}
#'         in the \code{pos_range}.
#'   \item \code{perfect_match_count}: Final number of sequences retained to call the new
#'         allele. These are unique sequences that have V segments that perfectly match
#'         the predicted germline in the \code{pos_range}.
#'   \item \code{perfect_match_freq}: \code{perfect_match_count / germline_call_count}
#'   \item \code{germline_call_count}: The number of sequences with the \code{germline_call}
#'         in the input data that were initially considered for the analysis.
#'   \item \code{germline_call_freq}: The fraction of sequences with the \code{germline_call}
#'         in the input data initially considered for the analysis.
#'   \item \code{germline_imgt}: Germline sequence for \code{germline_call}.
#'   \item \code{germline_imgt_count}: The number of times the \code{germline_imgt}
#'         sequence is found in the input data.
#'   \item \code{mut_min}: Minimum mutation considered by the algorithm.
#'   \item \code{mut_max}: Maximum mutation considered by the algorithm.
#'   \item \code{mut_pass_count}: Number of sequences in the mutation range.
#'   \item \code{pos_min}: First position of the sequence considered by the algorithm (IMGT numbering).
#'   \item \code{pos_max}: Last position of the sequence considered by the algorithm (IMGT numbering).
#'   \item \code{y_intercept}: The y-intercept above which positions were considered
#'         potentially polymorphic.
#'   \item \code{y_intercept_pass}: Number of positions that pass the \code{y_intercept} threshold.
#'   \item \code{snp_pass}: Number of sequences that pass the \code{y_intercept} threshold and are
#'         within the desired nucleotide range (\code{min_seqs}).
#'   \item \code{unmutated_count}: Number of unmutated sequences.
#'   \item \code{unmutated_freq}: Number of unmutated sequences over \code{germline_imgt_count}.
#'   \item \code{unmutated_snp_j_gene_length_count}: Number of distinct combinations
#'         of SNP, J gene, and junction length.
#'   \item \code{snp_min_seqs_j_max_pass}: Number of SNPs that pass both the \code{min_seqs}
#'         and \code{j_max} thresholds.
#'   \item \code{alpha}: Significance threshold to be used when constructing the
#'         confidence interval for the y-intercept.
#'   \item \code{min_seqs}: Input \code{min_seqs}. The minimum number of total sequences
#'         (within the desired mutational range and nucleotide range) required
#'         for the samples to be considered.
#'   \item \code{j_max}: Input \code{j_max}. The maximum fraction of sequences perfectly
#'         aligning to a potential novel allele that are allowed to utilize to a particular
#'         combination of junction length and J gene.
#'   \item \code{min_frac}: Input \code{min_frac}. The minimum fraction of sequences that must
#'         have usable nucleotides in a given position for that position to be considered.
#' }
#'
#' The following comments can appear in the \code{note} column:
#'
#' \itemize{
#'   \item \emph{Novel allele found}: A novel allele was detected.
#'   \item \emph{Same as:}: The same novel allele sequence
#'               has been identified multiple times. If this happens, the function
#'               will also throw the message 'Duplicated polymorphism(s) found'.
#'   \item \emph{Plurality sequence too rare}: No sequence is frequent enough to pass
#'         the J test (\code{j_max}).
#'   \item \emph{A J-junction combination is too prevalent}: Not enough J diversity (\code{j_max}).
#'   \item \emph{No positions pass y-intercept test}: No positions above \code{y_intercept}.
#'   \item \emph{Insufficient sequences in desired mutational range}:
#'         \code{mut_range} and \code{pos_range}.
#'   \item \emph{Not enough sequences}: Not enough sequences in the desired mutational
#'         range and nucleotide range (\code{min_seqs}).
#'   \item \emph{No unmutated versions of novel allele found}: All observed variants of the
#'         allele are mutated.
#' }
#'
#' @seealso \link{selectNovel} to filter the results to show only novel alleles.
#' \link{plotNovel} to visualize the data supporting any
#' novel alleles hypothesized to be present in the data and
#' \link{inferGenotype} and \link{inferGenotypeBayesian} to determine if the novel alleles are frequent
#' enough to be included in the subject's genotype.
#'
#' @examples
#' \donttest{
#' # Note: In this example, with SampleGermlineIGHV,
#' # which contains reference germlines retrieved on August 2014,
#' # TIgGER finds the allele IGHV1-8*02_G234T. This allele
#' # was added to IMGT as IGHV1-8*03 on March 28, 2018.
#'
#' # Find novel alleles and return relevant data
#' novel <- findNovelAlleles(AIRRDb, SampleGermlineIGHV)
#' selectNovel(novel)
#' }
#'
#' @export
findNovelAlleles <- function(data, germline_db,
                             v_call="v_call",
                             j_call="j_call",
                             seq="sequence_alignment",
                             junction="junction",
                             junction_length="junction_length",
                             germline_min=200,
                             min_seqs=50,
                             auto_mutrange=TRUE,
                             mut_range=1:10,
                             pos_range=1:312,
                             pos_range_max = NULL,
                             y_intercept=0.125,
                             alpha=0.05,
                             j_max=0.15,
                             min_frac=0.75,
                             nproc=1) {
    . = idx = NULL

    # Keep only the db columns needed
    data <- data %>%
        dplyr::select(!!!rlang::syms(c(seq,
                                       v_call,
                                       j_call,
                                       junction_length,
                                       junction,ifelse(is.null(pos_range_max),"",pos_range_max))))
    gc()

    # Keep only the columns we need and clean up the sequences
    missing <- c(seq, v_call, j_call, junction_length,pos_range_max) %>%
        setdiff(colnames(data))
    if (length(missing) != 0) {
        stop("Could not find required columns in the input data:\n  ",
             paste(missing, collapse="\n  "))
    }
    empty_junctions <- sum(data[[junction_length]] == 0, na.rm=TRUE)
    if (empty_junctions > 0) {
        stop(empty_junctions, " sequences have junction ", "length of zero. ",
             "Please remove these sequences.")
    }
    germlines <- cleanSeqs(germline_db)
    names(germlines) <- getAllele(names(germlines), first=FALSE, strip_d=FALSE)
    data[[seq]] <- cleanSeqs(data[[seq]])


    # Find which rows' calls contain which germline alleles
    cutoff <-
        ifelse(germline_min < 1, round(nrow(data)*germline_min), germline_min)
    allele_groups <- sapply(names(germlines), grep, data[[v_call]], fixed=TRUE,
                           simplify=FALSE)
    names(allele_groups) <- names(germlines)
    allele_groups <- allele_groups[sapply(allele_groups, length) >= cutoff]
    if(length(allele_groups) == 0){
        stop_message <- paste("Not enough sample sequences were assigned to any germline:\n",
                              " (1) germline_min is too large or\n",
                              " (2) sequences names don't match germlines.")
        stop(stop_message)
    }
    allele_groups <- allele_groups[sortAlleles(names(allele_groups))]

    # Prepare for parallel processing
    nproc <- max(1,min(nproc, alakazam::cpuCount()-1))

    if(nproc == 1) {
        foreach::registerDoSEQ()
    } else {
        #cluster_type = ifelse(Sys.info()['sysname'] == "Windows", "PSOCK", "FORK")
        cluster <- parallel::makeCluster(nproc, type="PSOCK")
        parallel::clusterExport(cluster, list("allele_groups",
                                              "germlines",
                                              "v_call","j_call",
                                              "junction", "junction_length",
                                              "pos_range_max",
                                              "seq",
                                              "data",
                                              "min_seqs",
                                              "auto_mutrange",
                                              "mut_range",
                                              "pos_range",
                                              "y_intercept",
                                              "alpha",
                                              "j_max",
                                              "germline_min",
                                              "min_frac",
                                              "findLowerY",
                                              "mutationRangeSubset",
                                              "positionMutations",
                                              "superSubstring"),
                                envir=environment())
        doParallel::registerDoParallel(cluster)
    }

    out_list <- foreach(idx=iterators::icount(length(allele_groups))) %dopar% {
        # out_list <- lapply(1:length(allele_groups), function(idx) {
        gc()
        #message(paste0("idx=",idx))
        # Subset of data being analyzed
        allele_name <- names(allele_groups)[idx]
        germline <- germlines[allele_name]
        indices <- allele_groups[[allele_name]]
        db_subset <- data[indices, ]

        # If mutrange is auto, find most popular mutation count and start from there
        gpm <- db_subset %>%
            dplyr::mutate(!!v_call := allele_name ) %>%
            getPopularMutationCount(germline,
                                    gene_min=0, seq_min=min_seqs,
                                    seq_p_of_max=1/8, full_return=TRUE,
                                    v_call=v_call,
                                    seq=seq)

        # Determine the mutation range(s) to scan
        mut_mins <- min(mut_range)
        if ( auto_mutrange & sum(gpm$mutation_count > 0) > 0 ){
            mut_mins <- c(mut_mins, gpm$mutation_count[gpm$mutation_count > 0]) %>%
                unique() %>%
                sort()
        }

        # Create the run's return object
        df_run_empty <- data.frame(germline_call = names(germline),
                                  note = "",
                                  polymorphism_call = NA,
                                  nt_substitutions=NA,
                                  novel_imgt = NA,
                                  novel_imgt_count=NA,
                                  novel_imgt_unique_j=NA,
                                  novel_imgt_unique_cdr3=NA,
                                  perfect_match_count = NA,
                                  perfect_match_freq = NA,
                                  germline_call_count = length(indices),
                                  germline_call_freq = round(length(indices)/nrow(data), 3),
                                  mut_min = NA,
                                  mut_max = NA,
                                  mut_pass_count=NA,
                                  germline_imgt = as.character(germline),
                                  germline_imgt_count=NA,
                                  pos_min = min(pos_range),
                                  pos_max = max(pos_range),
                                  y_intercept = y_intercept,
                                  y_intercept_pass = NA,
                                  snp_pass=NA,
                                  unmutated_count=NA,
                                  unmutated_freq=NA,
                                  unmutated_snp_j_gene_length_count=NA,
                                  snp_min_seqs_j_max_pass=NA,
                                  alpha = alpha,
                                  min_seqs = min_seqs,
                                  j_max = j_max,
                                  min_frac = min_frac,
                                  stringsAsFactors = FALSE)
        for (mut_min in rev(mut_mins)) {
            gc()
            # message(paste0("|-- mut_min=",mut_min))
            if (mut_min == rev(mut_mins)[1]){
                df_run <- df_run_empty
            } else {
                df_run <- dplyr::bind_rows(df_run_empty, df_run)
            }
            mut_max <- mut_min + diff(range(mut_range))
            df_run$mut_min[1] <- mut_min
            df_run$mut_max[1] <- mut_max

            # If no sequence is frequent enough to pass the J test, give up now
            if(nrow(gpm) < 1) {
                df_run$note[1] <- "Plurality sequence too rare."
                if(mut_mins[1] == mut_min){
                    return(df_run)
                } else {
                    next
                }
            }

            # Add a mutation count column and filter out sequences not in our range
            db_subset_mm <- mutationRangeSubset(db_subset, germline,
                                               mut_min:mut_max, pos_range,
                                               seq=seq, pos_range_max = pos_range_max)
            df_run$mut_pass_count[1] <- nrow(db_subset_mm)

            if(nrow(db_subset_mm) < min_seqs){
                df_run$note[1] <- paste0("Insufficient sequences (",nrow(db_subset_mm),") in desired mutational range.")
                if(mut_mins[1] == mut_min){
                    return(df_run)
                } else {
                    next
                }
            }

            # Duplicate each sequence for all the positions to be analyzed
            # and find which positions are mutated
            pos_db <- positionMutations(db_subset_mm, germline, pos_range,
                                        seq=seq, pos_range_max = pos_range_max)

            # Find positional mut freq vs seq mut count
            pos_muts <- pos_db %>%
                dplyr::group_by(!!rlang::sym("POSITION")) %>%
                dplyr::mutate(PASS = mean(!! rlang::sym("OBSERVED")) >= min_frac) %>%
                dplyr::group_by(!!!rlang::syms(c("MUT_COUNT", "POSITION"))) %>%
                dplyr::summarise(POS_MUT_RATE = mean(!!rlang::sym("MUTATED"))*unique(!!rlang::sym("PASS")) ) %>%
                dplyr::ungroup()

            rm(pos_db)
            gc()

            # Calculate y intercepts, find which pass the test
            pass_y <- pos_muts %>%
                dplyr::group_by(!!rlang::sym("POSITION")) %>%
                dplyr::summarise(Y_INT_MIN = findLowerY(!!rlang::sym("POS_MUT_RATE"),
                                                        !!rlang::sym("MUT_COUNT"),
                                                          mut_min, alpha)) %>%
                dplyr::filter(!!rlang::sym("Y_INT_MIN") > y_intercept)

            df_run$y_intercept_pass[1] <- nrow(pass_y)

            if(nrow(pass_y) < 1){
                df_run$note[1] <- "No positions pass y-intercept test."
                if(mut_mins[1] == mut_min){
                    return(df_run)
                } else {
                    next
                }
            }

            gl_substring <- superSubstring(germline, pass_y$POSITION)
            gl_minus_substring <- insertPolymorphisms(germline, pass_y$POSITION,
                                                     rep("N", nrow(pass_y)))

            # Find the potential SNP positions and remove anything that matches
            # the germline at all those positions or any combo that is too rare
            #message(idx)
            #message(pass_y$POSITION)
            db_y_subset_mm <- db_subset_mm %>%
                dplyr::group_by(1:n()) %>%
                dplyr::mutate(SNP_STRING = superSubstring(!! rlang::sym(seq),
                                                            pass_y$POSITION),
                              POSITION = TRUE)
            if (!is.null(pos_range_max)) {
                db_y_subset_mm <- db_y_subset_mm %>%
                    dplyr::mutate(POSITION = !any(!!rlang::sym(pos_range_max) < pass_y$POSITION))
            }
            db_y_subset_mm <- db_y_subset_mm %>%
                dplyr::filter(!!rlang::sym("SNP_STRING") != gl_substring, !!rlang::sym("POSITION")) %>%
                dplyr::group_by(!!rlang::sym("SNP_STRING")) %>%
                dplyr::mutate(STRING_COUNT = n()) %>%
                dplyr::filter(!!rlang::sym("STRING_COUNT") >= min_seqs)

            df_run$snp_pass[1] <- nrow(db_y_subset_mm)

            if (nrow(db_y_subset_mm) < 1 ){
                df_run$note[1] <- paste("Position(s) passed y-intercept (",
                                       paste(pass_y$POSITION, collapse = ","),
                                       ") but the plurality sequence is too rare.",
                                       sep="")
                if(mut_mins[1] == mut_min){
                    return(df_run)
                } else {
                    next
                }
            }

            # Get mutation count at all positions that are not potential SNPs
            pads <- paste(rep("-", min(pos_range)-1), collapse="")
            db_y_subset_mm$MUT_COUNT_MINUS_SUBSTRING <- db_y_subset_mm[[seq]] %>%
                substring(min(pos_range), max(pos_range)) %>%
                paste(pads, ., sep="") %>%
                getMutatedPositions(gl_minus_substring) %>%
                sapply(length)

            # Keep only unmutated sequences and then find the counts of J and
            # junction length for each of the SNP strings, and then check to
            # see which pass the j/junction and count requirements
            db_y_summary0 <- db_y_subset_mm %>%
                dplyr::filter(!!rlang::sym("MUT_COUNT_MINUS_SUBSTRING") == 0)

            df_run$unmutated_count[1] <- nrow(db_y_summary0)

            db_y_summary0 <- db_y_summary0 %>%
                dplyr::mutate(J_GENE = getGene(!! rlang::sym(j_call))) %>%
                dplyr::group_by(!!! rlang::syms(c("SNP_STRING", "J_GENE", junction_length))) %>%
                dplyr::summarise(COUNT = n())

            df_run$unmutated_snp_j_gene_length_count[1] <- nrow(db_y_summary0)

            db_y_summary0 <- db_y_summary0 %>%
                dplyr::group_by(!!rlang::sym("SNP_STRING")) %>%
                dplyr::mutate(FRACTION = !! rlang::sym("COUNT")/sum(!! rlang::sym("COUNT"))) %>%
                dplyr::summarise(TOTAL_COUNT = sum(!! rlang::sym("COUNT")), MAX_FRAC = max(!! rlang::sym("FRACTION")))

            if(nrow(db_y_summary0) < 1){
                df_run$note[1] <- paste("Position(s) passed y-intercept (",
                                       paste(pass_y$POSITION, collapse = ","),
                                       ") but no unmutated versions of novel allele",
                                       " found.", sep="")
                if(mut_mins[1] == mut_min){
                    return(df_run)
                } else {
                    next
                }
            }

            # db_y_summary = db_y_summary0 %>%
            #   filter_(~TOTAL_COUNT >= min_seqs & MAX_FRAC <= j_max)

            min_seqs_pass <- db_y_summary0$TOTAL_COUNT >= min_seqs
            j_max_pass <- db_y_summary0$MAX_FRAC <= j_max

            db_y_summary <- db_y_summary0[min_seqs_pass & j_max_pass, , drop=FALSE]

            df_run$snp_min_seqs_j_max_pass[1] <- nrow(db_y_summary)

            if(nrow(db_y_summary) < 1){
                msg <- c(NA, NA)
                names(msg) <- c("j_max", "min_seqs")

                if (sum(min_seqs_pass) == 0) {
                    msg['min_seqs'] <- paste0("Not enough sequences (maximum total count is ",
                                              max(db_y_summary0$TOTAL_COUNT),
                                              ").")
                }

                if (sum(j_max_pass) == 0) {
                    msg['j_max'] <- paste0("A J-junction combination is too prevalent (",
                                           round(100*max(db_y_summary0$MAX_FRAC),1),"% of sequences).")
                }

                msg <- paste(na.omit(msg), collapse=" and ")
                df_run$note[1] <- paste("Position(s) passed y-intercept (",
                                       paste(pass_y$POSITION, collapse = ","),
                                       ") but ",
                                       msg,".", sep="")
                df_run$perfect_match_count[1] <- max(db_y_summary0$TOTAL_COUNT)
                df_run$perfect_match_freq[1] <- df_run$perfect_match_count[1]/df_run$germline_call_count[1]
                if(mut_mins[1] == mut_min){
                    return(df_run)
                } else {
                    next
                }
            }

            germ_nts <- unlist(strsplit(gl_substring,""))
            for (r in 1:nrow(db_y_summary)) {
                if (r > 1){
                    df_run <- dplyr::bind_rows(df_run[1,], df_run)
                }
                # Create the new germline
                snp_nts <- unlist(strsplit(db_y_summary$SNP_STRING[r],""))
                remain_mut <- db_y_summary$SNP_STRING[r] %>%
                    getMutatedPositions(gl_substring) %>%
                    unlist() %>%
                    unique()
                germ <- insertPolymorphisms(germline, pass_y$POSITION, snp_nts)
                is_known_allele <- germ == germlines
                if (sum(is_known_allele) == 0 ) {
                    names(germ) <- mapply(paste, germ_nts[remain_mut],
                                         pass_y$POSITION[remain_mut],
                                         snp_nts[remain_mut], sep="") %>%
                        paste(collapse="_") %>%
                        paste(names(germline), ., sep="_")
                } else {
                    # If the match is with duplicated sequences in the reference germlines,
                    # use the first
                    known_allele_names <- sortAlleles(names(germlines)[is_known_allele],
                                                      method="position")
                    names(germ) = known_allele_names[1]
                }
                # Save the new germline to our data frame
                df_run$polymorphism_call[1] <- names(germ)
                df_run$novel_imgt[1] <-  as.character(germ)
                df_run$perfect_match_count[1] <- db_y_summary$TOTAL_COUNT[r]
                df_run$perfect_match_freq[1] <- df_run$perfect_match_count[1]/df_run$germline_call_count[1]
                df_run$note[1] = "Novel allele found!"
            }

        } # end for each starting mutation counts
        return(df_run)

    } # end foreach allele

    if(nproc > 1) { stopCluster(cluster) }
    out_df <- dplyr::bind_rows(out_list)
    getMuSpec <- function(poly_call, germ_call) {
        sapply(1:length(poly_call), function(i){
            p <- gsub(germ_call[i], "", poly_call[i], fixed = T)
            p <- strsplit(p,"_")[[1]][-1]
            m <- gsub("([[:alpha:]])([[:digit:]]*)([[:alpha:]])", "\\2\\1>\\3", p)
            paste(m, collapse=",")
        })
    }

    # The number of records in the sequence dataset matching
    # each exact novel_imgt sequence
    getDbMatch <- function(novel_imgt) {
        novel_imgt <- stri_sub(novel_imgt,  min(pos_range), max(pos_range))
        novel_imgt <- stri_replace_all_regex(novel_imgt, "[-\\.]","")
        data_seq <- stri_sub(data[[seq]],  min(pos_range), max(pos_range))
        data_seq <- stri_replace_all_regex(data_seq,"[-\\.]","")
        sapply(novel_imgt, function(n) {
            # sum(grepl(n, data_seq))
            sum(stri_detect_fixed(data_seq,n))
        })
    }

    # The number of distinct J in the sequence dataset associated
    # with the exact novel_imgt sequence
    getNumJ <- function(novel_imgt) {
        novel_imgt <- stri_sub(novel_imgt,  min(pos_range), max(pos_range))
        novel_imgt <- stri_replace_all_regex(novel_imgt, "[-\\.]","")
        data_seq <- stri_sub(data[[seq]],  min(pos_range), max(pos_range))
        data_seq <- stri_replace_all_regex(data_seq,"[-\\.]","")
        sapply(novel_imgt, function(n) {
            imgt_idx <- stri_detect_fixed(data_seq,n)
            length(unique(getGene(data[[j_call]][imgt_idx])))
        })
    }


    # The number of distinct CDR3 in the sequence dataset associated
    # with the exact novel_imgt sequence
    getNumCDR3 <- function(novel_imgt) {
        novel_imgt <- stri_sub(novel_imgt,  min(pos_range), max(pos_range))
        novel_imgt <- stri_replace_all_regex(novel_imgt, "[-\\.]","")
        data_seq <- stri_sub(data[[seq]],  min(pos_range), max(pos_range))
        data_seq <- stri_replace_all_regex(data_seq,"[-\\.]","")
        sapply(novel_imgt, function(n) {
            imgt_idx <- stri_detect_fixed(data_seq,n)
            seq <- data[[junction]][imgt_idx]
            seq <- stri_sub(seq, 4, stringi::stri_length(seq) - 3)
            length(unique(seq))
        })
    }

    idx <- which(!is.na(out_df$novel_imgt))
    if (length(idx)>0) {
        out_df$nt_substitutions[idx] <- getMuSpec(out_df$polymorphism_call[idx],
                                                  out_df$germline_call[idx])
        out_df$novel_imgt_count[idx] <- getDbMatch(out_df$novel_imgt[idx])
        out_df$novel_imgt_unique_j[idx] <- getNumJ(out_df$novel_imgt[idx])
        if (junction %in% colnames(data)) {
            out_df$novel_imgt_unique_cdr3[idx] <- getNumCDR3(out_df$novel_imgt[idx])
        }
    }
    out_df$germline_imgt_count <- getDbMatch(out_df$germline_imgt)
    out_df$unmutated_freq <- out_df$unmutated_count/out_df$germline_call_count

    dup_novel_imgt <- which(duplicated(out_df[['novel_imgt']]) & !is.na(out_df[['polymorphism_call']]))
    dup_novel <- which(out_df[['novel_imgt']] %in% out_df[['novel_imgt']][dup_novel_imgt])

    if (length(dup_novel) > 0 ) {
        message("Duplicated polymorphism(s) found!. See the field 'note' in your results for more details.")
        for (i in dup_novel) {
            this_polymorphism <- out_df[['polymorphism_call']][i]
            this_novel_imgt <- out_df[['novel_imgt']][i]
            other <- which(out_df[['novel_imgt']] == this_novel_imgt)
            other_polymorphism_call <- setdiff(out_df[['polymorphism_call']][other],
                                               this_polymorphism)
            new_note <- paste0(other_polymorphism_call, collapse=", ")
            out_df[['note']][i] <- paste0(out_df[['note']][i],". Same as: ", new_note)
        }
    }

    return(out_df)
}

#' Select rows containing novel alleles
#'
#' \code{selectNovel} takes the result from \link{findNovelAlleles} and
#' selects only the rows containing unique, novel alleles.
#'
#' @details
#' If, for instance, subject has in his genome \code{IGHV1-2*02} and a novel
#' allele equally close to \code{IGHV1-2*02} and \code{IGHV1-2*05}, the novel allele may be
#' detected by analyzing sequences that best align to either of these alleles.
#' If \code{keep_alleles} is \code{TRUE}, both polymorphic allele calls will
#' be retained. In the case that multiple mutation ranges are checked for the
#' same allele, only one mutation range will be kept in the output.
#'
#' @param   novel           \code{data.frame} of the type returned by
#'                          \link{findNovelAlleles}.
#' @param   keep_alleles    \code{logical} indicating if different alleles
#'                          leading to the same novel sequence should be kept.
#'                          See Details.
#'
#' @return  A \code{data.frame} containing only unique, novel alleles (if any)
#' that were in the input.
#'
#' @examples
#' novel <- selectNovel(SampleNovel)
#'
#' @export
selectNovel <- function(novel, keep_alleles=FALSE) {
    # Remove non-novel rows
    novel <- filter(novel, !is.na(!!rlang::sym("novel_imgt")))

    if (keep_alleles) {
        novel_set <- novel %>%
            group_by(!!rlang::sym("germline_call")) %>%
            distinct(!!rlang::sym("novel_imgt"), .keep_all=TRUE) %>%
            ungroup()
    } else {
        novel_set <- novel %>%
            distinct(!!rlang::sym("novel_imgt"), .keep_all=TRUE) %>%
            ungroup()
    }

    return(novel_set)
}

#' Visualize evidence of novel V alleles
#'
#' \code{plotNovel} is be used to visualize the evidence of any novel V
#' alleles found using \link{findNovelAlleles}. It can also be used to
#' visualize the results for alleles that did
#'
#' @details
#' The first panel in the plot shows, for all sequences which align to a particular
#' germline allele, the mutation frequency at each position along the aligned
#' sequence as a function of the sequence-wide mutation count. Each line is a position.
#' Positions that contain polymorphisms (rather than somatic hypermutations)
#' will exhibit a high apparent mutation frequency for a range of
#' sequence-wide mutation counts. The positions are color coded as follows:
#'
#' \itemize{
#'   \item  red:    the position(s) pass(ess) the novel allele test
#'   \item  yellow: the position(s) pass(ess) the y-intercept test but not
#'                         other tests
#'   \item  blue:   the position(s) didn't pass the y-intercept test and
#'                         was(were) not further considered
#' }
#'
#' The second panel shows the nucleotide usage at each of the polymorphic positions
#' as a function of sequence-wide mutation count. If no polymorphisms were identified,
#' the panel will show the mutation count.
#'
#' To avoid cases where a clonal expansion might lead to a false positive, TIgGER examines
#' the combinations of J gene and junction length among sequences which perfectly
#' match the proposed germline allele. Clonally related sequences usually share
#' the same V gene, J gene and junction length. Requiring the novel allele
#' to be found in different combinations of J gene and junction lengths
#' is a proxy for requiring it to be found in different clonal lineages.
#'
#' @param    data             \code{data.frame} containing repertoire data. See
#'                            \link{findNovelAlleles} for details.
#' @param    novel_row        single row from a data frame as output by
#'                            \link{findNovelAlleles} that contains a
#'                            polymorphism-containing germline allele.
#' @param    v_call           name of the column in \code{data} with V allele
#'                            calls. Default is \code{v_call}.
#' @param    j_call           name of the column in \code{data} with J allele calls.
#'                            Default is \code{j_call}.
#' @param    seq              name of the column in \code{data} with the
#'                            aligned, IMGT-numbered, V(D)J nucleotide sequence.
#'                            Default is \code{sequence_alignment}.
#' @param    junction         Junction region nucleotide sequence, which includes
#'                            the CDR3 and the two flanking conserved codons. Default
#'                            is \code{junction}.
#' @param    junction_length  number of junction nucleotides in the junction sequence.
#'                            Default is \code{junction_length}.
#' @param    pos_range_max    Name of the column in \code{data} with the ending
#'                            positions of the V alignment in the germline
#'                            (usually \code{v_germline_end}).
#' @param    ncol             number of columns to use when laying out the plots.
#' @param    multiplot        whether to return one single plot (\code{TRUE}) or a list
#'                            with the three individual plots (\code{FALSE}).
#' @examples
#' # Plot the evidence for the first (and only) novel allele in the example data
#' novel <- selectNovel(SampleNovel)
#' plotNovel(AIRRDb, novel[1, ], v_call="v_call", j_call="j_call",
#'           seq="sequence_alignment", junction="junction", junction_length="junction_length",
#'           multiplot=TRUE)
#'
#' @export
plotNovel <- function(data, novel_row, v_call="v_call", j_call="j_call",
                      seq="sequence_alignment",
                      junction="junction", junction_length="junction_length",
                      pos_range_max=NULL,
                      ncol=1, multiplot=TRUE) {
    . = NULL

    # Use the data frame
    if(length(novel_row) > 0) {
        if(is.data.frame(novel_row) & nrow(novel_row) == 1) {
            pos_range <- novel_row$pos_min:novel_row$pos_max
            germline <- novel_row$germline_imgt
            names(germline) <- novel_row$germline_call
            mut_range <- novel_row$mut_min[1]:novel_row$mut_max[1]
            novel_imgt <- novel_row$novel_imgt
            names(novel_imgt) <- novel_row$polymorphism_call
            min_frac <- novel_row$min_frac
            note <- novel_row$note
        } else {
            stop("novel_row is not a data frame with only one row.")
        }
    }

    germline <- cleanSeqs(germline)
    data[[seq]] <- cleanSeqs(data[[seq]])

    # Extract sequences assigned to the germline, determine which
    # have an appropriate range of mutations, and find the mutation
    # frequency of each position
    db_subset <- data %>%
        select(!!!rlang::syms(c(seq, v_call, j_call, junction_length, ifelse(is.null(pos_range_max), "",pos_range_max)))) %>%
        filter(grepl(names(germline),  .data[[v_call]], fixed=TRUE))
    pos_db <- db_subset %>%
        mutationRangeSubset(germline, mut_range, pos_range, seq=seq, pos_range_max=pos_range_max)
    if (nrow(pos_db) == 0) {
        warning(paste0("Insufficient sequences (",nrow(pos_db),") in desired mutational range."))
        return (invisible(NULL))
    }
    pos_db <- pos_db %>%
        positionMutations(germline, pos_range,seq=seq, pos_range_max=pos_range_max)
    pos_muts <- pos_db %>%
        group_by(!!rlang::sym("POSITION")) %>%
        mutate(PASS = mean(!!rlang::sym("OBSERVED")) >= min_frac) %>%
        group_by(!!!rlang::syms(c("MUT_COUNT", "POSITION"))) %>%
        summarise(POS_MUT_RATE = mean(!!rlang::sym("MUTATED"))*unique(!!rlang::sym("PASS"))) %>%
        ungroup()

    # Label the polymorphic positions as such
    pass_y <- unlist(strsplit(names(novel_imgt), "_"))[-1] %>%
        gsub("[^0-9]", "", .) %>%
        as.numeric()
    p_y_f <- unlist(strsplit(names(novel_imgt), "_"))[-1] %>%
        gsub("[0-9]+.", "", .)
    p_y_t <- unlist(strsplit(names(novel_imgt), "_"))[-1] %>%
        gsub(".[0-9]+", "", .)
    # Parse the note to find positions that passed y intercept if no novel found
    if (length(pass_y) == 0 & grepl("Position\\(s\\) passed y-intercept", note)) {
        pass_y <- note %>%
            gsub("Position\\(s\\) passed y-intercept \\(", "", .) %>%
            gsub("\\).*", "", .) %>% strsplit(",") %>% unlist %>% as.numeric
        p_y_f <- sapply(pass_y, function (x) substring(germline, x, x))
        p_y_t <- gsub(".", "?", p_y_f)
    }

    to_from <- paste(paste("Position", pass_y), paste(paste(p_y_f, "->"), p_y_t))
    names(to_from) <- pass_y
    pos_muts <- pos_muts %>%
        mutate(Polymorphic = ifelse(!!rlang::sym("POSITION") %in% pass_y, "True", "False"))

    pads <- paste(rep("-", min(pos_range)-1), collapse="")
    db_subset$MUT_COUNT_NOVEL <- db_subset[[seq]] %>%
        substring(min(pos_range), max(pos_range)) %>%
        paste(pads, ., sep="") %>%
        getMutatedPositions(novel_imgt) %>%
        sapply(length)
    db_subset <- db_subset %>%
        filter(!!rlang::sym("MUT_COUNT_NOVEL") == 0) %>%
        mutate(J_GENE = getGene(!!rlang::sym(j_call)))
    if (nrow(db_subset) == 0) {
        warning(paste0("Insufficient sequences (",nrow(db_subset),") with MUT_COUNT_NOVEL == 0."))
        return (invisible(NULL))
    }
    db_subset[[junction_length]] <- db_subset[[junction_length]] %>%
        factor(levels=min(db_subset[[junction_length]]):max(db_subset[[junction_length]]))
    pos_muts$Polymorphic <- pos_muts$Polymorphic %>%
        factor(levels = c("False", "True"))
    pos_db$NT <- pos_db$NT %>%
        factor(levels = names(DNA_COLORS))
    pos_muts$GERMLINE <- names(germline)

    x_range <- range(pos_muts$MUT_COUNT)
    x_limits <- c(x_range[1] - 0.5, x_range[2] + 0.5)
    x_breaks <- seq(x_range[1], x_range[2], by = 1)

    # MAKE THE FIRST PLOT
    if (!is.na(novel_imgt)) {
        POLYCOLORS <- setNames(DNA_COLORS[c(4,3)], c("False", "True")) # blue #3C88EE, red #EB413C
        p1 <- ggplot(pos_muts,
                     aes(x=!!rlang::sym("MUT_COUNT"),
                         y=!!rlang::sym("POS_MUT_RATE"),
                         group=!!rlang::sym("POSITION"),
                         color=!!rlang::sym("Polymorphic"))) +
            geom_line(data=filter(pos_muts, !!rlang::sym("Polymorphic") == "False"), linewidth=0.75) +
            geom_line(data=filter(pos_muts, !!rlang::sym("Polymorphic") == "True"), linewidth=0.75) +
            facet_grid(GERMLINE ~ .) +
            scale_color_manual(values = POLYCOLORS) +
            scale_x_continuous(limits = x_limits, breaks = x_breaks) +
            ylim(0,1) +
            xlab("Mutation Count (Sequence)") +
            ylab("Mutation Frequency (Position)") +
            theme_bw() +
            theme(legend.position=c(0.5,0.9), legend.justification=c(0.5,1),
                  legend.background=element_rect(fill = "transparent")) +
            guides(color = guide_legend(ncol = 2, reverse = TRUE))
    } else{
        POLYCOLORS <- setNames(DNA_COLORS[c(4,2)], c("False", "True")) # blue #3C88EE, yellow #FFB340
        p1 <- ggplot(pos_muts,
                     aes(x=!!rlang::sym("MUT_COUNT"),
                         y=!!rlang::sym("POS_MUT_RATE"),
                         group=!!rlang::sym("POSITION"),
                         color=!!rlang::sym("Polymorphic"))) +
            geom_line(linewidth=0.75) +
            facet_grid(GERMLINE ~ .) +
            scale_color_manual(values=POLYCOLORS) +
            scale_x_continuous(limits = x_limits, breaks = x_breaks) +
            ylim(0, 1) +
            xlab("Mutation Count (Sequence)") +
            ylab("Mutation Frequency (Position)") +
            theme_bw() +
            theme(legend.position=c(0.5, 0.9), legend.justification=c(0.5, 1),
                  legend.background=element_rect(fill="transparent")) +
            guides(color = guide_legend("Passed y-intercept test",
                                        ncol = 2, reverse = TRUE))
    }
    # MAKE THE SECOND PLOT
    p2_data <- mutate(filter(pos_db, !!rlang::sym("POSITION") %in% pass_y),
                      POSITION = to_from[as.character(!!rlang::sym("POSITION"))])
    positions <- unique(p2_data$POSITION)
    numeric_positions <- as.numeric(sub("Position ([0-9]+) .+","\\1",positions))

    p2_data$POSITION <- factor(p2_data$POSITION,
                               levels=positions[order(numeric_positions)],
                               ordered = TRUE)

    if (nrow(p2_data)) {
        p2 <- ggplot(p2_data, aes(x=!!rlang::sym("MUT_COUNT"),
                                  fill=!!rlang::sym("NT"))) +
            geom_bar(width=0.9) +
            guides(fill = guide_legend("Nucleotide", ncol=4)) +
            xlab("Mutation Count (Sequence)") +
            ylab("Sequence Count") +
            scale_x_continuous(limits = x_limits, breaks = x_breaks) +
            scale_fill_manual(values=DNA_COLORS, breaks=names(DNA_COLORS),
                              drop=FALSE) +
            theme_bw() +
            theme(legend.position=c(1,1), legend.justification=c(1,1),
                  legend.background=element_rect(fill="transparent"))
    } else {
        p2_data <- mutate(filter(pos_db,
                                  !!rlang::sym("POSITION") %in% names(which.max(table(pos_db$POSITION)))),
                          POSITION = "No positions pass y-intercept test.")
        p2 <- ggplot(p2_data, aes(x=!!rlang::sym("MUT_COUNT"))) +
            geom_bar(width=0.9) +
            xlab("Mutation Count (Sequence)") + ylab("Sequence Count") +
            scale_x_continuous(limits = x_limits, breaks = x_breaks) +
            theme_bw() +
            theme(legend.position=c(1,1), legend.justification=c(1,1),
                  legend.background=element_rect(fill = "transparent"))
    }

    if (length(positions) < 5) {
        p2 <- p2 +
            facet_grid(POSITION ~ .)
        ncols <- 1
        nrows <- length(positions)
    } else {
        ncols <- min(length(positions),6)
        nrows <- ceiling(length(positions)/ncols)
        p2 <- p2 +
            facet_wrap(POSITION ~ ., ncol=ncols)
    }

    # MAKE THE THIRD PLOT
    p3 <- ggplot(db_subset, aes(x=!!rlang::sym(junction_length),
                                fill=!!rlang::sym("J_GENE"))) +
        geom_bar(width=0.9) +
        guides(fill=guide_legend("J Gene", ncol=2)) +
        xlab("Junction Length") +
        ylab("Unmutated Sequence Count") +
        theme_bw() +
        theme(legend.position=c(1, 1), legend.justification=c(1, 1),
              legend.background=element_rect(fill="transparent"))

    p2_height <- max(1,0.6*nrows )

    heights <- c(1, p2_height, 1)
    if (multiplot) {
        multiplot(p1, p2, p3, cols = ncol, heights=heights)
    } else {
        list(p1, p2, p3)
    }
}

#' Infer a subject-specific genotype using a frequency method
#'
#' \code{inferGenotype} infers an subject's genotype using a frequency method.
#' The genotype is inferred by finding the minimum number set of alleles that
#' can explain the majority of each gene's calls. The most common allele of
#' each gene is included in the genotype first, and the next most common allele
#' is added until the desired fraction of alleles can be explained. In this
#' way, mistaken allele calls (resulting from sequences which
#' by chance have been mutated to look like another allele) can be removed.
#'
#' @details
#' Allele calls representing cases where multiple alleles have been
#' assigned to a single sample sequence are rare among unmutated
#' sequences but may result if nucleotides for certain positions are
#' not available. Calls containing multiple alleles are treated as
#' belonging to all groups. If \code{novel} is provided, all
#' sequences that are assigned to the same starting allele as any
#' novel germline allele will have the novel germline allele appended
#' to their assignment prior to searching for unmutated sequences.
#'
#' @param    data                 \code{data.frame} containing V allele
#'                                calls from a single subject.
#' @param    germline_db          named vector of sequences containing the
#'                                germline sequences named in
#'                                \code{allele_calls}. Only required if
#'                                \code{find_unmutated} is \code{TRUE}.
#' @param    novel                optional \code{data.frame} of the type
#'                                novel returned by
#'                                \link{findNovelAlleles} containing
#'                                germline sequences that will be utilized if
#'                                \code{find_unmutated} is \code{TRUE}. See
#'                                Details.
#' @param    v_call               column in \code{data} with V allele calls.
#'                                Default is \code{"v_call"}.
#' @param    seq                  name of the column in \code{data} with the
#'                                aligned, IMGT-numbered, V(D)J nucleotide sequence.
#'                                Default is \code{sequence_alignment}.
#' @param    fraction_to_explain  the portion of each gene that must be
#'                                explained by the alleles that will be included
#'                                in the genotype.
#' @param    gene_cutoff          either a number of sequences or a fraction of
#'                                the length of \code{allele_calls} denoting the
#'                                minimum number of times a gene must be
#'                                observed in \code{allele_calls} to be included
#'                                in the genotype.
#' @param    find_unmutated       if \code{TRUE}, use \code{germline_db} to
#'                                find which samples are unmutated. Not needed
#'                                if \code{allele_calls} only represent
#'                                unmutated samples.
#'
#' @return
#' A \code{data.frame} of alleles denoting the genotype of the subject containing
#' the following columns:
#'
#' \itemize{
#'   \item \code{gene}: The gene name without allele.
#'   \item \code{alleles}: Comma separated list of alleles for the given \code{gene}.
#'   \item \code{counts}: Comma separated list of observed sequences for each
#'         corresponding allele in the \code{alleles} list.
#'   \item \code{total}: The total count of observed sequences for the given \code{gene}.
#'   \item \code{note}: Any comments on the inference.
#' }
#'
#' @note
#' This method works best with data derived from blood, where a large
#' portion of sequences are expected to be unmutated. Ideally, there
#' should be hundreds of allele calls per gene in the input.
#'
#' @seealso \link{plotGenotype} for a colorful visualization and
#'          \link{genotypeFasta} to convert the genotype to nucleotide sequences.
#'          See \link{inferGenotypeBayesian} to infer a subject-specific genotype
#'          using a Bayesian approach.
#'
#' @examples
#' # Infer IGHV genotype, using only unmutated sequences, including novel alleles
#' inferGenotype(AIRRDb, germline_db=SampleGermlineIGHV, novel=SampleNovel,
#'               find_unmutated=TRUE)
#'
#' @export
inferGenotype <- function(data, germline_db=NA, novel=NA, v_call="v_call",
                          seq="sequence_alignment",
                          fraction_to_explain=0.875, gene_cutoff=1e-4,
                          find_unmutated=TRUE) {

    . = NULL


    # Check columns exist(can be NULL)
    check <- checkColumns(data, c(v_call, seq))
    if (check != TRUE) { stop(check) }

    allele_calls = getAllele(data[[v_call]], first=FALSE, strip_d=FALSE)
    # Find the unmutated subset, if requested
    if (find_unmutated) {
        if(is.na(germline_db[1])){
            stop("germline_db needed if find_unmutated is TRUE")
        }
        if (!is.null(nrow(novel))) {
            novel <- filter(novel, !is.na(!!rlang::sym("polymorphism_call"))) %>%
                select(!!!rlang::syms(c("germline_call", "polymorphism_call", "novel_imgt")))
            if (nrow(novel) > 0) {
                # Extract novel alleles if any and add them to germline_db
                novel_gl <- novel$novel_imgt
                names(novel_gl) <- novel$polymorphism_call
                germline_db <- c(germline_db, novel_gl)
                # Add the novel allele calls to allele calls of the same starting allele
                for(r in 1:nrow(novel)){
                    ind <- grep(novel$germline_call[r], allele_calls, fixed=TRUE)
                    allele_calls[ind] <- allele_calls[ind] %>%
                        sapply(paste, novel$polymorphism_call[r], sep=",")
                }
            }
        }
        # Find unmutated sequences
        allele_calls <- findUnmutatedCalls(allele_calls,
                                          as.character(data[[seq]]),
                                          germline_db)
        if(length(allele_calls) == 0){
            stop("No unmutated sequences found! Set 'find_unmutated' to 'FALSE'.")
        }
    }

    # Find which rows' calls contain which genes
    cutoff <- ifelse(gene_cutoff < 1, length(allele_calls)*gene_cutoff, gene_cutoff)
    gene_regex <- allele_calls %>% strsplit(",") %>% unlist() %>%
        getGene(strip_d=FALSE) %>%  unique() %>% paste("\\*", sep="")
    gene_groups <- sapply(gene_regex, grep, allele_calls, simplify=FALSE)
    names(gene_groups) <- gsub("\\*", "", gene_regex, fixed=TRUE)
    gene_groups <- gene_groups[sapply(gene_groups, length) >= cutoff]
    gene_groups <- gene_groups[sortAlleles(names(gene_groups))]

    # Make a table to store the resulting genotype
    gene <- names(gene_groups)
    alleles <- counts <- note <- rep("", length(gene))
    total <- sapply(gene_groups, length)
    genotype <- cbind(gene, alleles, counts, total, note)

    # For each gene, find which alleles to include
    for (g in gene) {
        # Keep only the part of the allele calls that uses the gene being analyzed
        ac <- allele_calls[gene_groups[[g]]] %>%
            strsplit(",") %>%
            lapply(function(x) x[grep(paste(g, "\\*", sep=""), x)]) %>%
            sapply(paste, collapse=",")
        target <- ceiling(fraction_to_explain*length(ac)) # how many we need to explain
        t_ac <- table(ac) # table of allele calls
        potentials <- unique(unlist(strsplit(names(t_ac),","))) # potential alleles
        # One allele? Easy!
        if (length(potentials) == 1 | length(t_ac) == 1) {
            genotype[genotype[,"gene"]==g,"alleles"] <- gsub("[^d\\*]*[d\\*]","",potentials )[1]
            genotype[genotype[,"gene"]==g,"counts"] <- t_ac
        } else {
            # More alleles? Let's find the fewest that can explain the needed fraction
            # Make a table of which alleles can explain which calls
            regexpotentials <- paste(gsub("\\*","\\\\*", potentials),"$",sep="")
            regexpotentials <-
                paste(regexpotentials,gsub("\\$",",",regexpotentials),sep="|")
            tmat =
                sapply(regexpotentials, function(x) grepl(x, names(t_ac),fixed=FALSE))
            seqs_expl = as.data.frame(apply(tmat, 2, function(x) x*t_ac))
            colnames(seqs_expl) = potentials

            # Cycle through the table, including alleles to explain more sequences,
            # until we explain enough sequences
            included <- counts <- character(0)
            tot_expl = 0
            while(tot_expl < target){
                allele_tot <- apply(seqs_expl, 2, sum)
                included <- c(included, names(which.max(allele_tot)))
                counts <- c(counts, max(allele_tot))
                tot_expl <- max(allele_tot)  + tot_expl
                seqs_expl <- seqs_expl[which(seqs_expl[,which.max(allele_tot)]==0),]
            }
            genotype[genotype[,"gene"]==g,"alleles"] <-
                paste(gsub("[^d\\*]*[d\\*]","",included ),collapse=",")
            genotype[genotype[,"gene"]==g,"counts"] <-
                paste(counts,collapse=",")
        }
    }

    geno <- as.data.frame(genotype, stringsAsFactors = FALSE)

    # Check for indistinguishable calls
    if (find_unmutated == TRUE) {
        seqs <- genotypeFasta(geno, germline_db)
        dist_mat <- seqs %>%
            sapply(function(x) sapply((getMutatedPositions(seqs, x)), length)) %>%
            as.matrix
        rownames(dist_mat) <- colnames(dist_mat)
        for (i in 1:nrow(dist_mat)){ dist_mat[i,i] = NA }
        same <- which(dist_mat == 0, arr.ind=TRUE)
        if (nrow(same) > 0 ) {
            for (r in 1:nrow(same)) {
                inds <- as.vector(same[r,])
                geno[getGene(rownames(dist_mat)[inds][1]),]$note <-
                    paste(rownames(dist_mat)[inds], collapse=" and ") %>%
                    paste("Cannot distinguish", .)
            }
        }
    }
    rownames(geno) <- NULL

    return(geno)
}


#' Show a colorful representation of a genotype
#'
#' \code{plotGenotype} plots a genotype table.
#'
#' @param    genotype     \code{data.frame} of alleles denoting a genotype,
#'                        as returned by \link{inferGenotype}.
#' @param    facet_by     column name in \code{genotype} to facet the plot by.
#'                        if \code{NULL}, then do not facet the plot.
#' @param    gene_sort    string defining the method to use when sorting alleles.
#'                        if \code{"name"} then sort in lexicographic order. If
#'                        \code{"position"} then sort by position in the locus, as
#'                        determined by the final two numbers in the gene name.
#' @param    text_size    point size of the plotted text.
#' @param    silent       if \code{TRUE} do not draw the plot and just return the ggplot
#'                        object; if \code{FALSE} draw the plot.
#' @param    ...          additional arguments to pass to ggplot2::theme.
#'
#' @return  A ggplot object defining the plot.
#'
#' @seealso \link{inferGenotype}
#'
#' @examples
#' # Plot genotype
#' plotGenotype(SampleGenotype)
#'
#' # Facet by subject
#' genotype_a <- genotype_b <- SampleGenotype
#' genotype_a$SUBJECT <- "A"
#' genotype_b$SUBJECT <- "B"
#' geno_sub <- rbind(genotype_a, genotype_b)
#' plotGenotype(geno_sub, facet_by="SUBJECT", gene_sort="pos")
#'
#' @export
plotGenotype <- function(genotype, facet_by=NULL, gene_sort=c("name", "position"),
                         text_size=12, silent=FALSE, ...) {
    # Check arguments
    gene_sort <- match.arg(gene_sort)

    # Split genes' alleles into their own rows
    alleles = strsplit(genotype$alleles, ",")
    geno2 = genotype
    r = 1
    for (g in 1:nrow(genotype)){
        for(a in 1:length(alleles[[g]])) {
            geno2[r, ] = genotype[g, ]
            geno2[r, ]$alleles = alleles[[g]][a]
            r = r + 1
        }
    }

    # Set the gene order
    geno2$gene = factor(geno2$gene,
                        levels=rev(sortAlleles(unique(geno2$gene), method=gene_sort)))

    # Create the base plot
    p = ggplot(geno2, aes(x=!!rlang::sym("gene"),
                          fill=!!rlang::sym("alleles"))) +
        theme_bw() +
        theme(axis.ticks=element_blank(),
              axis.text.x=element_blank(),
              panel.grid.major=element_blank(),
              panel.grid.minor=element_blank(),
              text=element_text(size=text_size),
              strip.background=element_blank(),
              strip.text=element_text(face="bold")) +
        geom_bar(position="fill") +
        coord_flip() + xlab("Gene") + ylab("") +
        scale_fill_hue(name="Allele", h=c(0, 270), h.start=10)

    # Plot, with facets by SUBJECT if that column is present
    if (!is.null(facet_by)) {
        p = p + facet_grid(paste0(".~", facet_by))
    }

    # Add additional theme elements
    p = p + do.call(theme, list(...))

    # Plot
    if (!silent) { plot(p) }

    invisible(p)
}

#' Return the nucleotide sequences of a genotype
#'
#' \code{genotypeFasta} converts a genotype table into a vector of nucleotide
#' sequences.
#'
#' @param    genotype     \code{data.frame} of alleles denoting a genotype,
#'                        as returned by \link{inferGenotype}.
#' @param    germline_db  vector of named nucleotide germline sequences
#'                        matching the alleles detailed in \code{genotype}.
#' @param    novel        an optional \code{data.frame} containing putative
#'                        novel alleles of the type returned by
#'                        \link{findNovelAlleles}.
#'
#' @return   A named vector of strings containing the germline nucleotide
#'           sequences of the alleles in the provided genotype.
#'
#' @seealso \link{inferGenotype}
#'
#' @examples
#' # Find the sequences that correspond to the genotype
#' genotype_db <- genotypeFasta(SampleGenotype, SampleGermlineIGHV, SampleNovel)
#'
#' @export
genotypeFasta <- function(genotype, germline_db, novel=NA){
    if(!is.null(nrow(novel))){
        # Extract novel alleles if any and add them to germline_db
        novel <- filter(novel, !is.na(!!rlang::sym("polymorphism_call"))) %>%
            select(!!!rlang::syms(c("germline_call", "polymorphism_call", "novel_imgt")))
        if(nrow(novel) > 0){
            novel_gl <- novel$novel_imgt
            names(novel_gl) <- novel$polymorphism_call
            germline_db <- c(germline_db, novel_gl)
        }
    }

    genotype$gene <- getGene(genotype$gene, first = T, strip_d = T)
    g_names <- names(germline_db)
    names(g_names) <- getAllele(names(germline_db), first = T, strip_d = T)

    table_calls <- mapply(paste, genotype$gene, strsplit(genotype$alleles, ","),
                         sep="*")
    table_calls_names <- unlist(table_calls)
    seq_names <- g_names[names(g_names) %in% table_calls_names]
    seqs <- germline_db[seq_names]
    not_found <- !table_calls_names %in% names(g_names)

    if ( any(not_found) ) {
        stop("The following genotype alleles were not found in germline_db: ",
             paste(table_calls_names[not_found], collapse = ", "))
    }

    return(seqs)
}

#' Correct allele calls based on a personalized genotype
#'
#' \code{reassignAlleles} uses a subject-specific genotype to correct
#' correct preliminary allele assignments of a set of sequences derived
#' from a single subject.
#'
#' @details
#' In order to save time, initial gene assignments are preserved and
#' the allele calls are chosen from among those provided in \code{genotype_db},
#' based on a simple alignment to the sample sequence.
#'
#' @param    data          \code{data.frame} containing V allele calls from a
#'                         single subject and the sample IMGT-gapped V(D)J sequences under
#'                         \code{seq}.
#' @param    genotype_db   vector of named nucleotide germline sequences
#'                         matching the calls detailed in \code{allele_calls}
#'                         and personalized to the subject
#' @param    v_call        name of the column in \code{data} with V allele
#'                         calls. Default is \code{v_call}.
#' @param    seq           name of the column in \code{data} with the
#'                         aligned, IMGT-numbered, V(D)J nucleotide sequence.
#'                         Default is SEQUENCE_IMGT
#' @param    method        method to use when realigning sequences to
#'                         the genotype_db sequences. Currently, only \code{"hamming"}
#'                         (for Hamming distance) is implemented.
#' @param    path          directory containing the tool used in the
#'                         realignment method, if needed. Hamming distance does
#'                         not require a path to a tool.
#' @param    keep_gene     string indicating if the gene (\code{"gene"}),
#'                         family (\code{"family"}) or complete repertoire
#'                         (\code{"repertoire"}) assignments should be performed.
#'                         Use of \code{"gene"} increases speed by minimizing required number of
#'                         alignments, as gene level assignments will be maintained when possible.
#'
#' @return   A modified input \code{data.frame} containing the best allele call from
#'           among the sequences listed in \code{genotype_db} in the
#'           \code{v_call_genotyped} column.
#'
#' @examples
#' # Extract the database sequences that correspond to the genotype
#' genotype_db <- genotypeFasta(SampleGenotype, SampleGermlineIGHV, novel=SampleNovel)
#'
#' # Use the personalized genotype to determine corrected allele assignments
#' output_db <- reassignAlleles(AIRRDb, genotype_db, v_call="v_call",
#'                              seq="sequence_alignment")
#'
#' @export
reassignAlleles <- function(data, genotype_db, v_call="v_call",
                            seq="sequence_alignment",
                            method="hamming", path=NA,
                            keep_gene=c("gene", "family", "repertoire")){
    # Check arguments
    keep_gene <- match.arg(keep_gene)

    # Extract data subset and prepare output vector
    v_sequences = as.character(data[[seq]])
    v_calls = getAllele(data[[v_call]], first=FALSE, strip_d=FALSE)
    v_call_genotyped = rep("", length(v_calls))

    if (keep_gene == "gene") {
        v = getGene(v_calls, first = TRUE, strip_d=FALSE)
        geno = getGene(names(genotype_db),strip_d=TRUE)
        names(geno) = names(genotype_db)
    } else if (keep_gene == "family") {
        v <- getFamily(v_calls, first = TRUE, strip_d = FALSE)
        geno = getFamily(names(genotype_db),strip_d=TRUE)
        names(geno) = names(genotype_db)
    } else if (keep_gene == "repertoire") {
        v <- rep(v_call, length(v_calls))
        geno = rep(v_call, length(genotype_db))
        names(geno) = names(genotype_db)
    } else {
        stop("Unknown keep_gene value: ", keep_gene)
    }

    # keep_gene == FALSE
    # Find which genotype genes/families are homozygous and assign those alleles first
    hetero = unique(geno[which(duplicated(geno))])
    homo = geno[!(geno %in% hetero)]
    homo_alleles = names(homo)
    names(homo_alleles) = homo
    homo_calls_i = which(v %in% homo)
    v_call_genotyped[homo_calls_i] = homo_alleles[v[homo_calls_i]]

    # Now realign the heterozygote sequences to each allele of that gene
    for (het in hetero){
        ind = which(v %in% het)
        if (length(ind) > 0){
            het_alleles = names(geno[which(geno == het)])
            het_seqs = genotype_db[het_alleles]
            if(method == "hamming"){
                dists = lapply(het_seqs, function(x)
                    sapply(getMutatedPositions(v_sequences[ind], x, match_instead=FALSE),
                           length))
                dist_mat = matrix(unlist(dists), ncol = length(het_seqs))
            } else {
                stop("Only Hamming distance is currently supported as a method.")
            }
            # The sapply-apply approach could become problematic when nrow(dist_mat)
            # is 1 and min(best_match) has multiple values, due to the fact that R
            # does not always keep data structures unmutable
            # Explicitly specifying a list and subsequently keeping it as a list by
            # using lapply avoids that problem
            best_match = vector("list", length=nrow(dist_mat))
            for (i in 1:nrow(dist_mat)) {
                best_match[[i]] = which(dist_mat[i, ]==min(dist_mat[i, ]))
            }
            best_alleles = lapply(best_match, function(x) het_alleles[x])
            v_call_genotyped[ind] = unlist(lapply(best_alleles, paste, collapse=","))
        }
    }

    # Now realign the gene-not-in-genotype calls to every genotype allele
    hetero_calls_i = which(v %in% hetero)
    not_called = setdiff(1:length(v), c(homo_calls_i, hetero_calls_i))
    if(length(not_called)>1){
        if(method ==  "hamming"){
            dists = lapply(genotype_db, function(x)
                sapply(getMutatedPositions(v_sequences[not_called], x, match_instead=FALSE),
                       length))
            dist_mat = matrix(unlist(dists), ncol = length(genotype_db))
        } else {
            stop("Only Hamming distance is currently supported as a method.")
        }
        # The sapply-apply approach could become problematic when nrow(dist_mat)
        # is 1 and min(best_match) has multiple values, due to the fact that R
        # does not always keep data structures unmutable
        # Explicitly specifying a list and subsequently keeping it as a list by
        # using lapply avoids that problem
        best_match = vector("list", length=nrow(dist_mat))
        for (i in 1:nrow(dist_mat)) {
            best_match[[i]] = which(dist_mat[i, ]==min(dist_mat[i, ]))
        }
        best_alleles = lapply(best_match, function(x) names(genotype_db[x]))
        v_call_genotyped[not_called] = unlist(lapply(best_alleles, paste, collapse=","))
    }

    if (all(v_call_genotyped == data[[v_call]])) {
        msg <- ("No allele assignment corrections made.")
        if (all(v %in% homo) & length(hetero) > 0) {
            keep_opt <- eval(formals(reassignAlleles)$keep_gene)
            i <- match(keep_gene, keep_opt)
            rec_opt <- paste(keep_opt[(i+1):length(keep_opt)], collapse = ", ")
            msg <- paste(msg, "Consider setting keep_gene to one of:", rec_opt)
        }
        warning(msg)
    }

    data$v_call_genotyped <- v_call_genotyped

    return(data)
}


# Other Mutation-Related Functions ----------------------------------------

#' Find the location of mutations in a sequence
#'
#' \code{getMutatedPositions} takes two vectors of aligned sequences and
#' compares pairs of sequences. It returns a list of the nucleotide positions of
#' any differences.
#'
#' @param    samples        vector of strings representing aligned sequences
#' @param    germlines      vector of strings representing aligned sequences
#'                          to which \code{samples} will be compared. If only
#'                          one string is submitted, it will be used for all
#'                          \code{samples}.
#' @param    ignored_regex  regular expression indicating what characters
#'                          should be ignored (such as gaps and N nucleotides).
#' @param    match_instead  if \code{TRUE}, the function returns the positions
#'                          that are the same instead of those that are
#'                          different.
#' @return   A list of the nucleotide positions of any differences between the
#'           input vectors.
#'
#' @examples
#' # Create strings to act as a sample sequences and a reference sequence
#' seqs <- c("----GATA", "GAGAGAGA", "TANA")
#' ref <- "GATAGATA"
#'
#' # Find the differences between the two
#' getMutatedPositions(seqs, ref)
#'
#' @export
getMutatedPositions <- function(samples, germlines, ignored_regex="[\\.N-]",
                                match_instead=FALSE) {

    # If only one germline sequence is given, use it for all the sample seqs
    if(length(germlines) == 1){ germlines = rep(germlines, length(samples)) }
    if(length(samples) != length(germlines)) {
        stop("Number of input sequences does not match number of germlines.")
    }

    # Truncate each pair of sequences to the length of the shorter
    germ_mins = lapply(germlines, nchar)
    samp_mins = lapply(samples, nchar)
    min_lens = mapply(min, germ_mins, samp_mins)
    germ = toupper(mapply(substr, germlines, 1, min_lens, SIMPLIFY=FALSE))
    samp = toupper(mapply(substr, samples, 1, min_lens, SIMPLIFY=FALSE))

    # Calculate positions of mutations (or matches), ignoring gaps, Ns, and CDR3
    samp_char = strsplit(samp,"")
    germ_char = strsplit(germ,"")
    if(!match_instead){
        muts = lapply(mapply("!=", samp_char, germ_char, SIMPLIFY=FALSE), which)
    } else {
        muts = lapply(mapply("==", samp_char, germ_char, SIMPLIFY=FALSE), which)
    }
    ignore_germ = gregexpr(ignored_regex, germ)
    ignore_samp = gregexpr(ignored_regex, samp)
    ignore = mapply(c, ignore_germ, ignore_samp, SIMPLIFY=FALSE)

    muts = mapply(function(x, y) x[!x%in%y], muts, ignore, SIMPLIFY=FALSE)
    return(muts)
}


#' Determine the mutation counts from allele calls
#'
#' \code{getMutCount} takes a set of nucleotide sequences and their allele calls
#' and determines the distance between that sequence and any germline alleles
#' contained within the call
#'
#' @param    samples       vector of IMGT-gapped sample V sequences
#' @param    allele_calls  vector of strings representing Ig allele calls for
#'                         the sequences in \code{samples}, where multiple
#'                         calls are separated by a comma
#' @param    germline_db   vector of named nucleotide germline sequences
#'                         matching the calls detailed in \code{allele_calls}
#'
#' @return   A list equal in length to \code{samples}, containing the Hamming
#'           distance to each germline allele contained within each call within
#'           each element of \code{samples}
#'
#' @examples
#' # Insert a mutation into a germline sequence
#' s2 <- s3 <- SampleGermlineIGHV[1]
#' stringi::stri_sub(s2, 103, 103) <- "G"
#' stringi::stri_sub(s3, 107, 107) <- "C"
#'
#' sample_seqs <- c(SampleGermlineIGHV[2], s2, s3)
#'
#' # Pretend that one sample sequence has received an ambiguous allele call
#' sample_alleles <- c(paste(names(SampleGermlineIGHV[1:2]), collapse=","),
#'                     names(SampleGermlineIGHV[2]),
#'                     names(SampleGermlineIGHV[1]))
#'
#' # Compare each sequence to its assigned germline(s) to determine the distance
#' getMutCount(sample_seqs, sample_alleles, SampleGermlineIGHV)
#'
#' @export
getMutCount <- function(samples, allele_calls, germline_db){

    call_list = strsplit(allele_calls, ",")

    germline_list = lapply(call_list, function(x) germline_db[x])

    mut_pos_list = list()
    mut_count_list = list()
    # First, find mutations of all sequences with call count of 1
    call_count = sapply(germline_list, length)
    cc1 = which(call_count == 1)
    if (length(cc1) > 0) {
        mut_pos_list[cc1] = getMutatedPositions(samples[cc1],
                                                unlist(germline_list[cc1]))
        mut_count_list[cc1] = lapply(mut_pos_list[cc1], length)
    }
    # Then find mutations of all sequences with call count > 1
    ccm = which(call_count > 1)
    if (length(ccm) > 0){
        mut_pos_list[ccm] = mapply(getMutatedPositions,
                                   germline_list[ccm], samples[ccm],
                                   SIMPLIFY=FALSE)
        mut_count_list[ccm] = lapply(mut_pos_list[ccm],
                                     function(x) lapply(x,length))
    }

    return(mut_count_list)
}

#' Determine which calls represent an unmutated allele
#'
#' \code{findUnmutatedCalls} determines which allele calls would represent a
#' perfect match with the germline sequence, given a vector of allele calls and
#' mutation counts. In the case of multiple alleles being assigned to a
#' sequence, only the subset that would represent a perfect match is returned.
#'
#' @param    allele_calls   vector of strings representing Ig allele calls,
#'                          where multiple calls are separated by a comma.
#' @param    germline_db    vector of named nucleotide germline sequences
#' @param    sample_seqs    V(D)J-rearranged sample sequences matching the order
#'                          of the given \code{allele_calls}.
#'
#' @return   A vector of strings containing the members of \code{allele_calls}
#'           that represent unmutated sequences.
#'
#' @examples
#' # Find which of the sample alleles are unmutated
#' calls <- findUnmutatedCalls(AIRRDb$v_call, AIRRDb$sequence_alignment,
#'                             germline_db=SampleGermlineIGHV)
#'
#' @export
findUnmutatedCalls <- function(allele_calls, sample_seqs, germline_db){
    . = NULL
    allele_calls = getAllele(allele_calls, first = FALSE)
    sample_seqs = as.character(sample_seqs)

    # Remove calls not in germline_db
    not_in_db = allele_calls %>%
        strsplit(",") %>%
        unlist %>%
        setdiff(names(germline_db))
    no_call = which(allele_calls == "")
    in_db = not_in_db %>%
        sapply(grep, allele_calls, fixed=TRUE) %>%
        unlist() %>%
        c(no_call) %>%
        unique() %>%
        setdiff(1:length(allele_calls), .)
    allele_calls = allele_calls[in_db]
    sample_seqs = sample_seqs[in_db]

    mut_counts = getMutCount(sample_seqs, allele_calls, germline_db)

    # Find which seqs are unmutated and which of the allele calls that represents
    unmut_i = which(sapply(mut_counts, function(x) min(unlist(x))) == 0)
    which_no_muts = sapply(mut_counts, function(x) grep("^0$", unlist(x)) )
    unmut_alleles = rep("", length(allele_calls))

    # How many alleles represent perfect matches?
    n_gl_unmut = sapply(which_no_muts, length)

    one_unmut = which(n_gl_unmut == 1)
    split_names = strsplit(allele_calls, ",")
    if (length(one_unmut) > 0){
        inds = unlist(which_no_muts[one_unmut])
        unmut_alleles[one_unmut] = mapply("[", split_names[one_unmut], inds)
    }

    more_unmut = which(n_gl_unmut > 1)
    if (length(more_unmut) > 0){
        inds = which_no_muts[more_unmut]
        unmut_multi = mapply(function(x,y) x[unlist(y)], split_names[more_unmut],
                             inds, SIMPLIFY = FALSE)
        unmut_alleles[more_unmut] = sapply(unmut_multi, paste, collapse=",")
    }

    unmut_alleles = unmut_alleles[unmut_i]

    return(unmut_alleles)

}

#' Find mutation counts for frequency sequences
#'
#' \code{getPopularMutationCount} determines which sequences occur frequently
#' for each V gene and returns the mutation count of those sequences.
#'
#' @param  data          \code{data.frame} in the Change-O format. See
#'                       \link{findNovelAlleles} for a list of required
#'                       columns.
#' @param  germline_db   named list of IMGT-gapped germline sequences.
#' @param  v_call        name of the column in \code{data} with V allele calls.
#'                       Default is \code{v_call}.
#' @param  seq           name of the column in \code{data} with the
#'                       aligned, IMGT-numbered, V(D)J nucleotide sequence.
#'                       Default is \code{sequence_alignment}.
#' @param  gene_min      portion of all unique sequences a gene must
#'                       constitute to avoid exclusion.
#' @param  seq_min       number of copies of the V that must be present for
#'                       to avoid exclusion.
#' @param  seq_p_of_max  ror each gene, the fraction of the most common V sequence
#'                       count that a sequence must meet to avoid exclusion.
#' @param  full_return   if \code{TRUE}, will return all \code{data} columns and
#'                       will include sequences with mutation count < 1.
#'
#' @return  A data frame of genes that have a frequent sequence mutation count
#'          above 1.
#'
#' @seealso \link{getMutatedPositions} can be used to find which positions
#'          of a set of sequences are mutated.
#'
#' @examples
#' getPopularMutationCount(AIRRDb, SampleGermlineIGHV)
#'
#' @export
getPopularMutationCount <- function(data, germline_db,
                                    v_call="v_call",
                                    seq="sequence_alignment",
                                    gene_min = 1e-03,
                                    seq_min = 50, seq_p_of_max = 1/8,
                                    full_return = FALSE){
    modified_db <- data %>%
        mutate(v_gene = getGene(!!rlang::sym(v_call))) %>%
        group_by(!!rlang::sym("v_gene")) %>%
        mutate(v_gene_n = n()) %>%
        group_by(1:n()) %>%
        mutate(v_sequence_imgt = substring(!!rlang::sym(seq), 1, 312)) %>%
        # Count occurrence of each unique IMGT-gapped V sequence
        group_by(!!!rlang::syms(c("v_gene", "v_sequence_imgt"))) %>%
        mutate(v_sequence_imgt_n = n()) %>%
        # Determine count of most common sequence
        group_by(!!rlang::sym("v_gene")) %>%
        mutate(v_sequence_imgt_n_max = max(!!rlang::sym("v_sequence_imgt_n"))) %>%
        # Remove rare V genes, rare sequences, and sequences not making up a
        # sufficient proportion of sequences as compared to the most common
        ungroup %>%
        distinct(!!rlang::sym("v_sequence_imgt"), .keep_all = TRUE) %>%
        filter(!!rlang::sym("v_gene_n") >= (nrow(data)*gene_min)) %>%
        filter(!!rlang::sym("v_sequence_imgt_n") >= seq_min) %>%
        mutate(v_sequence_imgt_p_max = !!rlang::sym("v_sequence_imgt_n")/!!rlang::sym("v_sequence_imgt_n_max")) %>%
        filter(!!rlang::sym("v_sequence_imgt_p_max") >= seq_p_of_max)
    # Determine the mutation counts of the V sequences and append them to the db
    mutation_count <- getMutCount(modified_db$v_sequence_imgt,
                                 modified_db[[v_call]],
                                 germline_db) %>%
        sapply(function(x) min(unlist(x)))
    if (length(mutation_count)==0){
        mutation_count <- integer(0)
    }
    merged_db <- bind_cols(modified_db, data.frame(mutation_count))
    # Strip down the data frame before returning it
    if (!full_return) {
        merged_db <- merged_db %>%
            filter(mutation_count > 0) %>%
            select(!!!rlang::syms(c("v_gene", "mutation_count")))
    }
    return(merged_db)
}

#' Insert polymorphisms into a nucleotide sequence
#'
#' \code{insertPolymorphisms} replaces nucleotides in the desired locations of a
#' provided sequence.
#'
#' @param    sequence     starting nucleotide sequence.
#' @param    positions    numeric vector of positions which to be changed.
#' @param    nucleotides  character vector of nucleotides to which to change the
#'                        positions.
#'
#' @return   A sequence with the desired nucleotides in the provided locations.
#'
#' @examples
#' insertPolymorphisms("HUGGED", c(1, 6, 2), c("T", "R", "I"))
#'
#' @export
insertPolymorphisms <- function(sequence, positions, nucleotides) {

    if(length(positions) != length(nucleotides)){
        stop("Number of nucleotides and number of positions do not match.")
    }
    names(positions) = nucleotides
    for (i in 1:length(positions)){
        substr(sequence, positions[i], positions[i]) = names(positions[i])
    }

    return(sequence)
}

# Formatting and Cleanup --------------------------------------------------

#' Read immunoglobulin sequences
#'
#' \code{readIgFasta} reads a fasta-formatted file of immunoglobulin (Ig)
#' sequences and returns a named vector of those sequences.
#'
#' @param    fasta_file       fasta-formatted file of immunoglobulin sequences.
#' @param    strip_down_name  if \code{TRUE}, will extract only the allele name
#'                            from the strings fasta file's sequence names.
#' @param    force_caps       if \code{TRUE}, will force nucleotides to
#'                            uppercase.
#'
#' @return   Named vector of strings representing Ig alleles.
#'
#' @seealso  \link{writeFasta} to do the inverse.
#'
#' @examples
#' \dontrun{
#' # germlines <- readIgFasta("ighv.fasta")
#' }
#'
#' @export
readIgFasta <- function(fasta_file, strip_down_name=TRUE, force_caps=TRUE) {
    all_char = readChar(fasta_file, file.info(fasta_file)$size)
    split_by_sequence = strsplit(all_char, "[ \t\r\n\v\f]?>")
    add_name_break = sapply(split_by_sequence, function(x) sub("[\r\n]",">",x))
    cleaned_up = sapply(add_name_break, function(x) gsub("[ \t\r\n\v\f]", "", x))
    broken_names = sapply(cleaned_up, strsplit, ">")

    seqs = sapply(broken_names, "[", 2)
    seq_names = sapply(broken_names, "[", 1)
    if(force_caps) { seqs = toupper(seqs) }
    if(strip_down_name){ seq_names = getAllele(seq_names, strip_d=FALSE) }
    names(seqs) = seq_names

    return(seqs[which(!is.na(seqs))])
}

#' Write to a fasta file
#'
#' \code{writeFasta} writes a named vector of sequences to a file in fasta
#' format.
#'
#' @param    named_sequences  vector of named string representing sequences
#' @param    file             the name of the output file.
#' @param    width            the number of characters to be printed per line.
#'                            if not between 1 and 255, width with be infinite.
#' @param    append           \code{logical} indicating if the output should be
#'                            appended to \code{file} instead of overwriting it
#'
#' @return   A named vector of strings representing Ig alleles.
#'
#' @seealso  \link{readIgFasta} to do the inverse.
#'
#' @examples
#' \dontrun{
#' # writeFasta(germlines, "ighv.fasta")
#' }
#'
#' @export
writeFasta <- function(named_sequences, file, width=60, append=FALSE){
    . = NULL
    seq_names = names(named_sequences) %>%
        paste(">", ., "\n", sep="")
    seqs = as.character(named_sequences)
    if(is.numeric(width) & width > 0 & width < 256){
        width_regex = paste("(.{", width, ",", width, "})", sep="")
        seqs = gsub(width_regex, "\\1\n", seqs)
    }
    seqs = seqs %>%
        paste("\n", sep="") %>%
        gsub("\n\n", "\n", .)
    paste(seq_names, seqs, sep="", collapse="") %>%
        cat(file=file, append=append)
}

#' Update IGHV allele names
#'
#' \code{updateAlleleNames} takes a set of IGHV allele calls and replaces any
#' outdated names (e.g. IGHV1-f) with the new IMGT names.
#'
#' @param    allele_calls  vector of strings representing IGHV allele names.
#'
#' @return   Vector of strings representing updated IGHV allele names.
#'
#' @note
#' IMGT has removed \code{IGHV2-5*10} and \code{IGHV2-5*07} as it has determined they
#' are actually alleles \code{02} and \code{04}, respectively. The updated allele
#' names are based on IMGT release 2014-08-4.
#'
#' @references
#' \enumerate{
#'   \item Xochelli et al. (2014) Immunoglobulin heavy variable (IGHV) genes
#'         and alleles: new entities, new names and implications for research and
#'         prognostication in chronic lymphocytic leukaemia. Immunogenetics. 67(1):61-6
#' }
#'
#' @seealso Like \code{updateAlleleNames}, \link{sortAlleles} can help
#'          format a list of allele names.
#'
#' @examples
#' # Create a vector that uses old gene/allele names.
#' alleles <- c("IGHV1-c*01", "IGHV1-f*02", "IGHV2-5*07")
#'
#' # Update the alleles to the new names
#' updateAlleleNames(alleles)
#'
#' @export
updateAlleleNames <- function(allele_calls) {
    . = NULL
    temporary_names = c("IGHV1-c*",
                        "IGHV1-f*",
                        "IGHV3-d*",
                        "IGHV3-h*",
                        "IGHV4-b*",
                        "IGHV5-a*",
                        "IGHV2-5*10",
                        "IGHV2-5*07")
    definitive_names = c("IGHV1-38-4*",
                         "IGHV1-69-2*",
                         "IGHV3-38-3*",
                         "IGHV3-69-1*",
                         "IGHV4-38-2*",
                         "IGHV5-10-1*",
                         "IGHV2-5*02",
                         "IGHV2-5*04")
    for (i in 1:length(temporary_names)){
        allele_calls = allele_calls %>%
            gsub(temporary_names[i], definitive_names[i], ., fixed = TRUE)
    }
    return(allele_calls)
}

#' Sort allele names
#'
#' \code{sortAlleles} returns a sorted vector of strings representing Ig allele
#' names. Names are first sorted by gene family, then by gene, then by allele.
#' Duplicated genes have their alleles are sorted as if they were part of their
#' non-duplicated counterparts (e.g. \code{IGHV1-69D*01} comes after \code{IGHV1-69*01}
#' but before \code{IGHV1-69*02}), and non-localized genes (e.g. \code{IGHV1-NL1*01})
#' come last within their gene family.
#'
#' @param    allele_calls  vector of strings representing Ig allele names.
#' @param    method        string defining the method to use when sorting alleles.
#'                         If \code{"name"} then sort in lexicographic order. If
#'                         \code{"position"} then sort by position in the locus, as
#'                         determined by the final two numbers in the gene name.
#' @return   A sorted vector of strings representing Ig allele names.
#'
#' @seealso Like \code{sortAlleles}, \link{updateAlleleNames} can help
#'          format a list of allele names.
#'
#' @examples
#' # Create a list of allele names
#' alleles <- c("IGHV1-69D*01","IGHV1-69*01","IGHV1-2*01","IGHV1-69-2*01",
#'              "IGHV2-5*01","IGHV1-NL1*01", "IGHV1-2*01,IGHV1-2*05",
#'              "IGHV1-2", "IGHV1-2*02", "IGHV1-69*02")
#'
#' # Sort the alleles by name
#' sortAlleles(alleles)
#'
#' # Sort the alleles by position in the locus
#' sortAlleles(alleles, method="pos")
#'
#' @export
sortAlleles <- function(allele_calls, method=c("name", "position")) {
    # Check arguments
    method <- match.arg(method)

    # Standardize format of submitted alleles, first
    SUBMITTED_CALLS <- getAllele(allele_calls, first = FALSE, strip_d= FALSE)
    allele_df <- data.frame(
        list("SUBMITTED_CALLS"=SUBMITTED_CALLS,
             "SUBMITTED_NAMES"=allele_calls),
        stringsAsFactors = FALSE) %>%
        arrange(SUBMITTED_CALLS) %>%
        # Determine the family
        mutate(FAMILY = getFamily(!! rlang::sym("SUBMITTED_CALLS"))) %>%
        # Determine the gene (exclude family); convert letters to numbers for sort
        mutate(GENE = getGene(!! rlang::sym("SUBMITTED_CALLS"))) %>%
        mutate(GENE1 = gsub("[^-]+([-S][^-\\*D]+).*","\\1",!! rlang::sym("SUBMITTED_CALLS"))) %>%
        mutate(GENE1 = sub("^-","",!!rlang::sym("GENE1"))) %>%
        mutate(GENE1 = as.numeric(gsub("[^0-9]+", "99", !!rlang::sym("GENE1")))) %>%
        # If there is a second gene number, determine that, too
        mutate(GENE2 = gsub("[^-]+[-S][^-]+-?","",!! rlang::sym("GENE"))) %>%
        mutate(GENE2 = as.numeric(gsub("[^0-9]+", "99", !!rlang::sym("GENE2")))) %>%
        mutate(ALLELE = getAllele(!!rlang::sym("SUBMITTED_CALLS"))) %>%
        mutate(ALLELE = sub("[^\\*]+\\*|[^\\*]+$","", !!rlang::sym("ALLELE"))) %>%
        mutate(ALLELE = as.numeric(sub("_.+$", "", !!rlang::sym("ALLELE"))))

    # Convert missing values to 0, sort data frame
    allele_df[is.na(allele_df)] <- 0
    if (method == "name") {
        sorted_df <- arrange(allele_df, !!!rlang::syms(c("FAMILY", "GENE1", "GENE2", "ALLELE")))
    } else if (method == "position") {
        sorted_df <- arrange(allele_df, desc(!!rlang::sym("GENE1")),
                             desc(!!rlang::sym("GENE2")),
                             desc(!!rlang::sym("FAMILY")),
                             desc(!!rlang::sym("ALLELE")))
    }

    return(sorted_df$SUBMITTED_NAMES)
}

#' Clean up nucleotide sequences
#'
#' \code{cleanSeqs} capitalizes nucleotides and replaces all characters
#' besides \code{c("A", "C", "G", "T", "-", ".")} with \code{"N"}.
#'
#' @param    seqs  vector of nucleotide sequences.
#'
#' @return   A modified vector of nucleotide sequences.
#'
#' @seealso \link{sortAlleles} and \link{updateAlleleNames} can
#'          help format a list of allele names.
#'
#' @examples
#' # Clean messy nucleotide sequences
#' seqs <- c("AGAT.taa-GAG...ATA", "GATACAGTXXZZAGNNPPACA")
#' cleanSeqs(seqs)
#'
#' @export
cleanSeqs <- function(seqs) {
    # . = NULL
    # seqs %>%
    #     toupper %>%
    #     gsub(".", "-", . , fixed = TRUE) %>%
    #     gsub("[^ACGT-]", "N", .) %>%
    #     return
    clean_seqs <- stri_replace_all_regex(stri_trans_toupper(seqs), "[^ACGT\\.\\-]", "N")
    if (!is.null(names(seqs))) {
        names(clean_seqs) <- names(seqs)
    }
    clean_seqs
}


# Private Functions -------------------------------------------------------

# Find mutations-by-position compared to a germline
#
# \code{positionMutations} duplicates the rows of a data frame for each
# position to be analyzed and determines if each sample is mutated at that
# position
#
# @param  data          \code{data.frame} containing repertoire data. See
#                       \link{findNovelAlleles} for a list of required
#                       columns.
# @param  germline      the germline to which all the sequences should be
#                       compared
# @param  pos_range     the range of positions within the sequence for which
#                       the rows should be duplicated and checked for mutation
# @param  seq           name of the column in \code{data} with the
#                       aligned, IMGT-numbered, V(D)J nucleotide sequence.
#                       Default is \code{sequence_alignment}
# @return  A data frame with rows duplicated for all the positions to be
# analyzed and a column indicating whether the position is mutated in
# comparison to the germline
#
positionMutations <- function(data, germline, pos_range, seq="sequence_alignment", pos_range_max = NULL){
    . = NULL
    pos_db = pos_range %>%
        length() %>%
        rep("data", .) %>%
        paste(collapse=",") %>%
        paste("bind_rows(",., ")") %>%
        parse(text=.) %>%
        eval()
    pos_db$POSITION = c(sapply(pos_range, rep, nrow(data)))
    # Find which positions are mutated
    pos_db <- pos_db %>%
        mutate(NT = substring(!!rlang::sym(seq),
                              !!rlang::sym("POSITION"),
                              !!rlang::sym("POSITION"))) %>%
        mutate(GERM_NT = substring(germline, !!rlang::sym("POSITION"), !!rlang::sym("POSITION"))) %>%
        mutate(MUTATED = (!!rlang::sym("NT") != !!rlang::sym("GERM_NT") &
                          !!rlang::sym("NT") != "N" &
                          !!rlang::sym("NT") != "-" &
                          !!rlang::sym("NT") != "." &
                          !!rlang::sym("NT") != "")) %>%
        mutate(OBSERVED = (!!rlang::sym("NT") != "-" &
                           !!rlang::sym("NT") != "." &
                           !!rlang::sym("NT") != ""))
    if(!is.null(pos_range_max)) pos_db <- pos_db %>% filter(!!rlang::sym("POSITION")<= !!rlang::sym(pos_range_max))
    if (any(pos_db$GERM_NT == "") && !grepl("\\.", germline) ) {
        stop("Empty ('') GERM_NT positions found. Check you are using gapped reference germlines.")
        }
    return(pos_db)
}

# Find sequences carrying certain levels of mutation
#
# \code{mutationRangeSubset} determines the mutations in a \code{data.frame} of
# sequences and returns the subset of sequences that meet the given mutation
# count limits
#
# @param  data          \code{data.frame} containing repertoire data. See
#                       \link{findNovelAlleles} for a list of required
#                       columns.
# @param  germline      germline to which all the sequences should be
#                       compared.
# @param  pos_range     range of positions within the sequences that should
#                       be analyzed for mutations
# @param  pos_range     range of mutation counts that sequences can have
#                       and still be included.
# @param  seq           name of the column in \code{data} with the
#                       aligned, IMGT-numbered, V(D)J nucleotide sequence.
#                       Default is \code{sequence_alignment}.
# @return
# A data.frame containing only the subset carrying the desired levels
# of mutation
#
mutationRangeSubset <- function(data, germline, mut_range, pos_range,
                                seq="sequence_alignment", pos_range_max = NULL){
    . = NULL
    pads = paste(rep("-", min(pos_range)-1), collapse="")
    if(is.null(pos_range_max)){
        data$MUT_COUNT = data[[seq]] %>%
            substring(min(pos_range), max(pos_range)) %>%
            paste(pads, ., sep="") %>%
            getMutatedPositions(germline) %>%
            sapply(length)
    }else{
        data$MUT_COUNT =  data %>%
            mutate(subseq = substring(!!rlang::sym(seq), min(pos_range), min(max(pos_range), !!rlang::sym(pos_range_max)))) %>%
            pull(!!rlang::sym("subseq")) %>%
            paste(pads, ., sep="") %>%
            getMutatedPositions(germline) %>%
            sapply(length)
    }

    data = data %>%
        filter(!!rlang::sym("MUT_COUNT") %in% mut_range)
    return(data)
}

# Find lower range of y-intercept confidence interval
#
# \code{findLowerY} finds the lower range of y-intercept confidence interval
#
# @details  If mut_min is 1, a y-intercept will be searched for at 0. If
# mut_min is above 1, then the "y-intercept" will be found at x = mut_min - 1.
#
# @param    x         vector of x values.
# @param    y         vector of y values.
# @param    mut_min   value where the the lowest mutation count should be
#                     found. See details.
# @param    alpha     alpha cutoff the be used in constructing the
#                     confidence interval
#
# @return  A data frame containing only the subset carrying the desired levels
# of mutation
#
findLowerY = function(x, y, mut_min, alpha){
    y = y + 1 - mut_min
    lowerY = suppressWarnings(confint(lm(x ~ y), level=1 - 2*alpha)[[1]])
    return(lowerY)
}

# enhanced substring extraction
#
# \code{superSubstring} is an enhanced version of \code{substring} in that
# it can find disjoint positions in one call.
#
# @param    string      single string.
# @param    positions   the positions to be extracted.
#
# @return   Substring.
superSubstring = function(string, positions){
    if(length(string) != 1){ stop("Please submit only one string.") }
    chars = sapply(positions, function(x) substring(string, x, x))
    return(paste(chars, collapse=""))
}


# Layout multiple ggplots
#
# \code{multiplot} is a function provided by http://www.cookbook-r.com/ which
# allows for plotting multiple ggplot objects in one panel.
#
# @param    ...       ggplot2 object(s).
# @param    plotlist  a list alternative to (...).
# @param    file      an unused parameter, but present in the provided function.
# @param    cols      Number of columns in layout.
# @param    layout    matrix specifying the layout. If present, \code{cols} is
#                     ignored.
# @param    heights   numeric vector A numeric vector or unit object
#                     describing the heights of the rows in the layout. Will
#                     be passed to grid.layout. Default is all plots have
#                     the same height.
multiplot <- function(..., plotlist=NULL, cols=1, layout=NULL, heights=NULL) {
    # Make a list from the ... arguments and plotlist
    plots <- c(list(...), plotlist)
    numPlots <- length(plots)
    ncol <- cols
    nrow <- ceiling(numPlots/cols)
    if (is.null(heights)) { heights = rep(1,nrow) }
    if (is.null(layout)) {
        # Make the panel
        # ncol: Number of columns of plots
        # nrow: Number of rows needed, calculated from # of cols
        layout <- matrix(seq(1, cols * nrow),
                         ncol = cols, nrow = nrow)
    }
    grob <- gridExtra::arrangeGrob(grobs=plots,
                                   nrow=nrow, ncol=ncol, layout_matrix = layout,
                                   heights=heights)
    p <- ggplot() +
        layer(data = data.frame(x = NA),
              stat = StatIdentity,
              position = PositionIdentity,
              # geom = GeomDrawGrob,
              geom = GeomCustomAnn,
              inherit.aes = FALSE,
              params = list(grob = grob,
                            xmin = 0,
                            xmax = 1,
                            ymin = 0,
                            ymax = 1)) +
        scale_x_continuous(expand=c(0,0)) +
        scale_y_continuous(expand=c(0,0))
    p
}


#' Subsample repertoire
#'
#' \code{subsampleDb} will sample the same number of sequences for each gene, family
#' or allele (specified with \code{mode}) in \code{data}. Samples or subjects can
#' be subsampled independent by setting \code{group}.
#'
#' \code{data} will be split into gene, allele or family subsets (\code{mode}) from
#' which the same number of sequences will be subsampled. If \code{mode=gene},
#' for each gene in the field \code{gene} from \code{data}, a maximum of
#' \code{max_n} sequences will be subsampled. Input sequences
#' that have multiple gene calls (ties), can be subsampled from any of their calls,
#' but these duplicated samplings will be removed, and the final
#' subsampled \code{data} will contain unique rows.
#'
#' @param   data   \code{data.frame} containing repertoire data.
#' @param   gene   name of the column in \code{data} with allele calls. Default
#'                 is \code{v_call}.
#' @param   mode   one of \code{c("gene", "family", "allele")} defining the degree of
#'                 specificity regarding allele calls when subsetting sequences.
#'                 Determines how \code{data} will be split into subsets from
#'                 which the same number of sequences will be subsampled. See
#'                 also \code{group}.
#' @param   min_n  minimum number of observations to sample from each group. A group with
#'                 less observations than the minimum is excluded.
#' @param   max_n  maximum number of observations to sample for all \code{mode} groups.
#'                 If \code{NULL}, it will be set automatically to the size of
#'                 the smallest group. If \code{max_n} is larger than the available
#'                 number of sequences for any \code{mode} group, it will be
#'                 automatically adjusted and the effective \code{max_n} used
#'                 will be the size of the smallest \code{mode} group.
#' @param   group  columns containing additional grouping variables, e.g. sample_id.
#'                 These groups will be subsampled independently. If
#'                 \code{max_n} is \code{NULL}, a \code{max_n} will be
#'                 automatically set for each \code{group}.
#'
#' @return  Subsampled version of the input \code{data}.
#'
#' @seealso \link{selectNovel}
#'
#' @examples
#' set.seed(1)
#' subsampleDb(AIRRDb)
#'
#' @export
subsampleDb <- function(data, gene="v_call", mode=c("gene", "allele", "family"),
                        min_n=1, max_n=NULL, group=NULL) {

    mode <- match.arg(mode)

    # Check columns exist(can be NULL)
    check <- checkColumns(data, c(gene, group))
    if (check != TRUE) { stop(check) }

    # If group is set,
    # split by group, then call this same function without grouping,
    # to apply the subsample within each group independently
    if (!is.null(group)) {
        group_id <- data %>%
            group_by(!!rlang::sym(group)) %>%
            group_indices()
        ss_data <- bind_rows(lapply(split(data, group_id), subsampleDb,
                             gene=gene, mode=mode, min_n=min_n, max_n=max_n, group=NULL)
        )
        return(ss_data)
    }

    # Extract gene, allele or family assignments
    gene_func <- switch(mode,
                        allele=getAllele,
                        gene=getGene,
                        family=getFamily)

    # Get temporary 'gene' names (SubSet Gene) used to
    # create subsampling groups
    data[["SS_GENE"]] <- gene_func(data[[gene]], first=FALSE)


    # Find which rows' calls contain which germline alleles
    genes <- unique(unlist(strsplit(data[["SS_GENE"]], ",")))
    allele_groups <- sapply(genes, grep, data[["SS_GENE"]],
                            fixed=TRUE, simplify=FALSE)
    allele_groups_sizes <- sapply(allele_groups, length)
    allele_groups <- allele_groups[ allele_groups_sizes >= min_n ]
    if(length(allele_groups) == 0){
        stop("Not enough sample sequences (min_n) were assigned to any gene. ",
                "Returned `NULL`")
    }

    # Determine the number of sequences to subsample
    if (!is.null(max_n)) {
        n <- min(allele_groups_sizes, max_n)
    } else {
        n <- min(allele_groups_sizes)
    }

    # Get indices subsampled sequences
    ss_idx <- lapply(allele_groups, sample, size=n, replace=FALSE, prob=NULL)
    ss_idx <- unique(unlist(ss_idx))

    data[ss_idx, ] %>% select(-!!rlang::sym("SS_GENE"))
}

# Find non triplet gaps in a nucleotide sequence
hasNonImgtGaps <- function (seq) {
    len <- ceiling(nchar(seq)/3)*3
    codons <- substring(seq, seq(1, len-2, 3), seq(3, len, 3))
    gaps_lengths <- nchar(gsub("[^\\.\\-]", "", codons))
    if (any(gaps_lengths %% 3 != 0)) {
        TRUE
    } else {
        FALSE
    }
}

# Compare two IMGT gapped sequences and find AA mutations
getMutatedAA <- function(ref_imgt, novel_imgt) {

    n_ref <- grepl("N", ref_imgt)
    if (any(n_ref)) {
        warning(sum(n_ref)," Ns found in ref_imgt")
    }

    n_novel <- grepl("N", novel_imgt)
    if (any(n_novel)) {
        warning(sum(n_novel)," Ns found in novel_imgt")
    }

    if (hasNonImgtGaps(ref_imgt)) {
        warning("Non IMGT gaps found in ref_imgt")
    }

    if (hasNonImgtGaps(novel_imgt)) {
        warning("Non IMGT gaps found in novel_imgt")
    }

    ref_imgt <- strsplit(alakazam::translateDNA(ref_imgt),"")[[1]]
    novel_imgt <- strsplit(alakazam::translateDNA(novel_imgt),"")[[1]]
    mutations <- c()
    diff_idx <- which(ref_imgt != novel_imgt)
    if (length(diff_idx)>0) {
        mutations <- paste0(diff_idx, ref_imgt[diff_idx],">",
                            replace(novel_imgt[diff_idx], is.na(novel_imgt[diff_idx]),"-"))
    }
    mutations
}


#' Generate evidence
#'
#' \code{generateEvidence} builds a table of evidence metrics for the final novel V
#' allele detection and genotyping inferences.
#'
#' @param data          a \code{data.frame} containing sequence data that has been
#'                      passed through \link{reassignAlleles} to correct the allele
#'                      assignments.
#' @param novel         the \code{data.frame} returned by \link{findNovelAlleles}.
#' @param genotype      the \code{data.frame} of alleles generated with \link{inferGenotype}
#'                      denoting the genotype of the subject.
#' @param genotype_db   a vector of named nucleotide germline sequences in the genotype.
#'                      Returned by \link{genotypeFasta}.
#' @param germline_db   the original uncorrected germline database used to by
#'                      \link{findNovelAlleles} to identify novel alleles.
#' @param j_call        name of the column in \code{data} with J allele calls.
#'                      Default is \code{j_call}.
#' @param junction      Junction region nucleotide sequence, which includes
#'                      the CDR3 and the two flanking conserved codons. Default
#'                      is \code{junction}.
#' @param fields        character vector of column names used to split the data to
#'                      identify novel alleles, if any. If \code{NULL} then the data is
#'                      not divided by grouping variables.
#'
#' @return
#' Returns the \code{genotype} input \code{data.frame} with the following additional columns
#' providing supporting evidence for each inferred allele:
#'
#' \itemize{
#'   \item \code{field_id}: Data subset identifier, defined with the input parameter \code{fields}.
#'   \item A variable number of columns, specified with the input parameter \code{fields}.
#'   \item \code{polymorphism_call}: The novel allele call.
#'   \item \code{novel_imgt}: The novel allele sequence.
#'   \item \code{closest_reference}: The closest reference gene and allele in
#'         the \code{germline_db} database.
#'   \item \code{closest_reference_imgt}: Sequence of the closest reference gene and
#'         allele in the \code{germline_db} database.
#'   \item \code{germline_call}: The input (uncorrected) V call.
#'   \item \code{germline_imgt}: Germline sequence for \code{germline_call}.
#'   \item \code{nt_diff}: Number of nucleotides that differ between the new allele and
#'         the closest reference (\code{closest_reference}) in the \code{germline_db} database.
#'   \item \code{nt_substitutions}: A comma separated list of specific nucleotide
#'         differences (e.g. \code{112G>A}) in the novel allele.
#'   \item \code{aa_diff}: Number of amino acids that differ between the new allele and the closest
#'         reference (\code{closest_reference}) in the \code{germline_db} database.
#'   \item \code{aa_substitutions}: A comma separated list with specific amino acid
#'         differences (e.g. \code{96A>N}) in the novel allele.
#'   \item \code{sequences}: Number of sequences unambiguously assigned to this allele.
#'   \item \code{unmutated_sequences}: Number of records with the unmutated novel allele sequence.
#'   \item \code{unmutated_frequency}: Proportion of records with the unmutated novel allele
#'         sequence (\code{unmutated_sequences / sequences}).
#'   \item \code{allelic_percentage}: Percentage at which the (unmutated) allele is observed
#'         in the sequence dataset compared  to other (unmutated) alleles.
#'   \item \code{unique_js}: Number of unique J sequences found associated with the
#'         novel allele. The sequences are those who have been unambiguously assigned
#'         to the novel allele (\code{polymorphism_call}).
#'   \item \code{unique_cdr3s}: Number of unique CDR3s associated with the inferred allele.
#'         The sequences are those who have been unambiguously assigned to the
#'         novel allele (polymorphism_call).
#'   \item \code{mut_min}: Minimum mutation considered by the algorithm.
#'   \item \code{mut_max}: Maximum mutation considered by the algorithm.
#'   \item \code{pos_min}: First position of the sequence considered by the algorithm (IMGT numbering).
#'   \item \code{pos_max}: Last position of the sequence considered by the algorithm (IMGT numbering).
#'   \item \code{y_intercept}: The y-intercept above which positions were considered
#'         potentially polymorphic.
#'   \item \code{alpha}: Significance threshold to be used when constructing the
#'         confidence interval for the y-intercept.
#'   \item \code{min_seqs}: Input \code{min_seqs}. The minimum number of total sequences
#'         (within the desired mutational range and nucleotide range) required
#'         for the samples to be considered.
#'   \item \code{j_max}: Input \code{j_max}. The maximum fraction of sequences perfectly
#'         aligning to a potential novel allele that are allowed to utilize to a particular
#'         combination of junction length and J gene.
#'   \item \code{min_frac}: Input \code{min_frac}. The minimum fraction of sequences that must
#'         have usable nucleotides in a given position for that position to be considered.
#'   \item \code{note}: Comments regarding the novel allele inference.
#' }
#'
#' @seealso
#' See \link{findNovelAlleles}, \link{inferGenotype} and \link{genotypeFasta}
#' for generating the required input.
#'
#' @examples
#' \donttest{
#' # Generate input data
#' novel <- findNovelAlleles(AIRRDb, SampleGermlineIGHV,
#'     v_call="v_call", j_call="j_call", junction="junction",
#'     junction_length="junction_length", seq="sequence_alignment")
#' genotype <- inferGenotype(AIRRDb, find_unmutated=TRUE,
#'                           germline_db=SampleGermlineIGHV,
#'                           novel=novel,
#'                           v_call="v_call", seq="sequence_alignment")
#' genotype_db <- genotypeFasta(genotype, SampleGermlineIGHV, novel)
#' data_db <- reassignAlleles(AIRRDb, genotype_db,
#' v_call="v_call", seq="sequence_alignment")
#'
#' # Assemble evidence table
#' evidence <- generateEvidence(data_db, novel, genotype,
#'                              genotype_db, SampleGermlineIGHV,
#'                              j_call = "j_call",
#'                              junction = "junction")
#' }
#'
#' @export
generateEvidence <- function(data, novel, genotype, genotype_db,
                             germline_db, j_call="j_call", junction="junction",
                             fields=NULL) {
    # Visibility hack
    . <- NULL

    # Define set of sequences containing genotype and uncorrected calls
    germline_set <- c(germline_db[!names(germline_db) %in% names(genotype_db)],
                      genotype_db)

    # Find closest reference
    .findClosestReference <- function(seq, allele_calls, ref_germ,
                                      exclude_self=F, multiple=F) {

        # filter `ref_germ`, use only same gene segment as `seq`
        diff_calls <- sub("[0-9]+$","",getFamily(names(ref_germ))) != sub("[0-9]+$","",getFamily(names(seq)))
        if (sum(diff_calls) > 0) {
            allele_calls <- allele_calls[allele_calls %in% names(ref_germ[diff_calls]) == FALSE]
            ref_germ <- ref_germ[!diff_calls]
        }

        closest <- getMutCount(seq,
                               paste(allele_calls, collapse=","),
                               ref_germ)
        min_dist <- min(unlist(closest))
        closest_idx <- which(unlist(closest) == min_dist)
        closest_names <- unique(allele_calls[closest_idx])
        if (exclude_self & names(seq) %in% closest_names) {
            warning("Excluding self")
            closest_names <- closest_names[closest_names!=names(seq)] # not self
        }
        if (length(closest_names) > 1) {
            warning(paste0("More than one closest reference found for ",
                           names(seq),": ",
                           paste(closest_names, collapse=",")))
            # Keep the one with less mutated positions
            mut_pos_count <- sapply(gsub("[^_]","",closest_names), nchar)
            closest_names <- closest_names[mut_pos_count==min(mut_pos_count)]
            # Pick same length
            if (length(closest_names) > 1 ) {
                idx <- which(
                    sapply(ref_germ[closest_names],nchar) == nchar(ref_germ[names(seq)])
                )
                if (length(idx) > 0 ) {
                    closest_names <- closest_names[idx]
                }
            }
            # Pick same allele
            if (length(closest_names) > 1 ) {
                idx <- which(
                    getAllele(closest_names) == gsub("_.+", "", getAllele(names(seq)))
                )
                if (length(idx) > 0 ) {
                    closest_names <- closest_names[idx]
                }
            }
            # Pick not duplicated
            if (length(closest_names) > 1 ) {
                idx <- !grepl("D\\*", closest_names)
                if (any(idx)) {
                    closest_names <- closest_names[idx]
                }
            }
            # If still more than one, err and TODO
            if (length(closest_names) > 1 & multiple==FALSE) {
                msg <- paste0("Multiple closest reference found for ",
                              names(seq),":\n",
                              paste(closest_names, collapse=","))
                stop(msg)
            }
            warning(paste0("Use: ",
                           paste(closest_names, collapse=","),
                           " (less mutated positions, not D, same length, same allele)"))

        }
        closest_names
    }

    # Subset to novel alleles
    unnest_cols <- c("alleles", "counts")
    final_gt <- genotype %>%
        dplyr::group_by(.data$gene) %>%
        dplyr::filter(!duplicated(.data$alleles)) %>%
        dplyr::ungroup() %>%
        dplyr::mutate(alleles=strsplit(as.character(.data$alleles), ","),
                      counts=strsplit(as.character(.data$counts), ",")) %>%
        tidyr::unnest(dplyr::all_of(unnest_cols)) %>%
        dplyr::mutate(polymorphism_call=paste0(.data$gene, "*" , .data$alleles)) %>%
        dplyr::filter(.data$polymorphism_call %in% novel$polymorphism_call)  %>%
        dplyr::rename(allele="alleles")


    # Add info from novel
    final_gt <- dplyr::inner_join(dplyr::rename(final_gt, note_gt="note"),
                                  novel,
                                  by=c(fields, "polymorphism_call"))

    # Add message if the same novel img sequence found from
    # different starting alleles, these will be novel imgt sequences
    # with more than one polymorphism call
    final_gt <- final_gt %>%
        dplyr::group_by(.data$novel_imgt) %>%
        dplyr::mutate(NUM_CALLS=length(unique(.data$polymorphism_call))) %>%
        dplyr::ungroup()
    idx_mult <- which(final_gt$NUM_CALLS > 1)
    final_gt$NUM_CALLS <- NULL
    if (length(idx_mult) > 0) {
        final_gt$note_gt[idx_mult] <- paste(
            final_gt$note_gt[idx_mult],
            " Found multiple polymorphism calls for the same novel_imgt.",
            sep="")
    }


    if (nrow(final_gt)>0) {

        .addEvidence <- function(df, germline_set, germline_db) {
            polymorphism <- df[['polymorphism_call']]
            novel_imgt <- df[["novel_imgt"]]
            names(novel_imgt) <- polymorphism

            v_call_genotyped <- data[["v_call_genotyped"]]

            sequences <- sum(v_call_genotyped == polymorphism)
            df[["sequences"]] <- sequences
            closest_ref_input <- .findClosestReference(novel_imgt,
                                                       names(germline_db),
                                                       germline_db,
                                                       exclude_self=F)
            closest_ref <- .findClosestReference(novel_imgt,
                                                 names(germline_set),
                                                 germline_set,
                                                 exclude_self=F, multiple=T)

            if (all(getGene(closest_ref_input) != getGene(closest_ref))) {
                warning("closest reference gene difference")
            }

            if (all(closest_ref != polymorphism)) {
                warning(paste0("closest reference allele (",
                               closest_ref
                               ,") different from polymorphism_call allele (",
                               polymorphism,")"))
            }

            df[["closest_reference"]] <- closest_ref_input

            nt_diff <- unlist(getMutatedPositions(novel_imgt, germline_set[[closest_ref_input]]))
            nt_diff_string <- ""
            if (nchar(novel_imgt) < nchar(germline_set)[[closest_ref_input]]) {
                nt_diff <- c(nt_diff, (nchar(novel_imgt)+1):nchar(germline_set[[closest_ref_input]]))
            }
            if (length(nt_diff) > 0 ) {
                ref_nt <- strsplit(germline_set[[closest_ref_input]],"")[[1]][nt_diff]
                novel_nt <- strsplit(germline_set[[polymorphism]],"")[[1]][nt_diff]
                nt_diff_string <- paste(paste(
                    nt_diff,
                    ref_nt,
                    ">",
                    replace(novel_nt, is.na(novel_nt), "-"),
                    sep=""), collapse=",")
            }

            df[["nt_diff"]] <- length(nt_diff)
            df[["nt_substitutions"]] <- nt_diff_string

            diff_aa <- getMutatedAA(germline_set[[closest_ref_input]], germline_set[[polymorphism]])

            if (length(diff_aa)>0) {
                df[["aa_diff"]] <- length(diff_aa)
                df[["aa_substitutions"]] <- paste(diff_aa,collapse=",")
            } else {
                df[["aa_diff"]] <- 0
                df[["aa_substitutions"]] <- ""
            }

            df[["counts"]] <- as.numeric(df[["counts"]])
            df[["total"]] <- as.numeric(df[["total"]])
            df[["unmutated_sequences"]] <- as.numeric(df[["counts"]])
            df[["unmutated_frequency"]] <- as.numeric(df[["counts"]])/sequences

            df[["allelic_percentage"]] <- 100*df[["unmutated_sequences"]]/as.numeric(df[["total"]])

            if (sequences > 0) {
                df[["unique_js"]] <- data %>%
                    dplyr::filter(.data$v_call_genotyped == polymorphism)  %>%
                    dplyr::distinct(.data[[j_call]]) %>%
                    nrow()
                df[["unique_cdr3s"]] <- data %>%
                    dplyr::filter(.data$v_call_genotyped == polymorphism)  %>%
                    dplyr::distinct(translateDNA(.data[[junction]], trim=TRUE)) %>%
                    nrow()
            } else {
                df[["unique_js"]]  <- NA
                df[["unique_cdr3s"]] <- NA
            }

            # Add closest germline
            df[["closest_reference_imgt"]] <- cleanSeqs(germline_set[[closest_ref_input]])

            data.frame(df, stringsAsFactors=FALSE)
        }

        final_gt <- final_gt %>%
            dplyr::rowwise() %>%
            do(.addEvidence(., germline_set=germline_set, germline_db=germline_db)) %>%
            dplyr::mutate(note=trimws(paste(.data$note_gt, .data$note, sep=" "))) %>%
            dplyr::select(-c("note_gt"))
    }

    return(final_gt)
}

#' @keywords internal
#' @aliases tigger-package
"_PACKAGE"

# Project documentation for tigger

#' tigger
#'
#' Here we provide a \strong{T}ool for \strong{I}mmuno\strong{g}lobulin
#' \strong{G}enotype \strong{E}lucidation via \strong{R}ep-Seq (TIgGER).
#' TIgGER infers the set of Ig alleles carried by an
#' individual (including any novel alleles) and then uses this set of alleles to
#' correct the initial assignments given to sample sequences by existing tools.
#'
#' @details
#' Immunoglobulin repertoire sequencing (AIRR-Seq, Rep-Seq) data is currently the
#' subject of much study. A key step in analyzing these data involves assigning
#' the closest known V(D)J germline alleles to the (often somatically mutated)
#' sample sequences using a tool such as IMGT/HighV-QUEST. However, if the
#' sample utilizes alleles not in the germline database used for alignment, this
#' step will fail. Additionally, this alignment has an associated error rate of
#' ~5%, notably among sequences carrying a large number of somatic
#' mutations. The purpose of TIgGER is to address these issues.
#'
#' @section  Allele detection and genotyping:
#' \itemize{
#'   \item  \link{findNovelAlleles}:       Detect novel alleles.
#'   \item  \link{plotNovel}:              Plot evidence of novel alleles.
#'   \item  \link{inferGenotype}:          Infer an Ig genotype using a frequency approach.
#'   \item  \link{inferGenotypeBayesian}:  Infer an Ig genotype using a Bayesian approach.
#'   \item  \link{plotGenotype}:           A colorful genotype visualization.
#'   \item  \link{genotypeFasta}:          Convert a genotype to sequences.
#'   \item  \link{reassignAlleles}:        Correct allele calls.
#'   \item  \link{generateEvidence}:       Generate evidence for the genotype and
#'                                         allele detection inference.
#' }
#'
#' @section  Mutation handling:
#' \itemize{
#'   \item  \link{getMutatedPositions}:      Find mutation locations.
#'   \item  \link{getMutCount}:              Find distance from germline.
#'   \item  \link{findUnmutatedCalls}:       Subset unmutated sequences.
#'   \item  \link{getPopularMutationCount}:  Find most common sequence's
#'                                           mutation count.
#'   \item  \link{insertPolymorphisms}:      Insert SNPs into a sequence.
#' }
#'
#' @section  Input, output and formatting:
#' \itemize{
#'   \item  \link{readIgFasta}:        Read a fasta file of Ig sequences.
#'   \item  \link{updateAlleleNames}:  Correct outdated allele names.
#'   \item  \link{sortAlleles}:        Sort allele names intelligently.
#'   \item  \link{cleanSeqs}:          Standardize sequence format.
#' }
#'
#' @name        tigger
#' @references
#' \enumerate{
#'   \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
#'         sequencing data reveals a high frequency of novel immunoglobulin V gene
#'         segment alleles. PNAS. 112(8):E862-70.
#' }
#'
#' @import      ggplot2
#' @importFrom  alakazam    getAllele getGene getFamily translateDNA DNA_COLORS checkColumns
#' @importFrom  doParallel  registerDoParallel
#' @importFrom  dplyr       do n desc %>%
#'                          glimpse distinct group_indices
#'                          as_data_frame data_frame
#'                          bind_cols bind_rows combine inner_join
#'                          filter select arrange
#'                          group_by ungroup
#'                          mutate pull rename slice
#'                          summarise transmute
#' @importFrom  foreach     foreach %dopar% registerDoSEQ
#' @importFrom  graphics    plot
#' @importFrom  gridExtra   arrangeGrob
#' @importFrom  gtools      ddirichlet
#' @importFrom  iterators   icount
#' @importFrom  lazyeval    interp
#' @importFrom  parallel    clusterEvalQ clusterExport makeCluster stopCluster
#' @importFrom  rlang       .data := sym syms
#' @importFrom  stats       na.omit setNames ecdf sd cor cov median mad
#'                          confint lm
#' @importFrom  stringi     stri_length stri_detect_fixed stri_replace_all_regex
#'                          stri_sub stri_sub<- stri_trans_toupper
#' @importFrom  tidyr       gather spread unnest
#' @importFrom  utils       citation
NULL

# Package loading actions
.onAttach <- function(libname, pkgname) {
    msg <- citation(pkgname)
    msg <- paste(c(format(msg, "citation")), collapse = "\n\n")
    packageStartupMessage(msg)
}

#' Infer a subject-specific genotype using a Bayesian approach
#'
#' \code{inferGenotypeBayesian} infers an subject's genotype by applying a Bayesian framework
#' with a Dirichlet prior for the multinomial distribution. Up to four distinct alleles are
#' allowed in an individual’s genotype. Four likelihood distributions were generated by
#' empirically fitting three high coverage genotypes from three individuals
#' (Laserson and Vigneault et al, 2014). A posterior probability is calculated for the
#' four most common alleles. The certainty of the highest probability model was
#' calculated using a Bayes factor (the most likely model divided by second-most likely model).
#' The larger the Bayes factor (K), the greater the certainty in the model.
#'
#' @details
#' Allele calls representing cases where multiple alleles have been
#' assigned to a single sample sequence are rare among unmutated
#' sequences but may result if nucleotides for certain positions are
#' not available. Calls containing multiple alleles are treated as
#' belonging to all groups. If \code{novel} is provided, all
#' sequences that are assigned to the same starting allele as any
#' novel germline allele will have the novel germline allele appended
#' to their assignment prior to searching for unmutated sequences.
#'
#' @param    data            a \code{data.frame} containing V allele
#'                           calls from a single subject. If \code{find_unmutated}
#'                           is \code{TRUE}, then the sample IMGT-gapped V(D)J sequence
#'                           should be provided in column \code{sequence_alignment}
#' @param    v_call          column in \code{data} with V allele calls.
#'                           Default is \code{"v_call"}.
#' @param    seq             name of the column in \code{data} with the
#'                           aligned, IMGT-numbered, V(D)J nucleotide sequence.
#'                           Default is \code{"sequence_alignment"}.
#' @param    find_unmutated  if \code{TRUE}, use \code{germline_db} to
#'                           find which samples are unmutated. Not needed
#'                           if \code{allele_calls} only represent
#'                           unmutated samples.
#' @param    germline_db     named vector of sequences containing the
#'                           germline sequences named in \code{allele_calls}.
#'                           Only required if \code{find_unmutated} is \code{TRUE}.
#' @param    novel           an optional \code{data.frame} of the type
#'                           novel returned by \link{findNovelAlleles} containing
#'                           germline sequences that will be utilized if
#'                           \code{find_unmutated} is \code{TRUE}. See Details.
#' @param    priors          a numeric vector of priors for the multinomial distribution.
#'                           The \code{priors} vector must be nine values that defined
#'                           the priors for the heterozygous (two allele),
#'                           trizygous (three allele), and quadrozygous (four allele)
#'                           distributions. The first two values of \code{priors} define
#'                           the prior for the heterozygous case, the next three values are for
#'                           the trizygous case, and the final four values are for the
#'                           quadrozygous case. Each set of priors should sum to one.
#'                           Note, each distribution prior is actually defined internally
#'                           by set of four numbers, with the unspecified final values
#'                           assigned to \code{0}; e.g., the heterozygous case is
#'                           \code{c(priors[1], priors[2], 0, 0)}. The prior for the
#'                           homozygous distribution is fixed at \code{c(1, 0, 0, 0)}.
#'
#' @return
#' A \code{data.frame} of alleles denoting the genotype of the subject with the log10
#' of the likelihood of each model and the log10 of the Bayes factor. The output
#' contains the following columns:
#'
#' \itemize{
#'   \item \code{gene}: The gene name without allele.
#'   \item \code{alleles}: Comma separated list of alleles for the given \code{gene}.
#'   \item \code{counts}: Comma separated list of observed sequences for each
#'         corresponding allele in the \code{alleles} list.
#'   \item \code{total}: The total count of observed sequences for the given \code{gene}.
#'   \item \code{note}: Any comments on the inference.
#'   \item \code{kh}: log10 likelihood that the \code{gene} is homozygous.
#'   \item \code{kd}: log10 likelihood that the \code{gene} is heterozygous.
#'   \item \code{kt}: log10 likelihood that the \code{gene} is trizygous
#'   \item \code{kq}: log10 likelihood that the \code{gene} is quadrozygous.
#'   \item \code{k_diff}: log10 ratio of the highest to second-highest zygosity likelihoods.
#' }
#'
#' @note
#' This method works best with data derived from blood, where a large
#' portion of sequences are expected to be unmutated. Ideally, there
#' should be hundreds of allele calls per gene in the input.
#'
#' @seealso \link{plotGenotype} for a colorful visualization and
#'          \link{genotypeFasta} to convert the genotype to nucleotide sequences.
#'          See \link{inferGenotype} to infer a subject-specific genotype using
#'          a frequency method
#'
#' @references
#' \enumerate{
#'   \item  Laserson U and Vigneault F, et al. High-resolution antibody dynamics of
#'          vaccine-induced immune responses. PNAS. 2014 111(13):4928-33.
#' }
#'
#' @examples
#' # Infer IGHV genotype, using only unmutated sequences, including novel alleles
#' inferGenotypeBayesian(AIRRDb, germline_db=SampleGermlineIGHV, novel=SampleNovel,
#'                       find_unmutated=TRUE, v_call="v_call", seq="sequence_alignment")
#'
#' @export
inferGenotypeBayesian <- function(data, germline_db=NA, novel=NA,
                                  v_call="v_call", seq="sequence_alignment",
                                  find_unmutated=TRUE,
                                  priors=c(0.6, 0.4, 0.4, 0.35, 0.25, 0.25, 0.25, 0.25, 0.25)){
    # Visibility hack
    . <- NULL

    allele_calls <- getAllele(data[[v_call]], first=FALSE, strip_d=FALSE)
    # Find the unmutated subset, if requested
    if(find_unmutated){
        if(is.na(germline_db[1])){
            stop("germline_db needed if find_unmutated is TRUE")
        }
        if (!is.null(nrow(novel))) {
            novel <- novel %>%
                filter(!is.na(!!rlang::sym("polymorphism_call"))) %>%
                select(!!!rlang::syms(c("germline_call", "polymorphism_call", "novel_imgt")))
            if(nrow(novel) > 0){
                # Extract novel alleles if any and add them to germline_db
                novel_gl <- novel$novel_imgt
                names(novel_gl) <- novel$polymorphism_call
                germline_db <- c(germline_db, novel_gl)
                # Add the novel allele calls to allele calls of the same starting allele
                for(r in 1:nrow(novel)){
                    ind <- grep(novel$germline_call[r], allele_calls, fixed=TRUE)
                    allele_calls[ind] <- allele_calls[ind] %>%
                        sapply(paste, novel$polymorphism_call[r], sep=",")
                }
            }
        }
        # Find unmutated sequences
        allele_calls <- findUnmutatedCalls(allele_calls,
                                          as.character(data[[seq]]),
                                          germline_db)
        if(length(allele_calls) == 0){
            stop("No unmutated sequences found! Set 'find_unmutated' to 'FALSE'.")
        }
    }

    # Find which rows' calls contain which genes
    gene_regex <- allele_calls %>% strsplit(",") %>% unlist() %>%
        getGene(strip_d=FALSE) %>%  unique() %>% paste("\\*", sep="")
    gene_groups <- sapply(gene_regex, grep, allele_calls, simplify=FALSE)
    names(gene_groups) <- gsub("\\*", "", gene_regex, fixed=TRUE)
    gene_groups <- gene_groups[sortAlleles(names(gene_groups))]

    # Make a table to store the resulting genotype
    gene <- names(gene_groups)
    #   alleles = counts = note = rep("", length(gene))
    #   total = sapply(gene_groups, length)
    #   genotype = cbind(gene, alleles, counts, total, note)
    alleles <- counts <- kh <- kd <- kt <- kq <- k_diff <- note <- rep("", length(gene))
    total <- sapply(gene_groups, length)
    genotype <- cbind(gene, alleles, counts, total, note, kh, kd, kt, kq, k_diff)

    # For each gene, find which alleles to include
    for (g in gene){
        # Keep only the part of the allele calls that uses the gene being analyzed
        ac <- allele_calls[gene_groups[[g]]] %>%
            strsplit(",") %>%
            lapply(function(x) x[grep(paste(g, "\\*", sep=""), x)]) %>%
            sapply(paste, collapse=",")
        t_ac <- table(ac) # table of allele calls
        potentials <- unique(unlist(strsplit(names(t_ac),","))) # potential alleles

        regexpotentials <- paste(gsub("\\*","\\\\*", potentials),"$",sep="")
        regexpotentials <-
            paste(regexpotentials,gsub("\\$",",",regexpotentials),sep="|")
        tmat <-
            sapply(regexpotentials, function(x) grepl(x, names(t_ac),fixed=FALSE))

        if (length(potentials) == 1 | length(t_ac) == 1){
            seqs_expl <- t(as.data.frame(apply(t(as.matrix(tmat)), 2, function(x) x *
                                                  t_ac)))
            rownames(seqs_expl) <- names(t_ac)[1]
        }else{
            seqs_expl <- as.data.frame(apply(tmat, 2, function(x) x *
                                                t_ac))
        }
        #       seqs_expl = as.data.frame(apply(tmat, 2, function(x) x*t_ac))
        colnames(seqs_expl) <- potentials
        # Add low (fake) counts
        sapply(colnames(seqs_expl), function(x){if(sum(rownames(seqs_expl) %in% paste(x)) == 0){
            seqs_expl <<- rbind(seqs_expl,rep(0,ncol(seqs_expl)));
            rownames(seqs_expl)[nrow(seqs_expl)] <<- paste(x)
            seqs_expl[rownames(seqs_expl) %in% paste(x),paste(x)] <<- 0.01

        }})

        # Build ratio dependent allele count distribution of multi assigned reads
        seqs_expl_single <- seqs_expl[grep(',',rownames(seqs_expl),invert = T),]

        seqs_expl_multi <- seqs_expl[grep(',',rownames(seqs_expl),invert = F),]
        if(is.null(nrow(seqs_expl_multi))){
            seqs_expl_multi <- t(as.data.frame(seqs_expl_multi))
            rownames(seqs_expl_multi) <- grep(',',rownames(seqs_expl),invert = F,value = T)
        }

        if(!is.null(nrow(seqs_expl_single))  && nrow(seqs_expl_single) !=0 && nrow(seqs_expl_single) != nrow(seqs_expl)){
            if(nrow(seqs_expl_multi)>1){
                seqs_expl_multi <- seqs_expl_multi[order(nchar(row.names(seqs_expl_multi))),]
            }
            sapply(1:nrow(seqs_expl_multi),function(x){
                genes <- unlist(strsplit(row.names(seqs_expl_multi)[x],','));
                counts <- seqs_expl_single[rownames(seqs_expl_single) %in% genes,genes]
                counts <- colSums(counts)
                counts_to_distribute <- seqs_expl_multi[x,genes]

                new_counts <- counts+((counts_to_distribute*counts)/sum(counts))
                for(i in 1:length(new_counts)){
                    gene_tmp <- names(new_counts)[i]
                    seqs_expl_single[rownames(seqs_expl_single) %in% gene_tmp,gene_tmp] <<- new_counts[i]
                }
            })
        }

        # Cycle through the table, including alleles to explain more sequences,
        # until we explain enough sequences
        #included = counts = character(0)
        #tot_expl = 0

        seqs_expl <- if(is.null(nrow(seqs_expl_single)) || nrow(seqs_expl_single) ==0 ){seqs_expl}else{seqs_expl_single}
        seqs_expl <- round(seqs_expl)
        if(sum(rowSums(seqs_expl) == 0 ) != 0){
            seqs_expl <- seqs_expl[rowSums(seqs_expl)!= 0, ]
        }

        allele_tot <- sort(apply(seqs_expl, 2, sum),decreasing=TRUE)
        len=min(length(allele_tot),4);
        #print(priors)
        probs <-get_probabilities_with_priors(sort(c(allele_tot,rep(0,4-len)),decreasing = T)[1:4],priors = priors)
        probs[probs==-Inf] <- -1000
        names(probs) <- c('H','D','T','Q')

        k <- sort(as.numeric(probs),decreasing = T);

        probs<-c(probs,k[1]-k[2])
        names(probs)[5] <- "k_diff"

        genotype[genotype[, "gene"] == g, "alleles"] <- paste(gsub("[^d\\*]*[d\\*]",
                                                                  "", names(allele_tot)[1:len]), collapse = ",")
        genotype[genotype[, "gene"] == g, "counts"] <- paste(as.numeric(allele_tot)[1:len],
                                                            collapse = ",")
        genotype[genotype[, "gene"] == g, "kh"] <- probs[1];
        genotype[genotype[, "gene"] == g, "kd"] <- probs[2];
        genotype[genotype[, "gene"] == g, "kt"] <- probs[3];
        genotype[genotype[, "gene"] == g, "kq"] <- probs[4];
        genotype[genotype[, "gene"] == g, "k_diff"] <- probs[5];
        #     }

    }


    geno <- as.data.frame(genotype, stringsAsFactors = FALSE)

    # Check for indistinguishable calls
    if(find_unmutated == TRUE){
        seqs <- genotypeFasta(geno, germline_db)
        dist_mat <- seqs %>%
            sapply(function(x) sapply((getMutatedPositions(seqs, x)), length))
        rownames(dist_mat) <- colnames(dist_mat)
        for (i in 1:nrow(dist_mat)){ dist_mat[i,i] = NA }
        same <- which(dist_mat == 0, arr.ind=TRUE)
        if (nrow(same) > 0 ) {
            for (r in 1:nrow(same)) {
                inds <- as.vector(same[r,])
                geno[getGene(rownames(dist_mat)[inds][1]),]$note <-
                    paste(rownames(dist_mat)[inds], collapse=" and ") %>%
                    paste("Cannot distinguish", .)
            }
        }
    }
    rownames(geno) <- NULL
    return(geno)
}


# Calculate models likelihood
#
#
# @param    X      a vector of counts
# @param    alpha_dirichlet      alpha parameter for dirichlet distribution
# @param    epsilon    epsilon
# @param    priors      a vector of priors
#
# @return  log10 of the likelihoods
get_probabilities_with_priors <- function(X, alpha_dirichlet=c(0.5,0.5,0.5,0.5)*2,
                                         epsilon=0.01,
                                         priors=c(0.5,0.5,0.33,0.33,0.33,0.25,0.25,0.25,0.25)){
    ## Hypotheses
    X<-sort(X,decreasing=TRUE)

    H1<-c(1,0,0,0)
    H2<-c(priors[1],priors[2],0,0)
    H3<-c(priors[3],priors[4],priors[5],0)
    H4<-c(priors[6],priors[7],priors[8],priors[9])

    E1<-ddirichlet((H1+epsilon)/sum(H1+epsilon),alpha_dirichlet+X)
    E2<-ddirichlet((H2+epsilon)/sum(H2+epsilon),alpha_dirichlet+X)
    E3<-ddirichlet((H3+epsilon)/sum(H3+epsilon),alpha_dirichlet+X)
    E4<-ddirichlet((H4+epsilon)/sum(H4+epsilon),alpha_dirichlet+X)



    while(sort(c(E1,E2,E3,E4),decreasing=TRUE)[2] == 0 ){

        X <- X/10
        E1<-ddirichlet((H1+epsilon)/sum(H1+epsilon),alpha_dirichlet+X)
        E2<-ddirichlet((H2+epsilon)/sum(H2+epsilon),alpha_dirichlet+X)
        E3<-ddirichlet((H3+epsilon)/sum(H3+epsilon),alpha_dirichlet+X)
        E4<-ddirichlet((H4+epsilon)/sum(H4+epsilon),alpha_dirichlet+X)

    }

    return(log10(c(E1,E2,E3,E4)))
}

#' Human IGHV germlines
#'
#' A \code{character} vector of all human IGHV germline gene segment alleles
#' in IMGT/GENE-DB (2019-06-01, 372 alleles).
#' See IMGT data updates: https://www.imgt.org/IMGTgenedbdoc/dataupdates.html.
#'
#' @name GermlineIGHV
#' @docType data
#' @format Values correspond to IMGT-gaped nucleotide sequences (with
#' nucleotides capitalized and gaps represented by ".") while names correspond
#' to stripped-down IMGT allele names (e.g. "IGHV1-18*01").
#'
#' @references
#' \enumerate{
#'   \item Xochelli, et al. (2014) Immunoglobulin heavy variable (IGHV) genes and
#'         alleles: new entities, new names and implications for research and
#'         prognostication in chronic lymphocytic leukaemia. Immunogenetics. 67(1):61-6.
#' }
#'
#' @keywords data
NULL

#' Example Human IGHV germlines
#'
#' A \code{character} vector of all 344 human IGHV germline gene segment alleles
#' in IMGT/GENE-DB release 2014-08-4.
#'
#' @name SampleGermlineIGHV
#' @docType data
#' @format Values correspond to IMGT-gaped nucleotide sequences (with
#' nucleotides capitalized and gaps represented by ".") while names correspond
#' to stripped-down IMGT allele names (e.g. "IGHV1-18*01").
#'
#' @references
#' \enumerate{
#'   \item Xochelli, et al. (2014) Immunoglobulin heavy variable (IGHV) genes and
#'         alleles: new entities, new names and implications for research and
#'         prognostication in chronic lymphocytic leukaemia. Immunogenetics. 67(1):61-6.
#' }
#'
#' @keywords data
NULL

#' Example human immune repertoire data
#'
#' A \code{data.frame} of example V(D)J immunoglobulin sequences derived from a
#' single individual (PGP1), sequenced on the Roche 454 platform, and assigned by
#' IMGT/HighV-QUEST to IGHV1 family alleles.
#'
#' @name SampleDb
#' @docType data
#' @format A \code{data.frame} where rows correspond to unique V(D)J sequences and
#' columns include:
#' \itemize{
#'   \item \code{"SEQUENCE_IMGT"}: IMGT-gapped V(D)J nucleotide sequence.
#'   \item \code{"V_CALL"}: IMGT/HighV-QUEST V segment allele calls.
#'   \item \code{"D_CALL"}: IMGT/HighV-QUEST D segment allele calls.
#'   \item \code{"J_CALL"}: IMGT/HighV-QUEST J segment allele calls.
#'   \item \code{"JUNCTION_LENGTH"}: Junction region length.
#' }
#' @seealso See \link{AIRRDb} for an AIRR formatted version of \code{SampleDb}.
#' @references
#' \enumerate{
#'   \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
#'         sequencing data reveals a high frequency of novel immunoglobulin V gene
#'         segment alleles. PNAS. 112(8):E862-70.
#' }
#'
#' @keywords data
NULL

#' Example human immune repertoire data
#'
#' A \code{data.frame} of example V(D)J immunoglobulin sequences derived from a
#' single individual (PGP1), sequenced on the Roche 454 platform, and assigned by
#' IMGT/HighV-QUEST to IGHV1 family alleles.
#'
#' @name AIRRDb
#' @docType data
#' @format A \code{data.frame} where rows correspond to unique V(D)J sequences and
#' columns include:
#' \itemize{
#'   \item \code{"sequence_alignment"}: IMGT-gapped V(D)J nucleotide sequence.
#'   \item \code{"v_call"}: IMGT/HighV-QUEST V segment allele calls.
#'   \item \code{"d_call"}: IMGT/HighV-QUEST D segment allele calls.
#'   \item \code{"j_call"}: IMGT/HighV-QUEST J segment allele calls.
#'   \item \code{"junction_length"}: Junction region length.
#' }
#' @seealso See \link{SampleDb} for Change-O formatted version of \code{AIRRDb}.
#' @references
#' \enumerate{
#'   \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
#'         sequencing data reveals a high frequency of novel immunoglobulin V gene
#'         segment alleles. PNAS. 112(8):E862-70.
#' }
#'
#' @keywords data
NULL

#' Example novel allele detection results
#'
#' A \code{data.frame} of novel allele detection results from \link{findNovelAlleles}.
#' Source data was a collection of V(D)J immunoglobulin sequences derived from a single
#' individual (PGP1), sequenced on the Roche 454 platform, and assigned by
#' IMGT/HighV-QUEST to IGHV1 family alleles.
#'
#' @name SampleNovel
#' @docType data
#' @format A \code{data.frame} where rows correspond to alleles checked for
#' polymorphisms and columns give results as well as parameters used to run
#' the test.
#'
#' @seealso See \link{findNovelAlleles} for detailed column descriptions.
#'
#' @references
#' \enumerate{
#'   \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
#'         sequencing data reveals a high frequency of novel immunoglobulin V gene
#'         segment alleles. PNAS. 112(8):E862-70.
#' }
#'
#' @keywords data
NULL

#' Example genotype inference results
#'
#' A \code{data.frame} of genotype inference results from \link{inferGenotype}
#' after novel allele detection via \link{findNovelAlleles}.
#' Source data was a collection of V(D)J immunoglobulin sequences derived from a single
#' individual (PGP1), sequenced on the Roche 454 platform, and assigned by
#' IMGT/HighV-QUEST to IGHV1 family alleles.
#'
#' @name SampleGenotype
#' @docType data
#' @format A \code{data.frame} where rows correspond to genes carried by an
#' individual and columns lists the alleles of those genes and their counts.
#'
#' @seealso See \link{inferGenotype} for detailed column descriptions.
#'
#' @references
#' \enumerate{
#'   \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
#'         sequencing data reveals a high frequency of novel immunoglobulin V gene
#'         segment alleles. PNAS. 112(8):E862-70.
#' }
#'
#' @keywords data
NULL

---
title: "Tool for Immunoglobulin Genotype Elucidation via Rep-Seq (TIgGER)"
author: "Daniel Gadala-Maria"
date: "Last modified `r Sys.Date()`"
output:
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes
  html_document:
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteEngine{knitr::rmarkdown} 
  %\VignetteIndexEntry{Polymorphism detection and genotyping} 
  %\usepackage[utf8]{inputenc}
---

## Introduction 

Adaptive immune receptor repertoire sequencing (AIRR-Seq, Rep-Seq) data is 
currently the subject of much study. A key step in analyzing these data involves 
assigning the closest known V(D)J germline alleles to the (often somatically mutated) 
sample sequences using a tool such as IMGT/HighV-QUEST ([[1]][1]). However, 
if the sample utilizes alleles not in the germline database used for alignment, 
this step will fail. Additionally, this alignment has an associated error rate 
of ~5% ([[2]][2]), notably among sequences carrying a large number of somatic 
mutations.

Here we provide a **T**ool for **I**mmuno**g**lobulin **G**enotype 
**E**lucidation via **R**ep-Seq (TIgGER). TIgGER addresses these issues by 
inferring the set of Immunoglobulin (Ig) alleles carried by an individual 
(including any novel alleles) and then using this set of alleles to correct 
the initial assignments given to sample sequences by existing tools. 

This vignette covers the following tasks:

1. Inferring the presence of novel IGHV alleles not in the germline database.
2. Inferring the personalized IGHV genotype of a sample.
3. Correcting the IGHV allele calls of a sample based on the IGHV genotype.

Additional information about the methods used by TIgGER is available in:

[Gadala-Maria D, Yaari G, Uduman M, Kleinstein SH (2015) Automated analysis of 
high-throughput B cell sequencing data reveals a high frequency of novel 
immunoglobulin V gene segment alleles. *PNAS* 
112(8):E862-70](http://www.pnas.org/content/early/2015/02/05/1417683112).

## Input

TIgGER requires two main inputs:

1. Pre-processed Ig sequence data
2. Database germline sequences

AIRR-seq data is input as a data frame following the AIRR or Change-O standard where 
each row represents a unique observation and and columns represent data about 
that observation. The required names of the required columns are provided below 
along with a description of each.

Column Name                                 | Description 
--------------------------------------------|---------------------------------------------------------
`SEQUENCE_IMGT` or `sequence_alignment`     | V(D)J sequence gapped in the IMGT gapped format ([[3]][3])
`V_CALL` or `v_call`                        | (Comma separated) name(s) of the nearest V allele(s) 
`J_CALL` or `j_call`                        | (Comma separated) name(s) of the nearest J allele(s) 
`JUNCTION` or `junction`                    | Junction nucleotide sequence
`JUNCTION_LENGTH` or `junction_length`      | Length of the junction region of the V(D)J sample

An example dataset is provided with the `tigger` package as `SampleDb` (Change-O format)
and `AIRRDb` (AIRR format). It contains unique functional sequences assigned to IGHV1 family genes 
isolated from individual PGP1 (referenced in Gadala-Maria *et al.* 2015).

The database of germline sequences should be provided in FASTA format with 
sequences gapped according to the IMGT numbering scheme ([[3]][3]). IGHV alleles in
the IMGT database (build 2014-08-4) are provided with this package as `SampleGermlineIGHV`. 
You may read in your own fasta file using `readIgFasta`.

```{r, eval=TRUE, message=FALSE, warning=FALSE}
# Load packages required for this example
library(tigger)
library(dplyr)
```

## Novel allele detection

Potential novel alleles can be detected by TIgGER. Some of these may be included
in the genotype later (see below). `findNovelAlleles` will return a `data.frame`
with a row for each allele tested for the presence of polymorphisms. If
polymorphisms are found and the novel sequence passes all quality tests (see
below), then the novel allele is named and the germline sequence is included in
the `data.frame`. Additionally, the result will contain metadata on the 
parameters used when searching for novel alleles (which can be optionally
changed in `findNovelAlleles`).

```{r, eval=TRUE, warning=FALSE}
# Detect novel alleles
novel <- findNovelAlleles(AIRRDb, SampleGermlineIGHV, nproc=1)
```

```{r, eval=TRUE, warning=FALSE}
# Extract and view the rows that contain successful novel allele calls
novel_rows <- selectNovel(novel)
```

The TIgGER procedure for identifying novel alleles (see citation above) involves
taking all sequences which align to a particular germline allele and, for each 
position along the aligned sequences, plotting the mutation frequency at that 
position as a function of the sequence-wide mutation count. While mutational 
hot-spots and cold-spots are both expected to have a *y*-intercept around zero, 
polymorphic positions will have a *y*-intercept larger than zero. The required 
minimum *y*-intercept may be specified in `findNovelAlleles` by `y_intercept`, but 
defaults to `1/8`. Passing this *y*-intercept threshold is the first of three 
pieces of evidence that support the novel allele.

The second piece of evidence supporting novel allele calls is the nucleotide
usage at the polymorphic positions as a function of sequence-wide mutation
count. We expect the polymorphic allele to be prevalent at all mutation counts,
and we expect the mutation count equal to the number of polymorphisms in the
novel sequence to be the most prevalent.

Finally, to avoid cases where a clonal expansion might lead to a false positive,
combinations of J gene and junction length are examined among sequences which 
*perfectly match* the proposed germline allele. A true novel allele is expected 
to utilize a wide range of J genes, and sequences with different junction length
can be ruled out as not being clonally related. The maximum portion of sequences
which can consist of a specific combination of J gene and junction length may be
specified in `findNovelAlleles` by `j_max`.

The three pieces of evidence described above can be viewed for any allele call
made by `findNovelAlleles` using the function `plotNovel`.


```{r, eval=TRUE, warning=FALSE, fig.width=6, fig.height=8}
# Plot evidence of the first (and only) novel allele from the example data
novel_row <- which(!is.na(novel$polymorphism_call))[1]
plotNovel(AIRRDb, novel[novel_row, ])
```


## Inferring genotypes

An individual's genotype can be inferred using the functions `inferGenotype` or 
`inferGenotypeBayesian`. Using one of this functions allows to remove from the
genotype rare/erroneous allele calls which may result from mutations in 
allele-differentiating regions. `inferGenotype` uses a frequency method to 
decide which alleles belong to the subjects genotype whereas 
`inferGenotypeBayesian` infers an subject's genotype applying a Bayesian 
framework and provides a confidence estimate associated with 
the genotype calls.


### Frequency genotyping approach

`inferGenotype` identifies the fewest alleles that account for 
nearly all (default is 7/8) of the allele calls made. The user may opt to only
use sequences which perfectly match germline alleles, and may opt to include 
potential novel alleles. (The genotype output is designed to be human readable, 
though `plotGenotype` can be used to make a colorful visualization.) For each 
allele, the number of sequences which  match the germline are listed in the same
order as the alleles are listed. The total number of sequences that match any 
allele of that gene is also given. To output these alleles as a names vector of 
nucleotide sequences, the user may use the function `genotypeFasta`. To save 
this vector to a fasta file, `writeFasta` may be used.

```{r, eval=TRUE, warning=FALSE, fig.width=4, fig.height=3}
# Infer the individual's genotype, using only unmutated sequences and checking
# for the use of the novel alleles inferred in the earlier step.
geno <- inferGenotype(AIRRDb, germline_db=SampleGermlineIGHV, novel=novel,
                      find_unmutated=TRUE)
# Save the genotype sequences to a vector
genotype_db <- genotypeFasta(geno, SampleGermlineIGHV, novel)
# Visualize the genotype and sequence counts
print(geno)
# Make a colorful visualization. Bars indicate presence, not proportion.
plotGenotype(geno, text_size = 10)
```

### Bayesian genotyping approach

The method `inferGenotypeBayesian` analyzes the posterior probabilities of 
possible allele distributions, considering up to four distinct alleles per 
V gene, corresponding to a gene duplication with both loci being heterozygous 
(i.e., homozygous, heterozygous with one copy of each allele, etc.). The 
posterior probabilities for these four possible models are compared and a Bayes
factor is calculated for the two most probable models. This Bayes factor 
reflects the confidence in the genotyping call of the method. The bayesian 
method doesn't use the strict cutoff criterion `fraction_to_explain` that 
`inferGenotype` uses wherein only the minimum set of alleles explaining 
88% (7/8) of apparently-unmutated sequences are included in the genotype.


```{r, eval=TRUE, warning=FALSE, fig.width=4, fig.height=3}
# Infer the individual's genotype, using the bayesian method
geno_bayesian <- inferGenotypeBayesian(AIRRDb, germline_db=SampleGermlineIGHV, 
                                       novel=novel, find_unmutated=TRUE)
# Visualize the genotype and sequence counts
print(geno_bayesian)
# Make a colorful visualization. Bars indicate presence, not proportion.
plotGenotype(geno_bayesian, text_size=10)
```

## Correcting allele calls

Finally, the original V allele calls may be limited to only those within the 
inferred genotype. This can be done by using the function `reassignAlleles`.
By correcting the calls in this manner, the user can greatly reduce the number of
ambiguous allele calls (where a single sample sequences is assigned to multiple
V alleles, thus preventing the mutations analysis of allele-differentiating
positions). Additionally, assignments to erroneous not-in-genotype alleles
(expected to be ~5% ([[2]][2]), as mentioned above, are corrected in this manner.

```{r, eval=TRUE, warning=FALSE}
# Use the personalized genotype to determine corrected allele assignments
# Updated genotype will be placed in the v_call_genotyped column
sample_db <- reassignAlleles(AIRRDb, genotype_db)
```

From here, one may proceed with further downstream analyses, but with the
advantage of having much-improved allele calls. Besides having discovered
alleles not in the IMGT database, the calls made by IMGT have been tailored to
the subject's genotype, greatly reducing the number of problematic calls, as
can be seen below.

```{r, eval=TRUE, warning=FALSE}
# Find the set of alleles in the original calls that were not in the genotype
not_in_genotype <- sample_db$v_call %>%
    strsplit(",") %>%
    unlist() %>%
    unique() %>%
    setdiff(names(genotype_db))

# Determine the fraction of calls that were ambiguous before/after correction
# and the fraction that contained original calls to non-genotype alleles. Note
# that by design, only genotype alleles are allowed in "after" calls.
data.frame(Ambiguous=c(mean(grepl(",", sample_db$v_call)),
                       mean(grepl(",", sample_db$v_call_genotyped))),
           NotInGenotype=c(mean(sample_db$v_call %in% not_in_genotype),
                           mean(sample_db$v_call_genotyped %in% not_in_genotype)),
           row.names=c("Before", "After")) %>% 
    t() %>% round(3)
```

## Evidence 

`generateEvidence` uses the final corrected calls, the novel alleles and genotype 
information, the final genotype sequences and the starting reference germlines
to build a table of evidence metrics supporting the final novel V allele detection.

```{r, eval=TRUE, warning=FALSE}
evidence <- generateEvidence(sample_db, novel, geno, genotype_db, SampleGermlineIGHV, fields = NULL)

evidence %>%
  select(gene, allele, polymorphism_call, sequences, unmutated_frequency)
```

In this example, `r evidence[["sequences"]]` sequences were unambiguously assigned
to allele `r evidence[["polymorphism_call"]]`, 
`r round(100*evidence[["unmutated_frequency"]],2)`%
of them unmutated.

## References

1. [Alamyar *et al.* (2010)](https://www.imgt.org/HighV-QUEST/)
2. [Munshaw and Kepler (2010)](https://pubmed.ncbi.nlm.nih.gov/20147303)
3. [Lefranc *et al.* (2003)](https://pubmed.ncbi.nlm.nih.gov/12477501)

[1]: https://www.imgt.org/HighV-QUEST/ "Alamyar et al. (2010)" 
[2]: https://pubmed.ncbi.nlm.nih.gov/20147303/ "Munshaw and Kepler (2010)" 
[3]: https://pubmed.ncbi.nlm.nih.gov/12477501 "Lefranc et al. (2003)"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{genotypeFasta}
\alias{genotypeFasta}
\title{Return the nucleotide sequences of a genotype}
\usage{
genotypeFasta(genotype, germline_db, novel = NA)
}
\arguments{
\item{genotype}{\code{data.frame} of alleles denoting a genotype,
as returned by \link{inferGenotype}.}

\item{germline_db}{vector of named nucleotide germline sequences
matching the alleles detailed in \code{genotype}.}

\item{novel}{an optional \code{data.frame} containing putative
novel alleles of the type returned by
\link{findNovelAlleles}.}
}
\value{
A named vector of strings containing the germline nucleotide
          sequences of the alleles in the provided genotype.
}
\description{
\code{genotypeFasta} converts a genotype table into a vector of nucleotide
sequences.
}
\examples{
# Find the sequences that correspond to the genotype
genotype_db <- genotypeFasta(SampleGenotype, SampleGermlineIGHV, SampleNovel)

}
\seealso{
\link{inferGenotype}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{findUnmutatedCalls}
\alias{findUnmutatedCalls}
\title{Determine which calls represent an unmutated allele}
\usage{
findUnmutatedCalls(allele_calls, sample_seqs, germline_db)
}
\arguments{
\item{allele_calls}{vector of strings representing Ig allele calls,
where multiple calls are separated by a comma.}

\item{sample_seqs}{V(D)J-rearranged sample sequences matching the order
of the given \code{allele_calls}.}

\item{germline_db}{vector of named nucleotide germline sequences}
}
\value{
A vector of strings containing the members of \code{allele_calls}
          that represent unmutated sequences.
}
\description{
\code{findUnmutatedCalls} determines which allele calls would represent a
perfect match with the germline sequence, given a vector of allele calls and
mutation counts. In the case of multiple alleles being assigned to a
sequence, only the subset that would represent a perfect match is returned.
}
\examples{
# Find which of the sample alleles are unmutated
calls <- findUnmutatedCalls(AIRRDb$v_call, AIRRDb$sequence_alignment,
                            germline_db=SampleGermlineIGHV)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{AIRRDb}
\alias{AIRRDb}
\title{Example human immune repertoire data}
\format{
A \code{data.frame} where rows correspond to unique V(D)J sequences and
columns include:
\itemize{
  \item \code{"sequence_alignment"}: IMGT-gapped V(D)J nucleotide sequence.
  \item \code{"v_call"}: IMGT/HighV-QUEST V segment allele calls.
  \item \code{"d_call"}: IMGT/HighV-QUEST D segment allele calls.
  \item \code{"j_call"}: IMGT/HighV-QUEST J segment allele calls.
  \item \code{"junction_length"}: Junction region length.
}
}
\description{
A \code{data.frame} of example V(D)J immunoglobulin sequences derived from a
single individual (PGP1), sequenced on the Roche 454 platform, and assigned by
IMGT/HighV-QUEST to IGHV1 family alleles.
}
\references{
\enumerate{
  \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
        sequencing data reveals a high frequency of novel immunoglobulin V gene
        segment alleles. PNAS. 112(8):E862-70.
}
}
\seealso{
See \link{SampleDb} for Change-O formatted version of \code{AIRRDb}.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{sortAlleles}
\alias{sortAlleles}
\title{Sort allele names}
\usage{
sortAlleles(allele_calls, method = c("name", "position"))
}
\arguments{
\item{allele_calls}{vector of strings representing Ig allele names.}

\item{method}{string defining the method to use when sorting alleles.
If \code{"name"} then sort in lexicographic order. If
\code{"position"} then sort by position in the locus, as
determined by the final two numbers in the gene name.}
}
\value{
A sorted vector of strings representing Ig allele names.
}
\description{
\code{sortAlleles} returns a sorted vector of strings representing Ig allele
names. Names are first sorted by gene family, then by gene, then by allele.
Duplicated genes have their alleles are sorted as if they were part of their
non-duplicated counterparts (e.g. \code{IGHV1-69D*01} comes after \code{IGHV1-69*01}
but before \code{IGHV1-69*02}), and non-localized genes (e.g. \code{IGHV1-NL1*01})
come last within their gene family.
}
\examples{
# Create a list of allele names
alleles <- c("IGHV1-69D*01","IGHV1-69*01","IGHV1-2*01","IGHV1-69-2*01",
             "IGHV2-5*01","IGHV1-NL1*01", "IGHV1-2*01,IGHV1-2*05",
             "IGHV1-2", "IGHV1-2*02", "IGHV1-69*02")

# Sort the alleles by name
sortAlleles(alleles)

# Sort the alleles by position in the locus
sortAlleles(alleles, method="pos")

}
\seealso{
Like \code{sortAlleles}, \link{updateAlleleNames} can help
         format a list of allele names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tigger.R
\name{tigger}
\alias{tigger}
\title{tigger}
\description{
Here we provide a \strong{T}ool for \strong{I}mmuno\strong{g}lobulin
\strong{G}enotype \strong{E}lucidation via \strong{R}ep-Seq (TIgGER).
TIgGER infers the set of Ig alleles carried by an
individual (including any novel alleles) and then uses this set of alleles to
correct the initial assignments given to sample sequences by existing tools.
}
\details{
Immunoglobulin repertoire sequencing (AIRR-Seq, Rep-Seq) data is currently the
subject of much study. A key step in analyzing these data involves assigning
the closest known V(D)J germline alleles to the (often somatically mutated)
sample sequences using a tool such as IMGT/HighV-QUEST. However, if the
sample utilizes alleles not in the germline database used for alignment, this
step will fail. Additionally, this alignment has an associated error rate of
~5%, notably among sequences carrying a large number of somatic
mutations. The purpose of TIgGER is to address these issues.
}
\section{Allele detection and genotyping}{

\itemize{
  \item  \link{findNovelAlleles}:       Detect novel alleles.
  \item  \link{plotNovel}:              Plot evidence of novel alleles.
  \item  \link{inferGenotype}:          Infer an Ig genotype using a frequency approach.
  \item  \link{inferGenotypeBayesian}:  Infer an Ig genotype using a Bayesian approach.
  \item  \link{plotGenotype}:           A colorful genotype visualization.
  \item  \link{genotypeFasta}:          Convert a genotype to sequences.
  \item  \link{reassignAlleles}:        Correct allele calls.
  \item  \link{generateEvidence}:       Generate evidence for the genotype and
                                        allele detection inference.
}
}

\section{Mutation handling}{

\itemize{
  \item  \link{getMutatedPositions}:      Find mutation locations.
  \item  \link{getMutCount}:              Find distance from germline.
  \item  \link{findUnmutatedCalls}:       Subset unmutated sequences.
  \item  \link{getPopularMutationCount}:  Find most common sequence's
                                          mutation count.
  \item  \link{insertPolymorphisms}:      Insert SNPs into a sequence.
}
}

\section{Input, output and formatting}{

\itemize{
  \item  \link{readIgFasta}:        Read a fasta file of Ig sequences.
  \item  \link{updateAlleleNames}:  Correct outdated allele names.
  \item  \link{sortAlleles}:        Sort allele names intelligently.
  \item  \link{cleanSeqs}:          Standardize sequence format.
}
}

\references{
\enumerate{
  \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
        sequencing data reveals a high frequency of novel immunoglobulin V gene
        segment alleles. PNAS. 112(8):E862-70.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{cleanSeqs}
\alias{cleanSeqs}
\title{Clean up nucleotide sequences}
\usage{
cleanSeqs(seqs)
}
\arguments{
\item{seqs}{vector of nucleotide sequences.}
}
\value{
A modified vector of nucleotide sequences.
}
\description{
\code{cleanSeqs} capitalizes nucleotides and replaces all characters
besides \code{c("A", "C", "G", "T", "-", ".")} with \code{"N"}.
}
\examples{
# Clean messy nucleotide sequences
seqs <- c("AGAT.taa-GAG...ATA", "GATACAGTXXZZAGNNPPACA")
cleanSeqs(seqs)

}
\seealso{
\link{sortAlleles} and \link{updateAlleleNames} can
         help format a list of allele names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{plotNovel}
\alias{plotNovel}
\title{Visualize evidence of novel V alleles}
\usage{
plotNovel(
  data,
  novel_row,
  v_call = "v_call",
  j_call = "j_call",
  seq = "sequence_alignment",
  junction = "junction",
  junction_length = "junction_length",
  pos_range_max = NULL,
  ncol = 1,
  multiplot = TRUE
)
}
\arguments{
\item{data}{\code{data.frame} containing repertoire data. See
\link{findNovelAlleles} for details.}

\item{novel_row}{single row from a data frame as output by
\link{findNovelAlleles} that contains a
polymorphism-containing germline allele.}

\item{v_call}{name of the column in \code{data} with V allele
calls. Default is \code{v_call}.}

\item{j_call}{name of the column in \code{data} with J allele calls.
Default is \code{j_call}.}

\item{seq}{name of the column in \code{data} with the
aligned, IMGT-numbered, V(D)J nucleotide sequence.
Default is \code{sequence_alignment}.}

\item{junction}{Junction region nucleotide sequence, which includes
the CDR3 and the two flanking conserved codons. Default
is \code{junction}.}

\item{junction_length}{number of junction nucleotides in the junction sequence.
Default is \code{junction_length}.}

\item{pos_range_max}{Name of the column in \code{data} with the ending
positions of the V alignment in the germline
(usually \code{v_germline_end}).}

\item{ncol}{number of columns to use when laying out the plots.}

\item{multiplot}{whether to return one single plot (\code{TRUE}) or a list
with the three individual plots (\code{FALSE}).}
}
\description{
\code{plotNovel} is be used to visualize the evidence of any novel V
alleles found using \link{findNovelAlleles}. It can also be used to
visualize the results for alleles that did
}
\details{
The first panel in the plot shows, for all sequences which align to a particular
germline allele, the mutation frequency at each position along the aligned
sequence as a function of the sequence-wide mutation count. Each line is a position.
Positions that contain polymorphisms (rather than somatic hypermutations)
will exhibit a high apparent mutation frequency for a range of
sequence-wide mutation counts. The positions are color coded as follows:

\itemize{
  \item  red:    the position(s) pass(ess) the novel allele test
  \item  yellow: the position(s) pass(ess) the y-intercept test but not
                        other tests
  \item  blue:   the position(s) didn't pass the y-intercept test and
                        was(were) not further considered
}

The second panel shows the nucleotide usage at each of the polymorphic positions
as a function of sequence-wide mutation count. If no polymorphisms were identified,
the panel will show the mutation count.

To avoid cases where a clonal expansion might lead to a false positive, TIgGER examines
the combinations of J gene and junction length among sequences which perfectly
match the proposed germline allele. Clonally related sequences usually share
the same V gene, J gene and junction length. Requiring the novel allele
to be found in different combinations of J gene and junction lengths
is a proxy for requiring it to be found in different clonal lineages.
}
\examples{
# Plot the evidence for the first (and only) novel allele in the example data
novel <- selectNovel(SampleNovel)
plotNovel(AIRRDb, novel[1, ], v_call="v_call", j_call="j_call",
          seq="sequence_alignment", junction="junction", junction_length="junction_length",
          multiplot=TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{reassignAlleles}
\alias{reassignAlleles}
\title{Correct allele calls based on a personalized genotype}
\usage{
reassignAlleles(
  data,
  genotype_db,
  v_call = "v_call",
  seq = "sequence_alignment",
  method = "hamming",
  path = NA,
  keep_gene = c("gene", "family", "repertoire")
)
}
\arguments{
\item{data}{\code{data.frame} containing V allele calls from a
single subject and the sample IMGT-gapped V(D)J sequences under
\code{seq}.}

\item{genotype_db}{vector of named nucleotide germline sequences
matching the calls detailed in \code{allele_calls}
and personalized to the subject}

\item{v_call}{name of the column in \code{data} with V allele
calls. Default is \code{v_call}.}

\item{seq}{name of the column in \code{data} with the
aligned, IMGT-numbered, V(D)J nucleotide sequence.
Default is SEQUENCE_IMGT}

\item{method}{method to use when realigning sequences to
the genotype_db sequences. Currently, only \code{"hamming"}
(for Hamming distance) is implemented.}

\item{path}{directory containing the tool used in the
realignment method, if needed. Hamming distance does
not require a path to a tool.}

\item{keep_gene}{string indicating if the gene (\code{"gene"}),
family (\code{"family"}) or complete repertoire
(\code{"repertoire"}) assignments should be performed.
Use of \code{"gene"} increases speed by minimizing required number of
alignments, as gene level assignments will be maintained when possible.}
}
\value{
A modified input \code{data.frame} containing the best allele call from
          among the sequences listed in \code{genotype_db} in the
          \code{v_call_genotyped} column.
}
\description{
\code{reassignAlleles} uses a subject-specific genotype to correct
correct preliminary allele assignments of a set of sequences derived
from a single subject.
}
\details{
In order to save time, initial gene assignments are preserved and
the allele calls are chosen from among those provided in \code{genotype_db},
based on a simple alignment to the sample sequence.
}
\examples{
# Extract the database sequences that correspond to the genotype
genotype_db <- genotypeFasta(SampleGenotype, SampleGermlineIGHV, novel=SampleNovel)

# Use the personalized genotype to determine corrected allele assignments
output_db <- reassignAlleles(AIRRDb, genotype_db, v_call="v_call",
                             seq="sequence_alignment")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{inferGenotype}
\alias{inferGenotype}
\title{Infer a subject-specific genotype using a frequency method}
\usage{
inferGenotype(
  data,
  germline_db = NA,
  novel = NA,
  v_call = "v_call",
  seq = "sequence_alignment",
  fraction_to_explain = 0.875,
  gene_cutoff = 1e-04,
  find_unmutated = TRUE
)
}
\arguments{
\item{data}{\code{data.frame} containing V allele
calls from a single subject.}

\item{germline_db}{named vector of sequences containing the
germline sequences named in
\code{allele_calls}. Only required if
\code{find_unmutated} is \code{TRUE}.}

\item{novel}{optional \code{data.frame} of the type
novel returned by
\link{findNovelAlleles} containing
germline sequences that will be utilized if
\code{find_unmutated} is \code{TRUE}. See
Details.}

\item{v_call}{column in \code{data} with V allele calls.
Default is \code{"v_call"}.}

\item{seq}{name of the column in \code{data} with the
aligned, IMGT-numbered, V(D)J nucleotide sequence.
Default is \code{sequence_alignment}.}

\item{fraction_to_explain}{the portion of each gene that must be
explained by the alleles that will be included
in the genotype.}

\item{gene_cutoff}{either a number of sequences or a fraction of
the length of \code{allele_calls} denoting the
minimum number of times a gene must be
observed in \code{allele_calls} to be included
in the genotype.}

\item{find_unmutated}{if \code{TRUE}, use \code{germline_db} to
find which samples are unmutated. Not needed
if \code{allele_calls} only represent
unmutated samples.}
}
\value{
A \code{data.frame} of alleles denoting the genotype of the subject containing
the following columns:

\itemize{
  \item \code{gene}: The gene name without allele.
  \item \code{alleles}: Comma separated list of alleles for the given \code{gene}.
  \item \code{counts}: Comma separated list of observed sequences for each
        corresponding allele in the \code{alleles} list.
  \item \code{total}: The total count of observed sequences for the given \code{gene}.
  \item \code{note}: Any comments on the inference.
}
}
\description{
\code{inferGenotype} infers an subject's genotype using a frequency method.
The genotype is inferred by finding the minimum number set of alleles that
can explain the majority of each gene's calls. The most common allele of
each gene is included in the genotype first, and the next most common allele
is added until the desired fraction of alleles can be explained. In this
way, mistaken allele calls (resulting from sequences which
by chance have been mutated to look like another allele) can be removed.
}
\details{
Allele calls representing cases where multiple alleles have been
assigned to a single sample sequence are rare among unmutated
sequences but may result if nucleotides for certain positions are
not available. Calls containing multiple alleles are treated as
belonging to all groups. If \code{novel} is provided, all
sequences that are assigned to the same starting allele as any
novel germline allele will have the novel germline allele appended
to their assignment prior to searching for unmutated sequences.
}
\note{
This method works best with data derived from blood, where a large
portion of sequences are expected to be unmutated. Ideally, there
should be hundreds of allele calls per gene in the input.
}
\examples{
# Infer IGHV genotype, using only unmutated sequences, including novel alleles
inferGenotype(AIRRDb, germline_db=SampleGermlineIGHV, novel=SampleNovel,
              find_unmutated=TRUE)

}
\seealso{
\link{plotGenotype} for a colorful visualization and
         \link{genotypeFasta} to convert the genotype to nucleotide sequences.
         See \link{inferGenotypeBayesian} to infer a subject-specific genotype
         using a Bayesian approach.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{insertPolymorphisms}
\alias{insertPolymorphisms}
\title{Insert polymorphisms into a nucleotide sequence}
\usage{
insertPolymorphisms(sequence, positions, nucleotides)
}
\arguments{
\item{sequence}{starting nucleotide sequence.}

\item{positions}{numeric vector of positions which to be changed.}

\item{nucleotides}{character vector of nucleotides to which to change the
positions.}
}
\value{
A sequence with the desired nucleotides in the provided locations.
}
\description{
\code{insertPolymorphisms} replaces nucleotides in the desired locations of a
provided sequence.
}
\examples{
insertPolymorphisms("HUGGED", c(1, 6, 2), c("T", "R", "I"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{readIgFasta}
\alias{readIgFasta}
\title{Read immunoglobulin sequences}
\usage{
readIgFasta(fasta_file, strip_down_name = TRUE, force_caps = TRUE)
}
\arguments{
\item{fasta_file}{fasta-formatted file of immunoglobulin sequences.}

\item{strip_down_name}{if \code{TRUE}, will extract only the allele name
from the strings fasta file's sequence names.}

\item{force_caps}{if \code{TRUE}, will force nucleotides to
uppercase.}
}
\value{
Named vector of strings representing Ig alleles.
}
\description{
\code{readIgFasta} reads a fasta-formatted file of immunoglobulin (Ig)
sequences and returns a named vector of those sequences.
}
\examples{
\dontrun{
# germlines <- readIgFasta("ighv.fasta")
}

}
\seealso{
\link{writeFasta} to do the inverse.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{selectNovel}
\alias{selectNovel}
\title{Select rows containing novel alleles}
\usage{
selectNovel(novel, keep_alleles = FALSE)
}
\arguments{
\item{novel}{\code{data.frame} of the type returned by
\link{findNovelAlleles}.}

\item{keep_alleles}{\code{logical} indicating if different alleles
leading to the same novel sequence should be kept.
See Details.}
}
\value{
A \code{data.frame} containing only unique, novel alleles (if any)
that were in the input.
}
\description{
\code{selectNovel} takes the result from \link{findNovelAlleles} and
selects only the rows containing unique, novel alleles.
}
\details{
If, for instance, subject has in his genome \code{IGHV1-2*02} and a novel
allele equally close to \code{IGHV1-2*02} and \code{IGHV1-2*05}, the novel allele may be
detected by analyzing sequences that best align to either of these alleles.
If \code{keep_alleles} is \code{TRUE}, both polymorphic allele calls will
be retained. In the case that multiple mutation ranges are checked for the
same allele, only one mutation range will be kept in the output.
}
\examples{
novel <- selectNovel(SampleNovel)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{subsampleDb}
\alias{subsampleDb}
\title{Subsample repertoire}
\usage{
subsampleDb(
  data,
  gene = "v_call",
  mode = c("gene", "allele", "family"),
  min_n = 1,
  max_n = NULL,
  group = NULL
)
}
\arguments{
\item{data}{\code{data.frame} containing repertoire data.}

\item{gene}{name of the column in \code{data} with allele calls. Default
is \code{v_call}.}

\item{mode}{one of \code{c("gene", "family", "allele")} defining the degree of
specificity regarding allele calls when subsetting sequences.
Determines how \code{data} will be split into subsets from
which the same number of sequences will be subsampled. See
also \code{group}.}

\item{min_n}{minimum number of observations to sample from each group. A group with
less observations than the minimum is excluded.}

\item{max_n}{maximum number of observations to sample for all \code{mode} groups.
If \code{NULL}, it will be set automatically to the size of
the smallest group. If \code{max_n} is larger than the available
number of sequences for any \code{mode} group, it will be
automatically adjusted and the effective \code{max_n} used
will be the size of the smallest \code{mode} group.}

\item{group}{columns containing additional grouping variables, e.g. sample_id.
These groups will be subsampled independently. If
\code{max_n} is \code{NULL}, a \code{max_n} will be
automatically set for each \code{group}.}
}
\value{
Subsampled version of the input \code{data}.
}
\description{
\code{subsampleDb} will sample the same number of sequences for each gene, family
or allele (specified with \code{mode}) in \code{data}. Samples or subjects can
be subsampled independent by setting \code{group}.
}
\details{
\code{data} will be split into gene, allele or family subsets (\code{mode}) from
which the same number of sequences will be subsampled. If \code{mode=gene},
for each gene in the field \code{gene} from \code{data}, a maximum of
\code{max_n} sequences will be subsampled. Input sequences
that have multiple gene calls (ties), can be subsampled from any of their calls,
but these duplicated samplings will be removed, and the final
subsampled \code{data} will contain unique rows.
}
\examples{
set.seed(1)
subsampleDb(AIRRDb)

}
\seealso{
\link{selectNovel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{getMutCount}
\alias{getMutCount}
\title{Determine the mutation counts from allele calls}
\usage{
getMutCount(samples, allele_calls, germline_db)
}
\arguments{
\item{samples}{vector of IMGT-gapped sample V sequences}

\item{allele_calls}{vector of strings representing Ig allele calls for
the sequences in \code{samples}, where multiple
calls are separated by a comma}

\item{germline_db}{vector of named nucleotide germline sequences
matching the calls detailed in \code{allele_calls}}
}
\value{
A list equal in length to \code{samples}, containing the Hamming
          distance to each germline allele contained within each call within
          each element of \code{samples}
}
\description{
\code{getMutCount} takes a set of nucleotide sequences and their allele calls
and determines the distance between that sequence and any germline alleles
contained within the call
}
\examples{
# Insert a mutation into a germline sequence
s2 <- s3 <- SampleGermlineIGHV[1]
stringi::stri_sub(s2, 103, 103) <- "G"
stringi::stri_sub(s3, 107, 107) <- "C"

sample_seqs <- c(SampleGermlineIGHV[2], s2, s3)

# Pretend that one sample sequence has received an ambiguous allele call
sample_alleles <- c(paste(names(SampleGermlineIGHV[1:2]), collapse=","),
                    names(SampleGermlineIGHV[2]),
                    names(SampleGermlineIGHV[1]))

# Compare each sequence to its assigned germline(s) to determine the distance
getMutCount(sample_seqs, sample_alleles, SampleGermlineIGHV)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{plotGenotype}
\alias{plotGenotype}
\title{Show a colorful representation of a genotype}
\usage{
plotGenotype(
  genotype,
  facet_by = NULL,
  gene_sort = c("name", "position"),
  text_size = 12,
  silent = FALSE,
  ...
)
}
\arguments{
\item{genotype}{\code{data.frame} of alleles denoting a genotype,
as returned by \link{inferGenotype}.}

\item{facet_by}{column name in \code{genotype} to facet the plot by.
if \code{NULL}, then do not facet the plot.}

\item{gene_sort}{string defining the method to use when sorting alleles.
if \code{"name"} then sort in lexicographic order. If
\code{"position"} then sort by position in the locus, as
determined by the final two numbers in the gene name.}

\item{text_size}{point size of the plotted text.}

\item{silent}{if \code{TRUE} do not draw the plot and just return the ggplot
object; if \code{FALSE} draw the plot.}

\item{...}{additional arguments to pass to ggplot2::theme.}
}
\value{
A ggplot object defining the plot.
}
\description{
\code{plotGenotype} plots a genotype table.
}
\examples{
# Plot genotype
plotGenotype(SampleGenotype)

# Facet by subject
genotype_a <- genotype_b <- SampleGenotype
genotype_a$SUBJECT <- "A"
genotype_b$SUBJECT <- "B"
geno_sub <- rbind(genotype_a, genotype_b)
plotGenotype(geno_sub, facet_by="SUBJECT", gene_sort="pos")

}
\seealso{
\link{inferGenotype}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesian.R
\name{inferGenotypeBayesian}
\alias{inferGenotypeBayesian}
\title{Infer a subject-specific genotype using a Bayesian approach}
\usage{
inferGenotypeBayesian(
  data,
  germline_db = NA,
  novel = NA,
  v_call = "v_call",
  seq = "sequence_alignment",
  find_unmutated = TRUE,
  priors = c(0.6, 0.4, 0.4, 0.35, 0.25, 0.25, 0.25, 0.25, 0.25)
)
}
\arguments{
\item{data}{a \code{data.frame} containing V allele
calls from a single subject. If \code{find_unmutated}
is \code{TRUE}, then the sample IMGT-gapped V(D)J sequence
should be provided in column \code{sequence_alignment}}

\item{germline_db}{named vector of sequences containing the
germline sequences named in \code{allele_calls}.
Only required if \code{find_unmutated} is \code{TRUE}.}

\item{novel}{an optional \code{data.frame} of the type
novel returned by \link{findNovelAlleles} containing
germline sequences that will be utilized if
\code{find_unmutated} is \code{TRUE}. See Details.}

\item{v_call}{column in \code{data} with V allele calls.
Default is \code{"v_call"}.}

\item{seq}{name of the column in \code{data} with the
aligned, IMGT-numbered, V(D)J nucleotide sequence.
Default is \code{"sequence_alignment"}.}

\item{find_unmutated}{if \code{TRUE}, use \code{germline_db} to
find which samples are unmutated. Not needed
if \code{allele_calls} only represent
unmutated samples.}

\item{priors}{a numeric vector of priors for the multinomial distribution.
The \code{priors} vector must be nine values that defined
the priors for the heterozygous (two allele),
trizygous (three allele), and quadrozygous (four allele)
distributions. The first two values of \code{priors} define
the prior for the heterozygous case, the next three values are for
the trizygous case, and the final four values are for the
quadrozygous case. Each set of priors should sum to one.
Note, each distribution prior is actually defined internally
by set of four numbers, with the unspecified final values
assigned to \code{0}; e.g., the heterozygous case is
\code{c(priors[1], priors[2], 0, 0)}. The prior for the
homozygous distribution is fixed at \code{c(1, 0, 0, 0)}.}
}
\value{
A \code{data.frame} of alleles denoting the genotype of the subject with the log10
of the likelihood of each model and the log10 of the Bayes factor. The output
contains the following columns:

\itemize{
  \item \code{gene}: The gene name without allele.
  \item \code{alleles}: Comma separated list of alleles for the given \code{gene}.
  \item \code{counts}: Comma separated list of observed sequences for each
        corresponding allele in the \code{alleles} list.
  \item \code{total}: The total count of observed sequences for the given \code{gene}.
  \item \code{note}: Any comments on the inference.
  \item \code{kh}: log10 likelihood that the \code{gene} is homozygous.
  \item \code{kd}: log10 likelihood that the \code{gene} is heterozygous.
  \item \code{kt}: log10 likelihood that the \code{gene} is trizygous
  \item \code{kq}: log10 likelihood that the \code{gene} is quadrozygous.
  \item \code{k_diff}: log10 ratio of the highest to second-highest zygosity likelihoods.
}
}
\description{
\code{inferGenotypeBayesian} infers an subject's genotype by applying a Bayesian framework
with a Dirichlet prior for the multinomial distribution. Up to four distinct alleles are
allowed in an individual’s genotype. Four likelihood distributions were generated by
empirically fitting three high coverage genotypes from three individuals
(Laserson and Vigneault et al, 2014). A posterior probability is calculated for the
four most common alleles. The certainty of the highest probability model was
calculated using a Bayes factor (the most likely model divided by second-most likely model).
The larger the Bayes factor (K), the greater the certainty in the model.
}
\details{
Allele calls representing cases where multiple alleles have been
assigned to a single sample sequence are rare among unmutated
sequences but may result if nucleotides for certain positions are
not available. Calls containing multiple alleles are treated as
belonging to all groups. If \code{novel} is provided, all
sequences that are assigned to the same starting allele as any
novel germline allele will have the novel germline allele appended
to their assignment prior to searching for unmutated sequences.
}
\note{
This method works best with data derived from blood, where a large
portion of sequences are expected to be unmutated. Ideally, there
should be hundreds of allele calls per gene in the input.
}
\examples{
# Infer IGHV genotype, using only unmutated sequences, including novel alleles
inferGenotypeBayesian(AIRRDb, germline_db=SampleGermlineIGHV, novel=SampleNovel,
                      find_unmutated=TRUE, v_call="v_call", seq="sequence_alignment")

}
\references{
\enumerate{
  \item  Laserson U and Vigneault F, et al. High-resolution antibody dynamics of
         vaccine-induced immune responses. PNAS. 2014 111(13):4928-33.
}
}
\seealso{
\link{plotGenotype} for a colorful visualization and
         \link{genotypeFasta} to convert the genotype to nucleotide sequences.
         See \link{inferGenotype} to infer a subject-specific genotype using
         a frequency method
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{getPopularMutationCount}
\alias{getPopularMutationCount}
\title{Find mutation counts for frequency sequences}
\usage{
getPopularMutationCount(
  data,
  germline_db,
  v_call = "v_call",
  seq = "sequence_alignment",
  gene_min = 0.001,
  seq_min = 50,
  seq_p_of_max = 1/8,
  full_return = FALSE
)
}
\arguments{
\item{data}{\code{data.frame} in the Change-O format. See
\link{findNovelAlleles} for a list of required
columns.}

\item{germline_db}{named list of IMGT-gapped germline sequences.}

\item{v_call}{name of the column in \code{data} with V allele calls.
Default is \code{v_call}.}

\item{seq}{name of the column in \code{data} with the
aligned, IMGT-numbered, V(D)J nucleotide sequence.
Default is \code{sequence_alignment}.}

\item{gene_min}{portion of all unique sequences a gene must
constitute to avoid exclusion.}

\item{seq_min}{number of copies of the V that must be present for
to avoid exclusion.}

\item{seq_p_of_max}{ror each gene, the fraction of the most common V sequence
count that a sequence must meet to avoid exclusion.}

\item{full_return}{if \code{TRUE}, will return all \code{data} columns and
will include sequences with mutation count < 1.}
}
\value{
A data frame of genes that have a frequent sequence mutation count
         above 1.
}
\description{
\code{getPopularMutationCount} determines which sequences occur frequently
for each V gene and returns the mutation count of those sequences.
}
\examples{
getPopularMutationCount(AIRRDb, SampleGermlineIGHV)

}
\seealso{
\link{getMutatedPositions} can be used to find which positions
         of a set of sequences are mutated.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tigger.R
\docType{package}
\name{tigger-package}
\alias{tigger-package}
\title{tigger: Infers Novel Immunoglobulin Alleles from Sequencing Data}
\description{
Infers the V genotype of an individual from immunoglobulin (Ig) repertoire sequencing data (AIRR-Seq, Rep-Seq). Includes detection of any novel alleles. This information is then used to correct existing V allele calls from among the sample sequences. Citations: Gadala-Maria, et al (2015) \doi{10.1073/pnas.1417683112}, Gadala-Maria, et al (2019) \doi{10.3389/fimmu.2019.00129}.
}
\seealso{
Useful links:
\itemize{
  \item \url{http://tigger.readthedocs.io}
  \item Report bugs at \url{https://github.com/immcantation/tigger/issues}
}

}
\author{
\strong{Maintainer}: Susanna Marquez \email{susanna.marquez@yale.edu}

Authors:
\itemize{
  \item Daniel Gadala-Maria \email{daniel.gadala-maria@yale.edu}
  \item Moriah Cohen \email{moriah.cohen@biu.ac.il}
  \item Jason Vander Heiden \email{jason.vanderheiden@gmail.com}
  \item Gur Yaari \email{gur.yaari@biu.ac.il}
  \item Steven Kleinstein \email{steven.kleinstein@yale.edu} [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{writeFasta}
\alias{writeFasta}
\title{Write to a fasta file}
\usage{
writeFasta(named_sequences, file, width = 60, append = FALSE)
}
\arguments{
\item{named_sequences}{vector of named string representing sequences}

\item{file}{the name of the output file.}

\item{width}{the number of characters to be printed per line.
if not between 1 and 255, width with be infinite.}

\item{append}{\code{logical} indicating if the output should be
appended to \code{file} instead of overwriting it}
}
\value{
A named vector of strings representing Ig alleles.
}
\description{
\code{writeFasta} writes a named vector of sequences to a file in fasta
format.
}
\examples{
\dontrun{
# writeFasta(germlines, "ighv.fasta")
}

}
\seealso{
\link{readIgFasta} to do the inverse.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evidence.R
\name{generateEvidence}
\alias{generateEvidence}
\title{Generate evidence}
\usage{
generateEvidence(
  data,
  novel,
  genotype,
  genotype_db,
  germline_db,
  j_call = "j_call",
  junction = "junction",
  fields = NULL
)
}
\arguments{
\item{data}{a \code{data.frame} containing sequence data that has been
passed through \link{reassignAlleles} to correct the allele
assignments.}

\item{novel}{the \code{data.frame} returned by \link{findNovelAlleles}.}

\item{genotype}{the \code{data.frame} of alleles generated with \link{inferGenotype}
denoting the genotype of the subject.}

\item{genotype_db}{a vector of named nucleotide germline sequences in the genotype.
Returned by \link{genotypeFasta}.}

\item{germline_db}{the original uncorrected germline database used to by
\link{findNovelAlleles} to identify novel alleles.}

\item{j_call}{name of the column in \code{data} with J allele calls.
Default is \code{j_call}.}

\item{junction}{Junction region nucleotide sequence, which includes
the CDR3 and the two flanking conserved codons. Default
is \code{junction}.}

\item{fields}{character vector of column names used to split the data to
identify novel alleles, if any. If \code{NULL} then the data is
not divided by grouping variables.}
}
\value{
Returns the \code{genotype} input \code{data.frame} with the following additional columns
providing supporting evidence for each inferred allele:

\itemize{
  \item \code{field_id}: Data subset identifier, defined with the input parameter \code{fields}.
  \item A variable number of columns, specified with the input parameter \code{fields}.
  \item \code{polymorphism_call}: The novel allele call.
  \item \code{novel_imgt}: The novel allele sequence.
  \item \code{closest_reference}: The closest reference gene and allele in
        the \code{germline_db} database.
  \item \code{closest_reference_imgt}: Sequence of the closest reference gene and
        allele in the \code{germline_db} database.
  \item \code{germline_call}: The input (uncorrected) V call.
  \item \code{germline_imgt}: Germline sequence for \code{germline_call}.
  \item \code{nt_diff}: Number of nucleotides that differ between the new allele and
        the closest reference (\code{closest_reference}) in the \code{germline_db} database.
  \item \code{nt_substitutions}: A comma separated list of specific nucleotide
        differences (e.g. \code{112G>A}) in the novel allele.
  \item \code{aa_diff}: Number of amino acids that differ between the new allele and the closest
        reference (\code{closest_reference}) in the \code{germline_db} database.
  \item \code{aa_substitutions}: A comma separated list with specific amino acid
        differences (e.g. \code{96A>N}) in the novel allele.
  \item \code{sequences}: Number of sequences unambiguously assigned to this allele.
  \item \code{unmutated_sequences}: Number of records with the unmutated novel allele sequence.
  \item \code{unmutated_frequency}: Proportion of records with the unmutated novel allele
        sequence (\code{unmutated_sequences / sequences}).
  \item \code{allelic_percentage}: Percentage at which the (unmutated) allele is observed
        in the sequence dataset compared  to other (unmutated) alleles.
  \item \code{unique_js}: Number of unique J sequences found associated with the
        novel allele. The sequences are those who have been unambiguously assigned
        to the novel allele (\code{polymorphism_call}).
  \item \code{unique_cdr3s}: Number of unique CDR3s associated with the inferred allele.
        The sequences are those who have been unambiguously assigned to the
        novel allele (polymorphism_call).
  \item \code{mut_min}: Minimum mutation considered by the algorithm.
  \item \code{mut_max}: Maximum mutation considered by the algorithm.
  \item \code{pos_min}: First position of the sequence considered by the algorithm (IMGT numbering).
  \item \code{pos_max}: Last position of the sequence considered by the algorithm (IMGT numbering).
  \item \code{y_intercept}: The y-intercept above which positions were considered
        potentially polymorphic.
  \item \code{alpha}: Significance threshold to be used when constructing the
        confidence interval for the y-intercept.
  \item \code{min_seqs}: Input \code{min_seqs}. The minimum number of total sequences
        (within the desired mutational range and nucleotide range) required
        for the samples to be considered.
  \item \code{j_max}: Input \code{j_max}. The maximum fraction of sequences perfectly
        aligning to a potential novel allele that are allowed to utilize to a particular
        combination of junction length and J gene.
  \item \code{min_frac}: Input \code{min_frac}. The minimum fraction of sequences that must
        have usable nucleotides in a given position for that position to be considered.
  \item \code{note}: Comments regarding the novel allele inference.
}
}
\description{
\code{generateEvidence} builds a table of evidence metrics for the final novel V
allele detection and genotyping inferences.
}
\examples{
\donttest{
# Generate input data
novel <- findNovelAlleles(AIRRDb, SampleGermlineIGHV,
    v_call="v_call", j_call="j_call", junction="junction",
    junction_length="junction_length", seq="sequence_alignment")
genotype <- inferGenotype(AIRRDb, find_unmutated=TRUE,
                          germline_db=SampleGermlineIGHV,
                          novel=novel,
                          v_call="v_call", seq="sequence_alignment")
genotype_db <- genotypeFasta(genotype, SampleGermlineIGHV, novel)
data_db <- reassignAlleles(AIRRDb, genotype_db,
v_call="v_call", seq="sequence_alignment")

# Assemble evidence table
evidence <- generateEvidence(data_db, novel, genotype,
                             genotype_db, SampleGermlineIGHV,
                             j_call = "j_call",
                             junction = "junction")
}

}
\seealso{
See \link{findNovelAlleles}, \link{inferGenotype} and \link{genotypeFasta}
for generating the required input.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{getMutatedPositions}
\alias{getMutatedPositions}
\title{Find the location of mutations in a sequence}
\usage{
getMutatedPositions(
  samples,
  germlines,
  ignored_regex = "[\\\\.N-]",
  match_instead = FALSE
)
}
\arguments{
\item{samples}{vector of strings representing aligned sequences}

\item{germlines}{vector of strings representing aligned sequences
to which \code{samples} will be compared. If only
one string is submitted, it will be used for all
\code{samples}.}

\item{ignored_regex}{regular expression indicating what characters
should be ignored (such as gaps and N nucleotides).}

\item{match_instead}{if \code{TRUE}, the function returns the positions
that are the same instead of those that are
different.}
}
\value{
A list of the nucleotide positions of any differences between the
          input vectors.
}
\description{
\code{getMutatedPositions} takes two vectors of aligned sequences and
compares pairs of sequences. It returns a list of the nucleotide positions of
any differences.
}
\examples{
# Create strings to act as a sample sequences and a reference sequence
seqs <- c("----GATA", "GAGAGAGA", "TANA")
ref <- "GATAGATA"

# Find the differences between the two
getMutatedPositions(seqs, ref)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{updateAlleleNames}
\alias{updateAlleleNames}
\title{Update IGHV allele names}
\usage{
updateAlleleNames(allele_calls)
}
\arguments{
\item{allele_calls}{vector of strings representing IGHV allele names.}
}
\value{
Vector of strings representing updated IGHV allele names.
}
\description{
\code{updateAlleleNames} takes a set of IGHV allele calls and replaces any
outdated names (e.g. IGHV1-f) with the new IMGT names.
}
\note{
IMGT has removed \code{IGHV2-5*10} and \code{IGHV2-5*07} as it has determined they
are actually alleles \code{02} and \code{04}, respectively. The updated allele
names are based on IMGT release 2014-08-4.
}
\examples{
# Create a vector that uses old gene/allele names.
alleles <- c("IGHV1-c*01", "IGHV1-f*02", "IGHV2-5*07")

# Update the alleles to the new names
updateAlleleNames(alleles)

}
\references{
\enumerate{
  \item Xochelli et al. (2014) Immunoglobulin heavy variable (IGHV) genes
        and alleles: new entities, new names and implications for research and
        prognostication in chronic lymphocytic leukaemia. Immunogenetics. 67(1):61-6
}
}
\seealso{
Like \code{updateAlleleNames}, \link{sortAlleles} can help
         format a list of allele names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions.R
\name{findNovelAlleles}
\alias{findNovelAlleles}
\title{Find novel alleles from repertoire sequencing data}
\usage{
findNovelAlleles(
  data,
  germline_db,
  v_call = "v_call",
  j_call = "j_call",
  seq = "sequence_alignment",
  junction = "junction",
  junction_length = "junction_length",
  germline_min = 200,
  min_seqs = 50,
  auto_mutrange = TRUE,
  mut_range = 1:10,
  pos_range = 1:312,
  pos_range_max = NULL,
  y_intercept = 0.125,
  alpha = 0.05,
  j_max = 0.15,
  min_frac = 0.75,
  nproc = 1
)
}
\arguments{
\item{data}{\code{data.frame} containing repertoire data. See details.}

\item{germline_db}{vector of named nucleotide germline sequences
matching the V calls in \code{data}. These should be
the gapped reference germlines used to make the V calls.}

\item{v_call}{name of the column in \code{data} with V allele calls.
Default is \code{v_call}.}

\item{j_call}{name of the column in \code{data} with J allele calls.
Default is \code{j_call}.}

\item{seq}{name of the column in \code{data} with the
aligned, IMGT-numbered, V(D)J nucleotide sequence.
Default is \code{sequence_alignment}.}

\item{junction}{Junction region nucleotide sequence, which includes
the CDR3 and the two flanking conserved codons. Default
is \code{junction}.}

\item{junction_length}{Number of junction nucleotides in the junction sequence.
Default is \code{junction_length}.}

\item{germline_min}{the minimum number of sequences that must have a
particular germline allele call for the allele to
be analyzed.}

\item{min_seqs}{minimum number of total sequences (within the
desired mutational range and nucleotide range)
required for the samples to be considered.}

\item{auto_mutrange}{if \code{TRUE}, the algorithm will attempt to
determine the appropriate mutation range
automatically using the mutation count of the most
common sequence assigned to each allele analyzed.}

\item{mut_range}{range of mutations that samples may carry and
be considered by the algorithm.}

\item{pos_range}{range of IMGT-numbered positions that should be
considered by the algorithm.}

\item{pos_range_max}{Name of the column in \code{data} with the ending
positions of the V alignment in the germline
(usually \code{v_germline_end}). The end of the alignment will
be used to limit the range of positions to be
considered to count mutations. With \code{NULL} all
positions in the IMGT V region will be considered. In
this case, in sequences where the V was trimmed
on the 3', mutated nucleotides could include nucleotides
from the CDR3.}

\item{y_intercept}{y-intercept threshold above which positions should be
considered potentially polymorphic.}

\item{alpha}{alpha value used for determining whether the
fit y-intercept is greater than the \code{y_intercept}
threshold.}

\item{j_max}{maximum fraction of sequences perfectly aligning
to a potential novel allele that are allowed to
utilize to a particular combination of junction
length and J gene. The closer to 1, the less strict
the filter for junction length and J gene diversity
will be.}

\item{min_frac}{minimum fraction of sequences that must have
usable nucleotides in a given position for that
position to considered.}

\item{nproc}{number of processors to use.}
}
\value{
A \code{data.frame} with a row for each known allele analyzed.
Besides metadata on the the parameters used in the search, each row will have
either a note as to where the polymorphism-finding algorithm exited or a
nucleotide sequence for the predicted novel allele, along with columns providing
additional evidence.

The output contains the following columns:
\itemize{
  \item \code{germline_call}: The input (uncorrected) V call.
  \item \code{note}: Comments regarding the inference.
  \item \code{polymorphism_call}: The novel allele call.
  \item \code{nt_substitutions}: Mutations identified in the novel allele, relative
        to the reference germline (\code{germline_call})
  \item \code{novel_imgt}: The novel allele sequence.
  \item \code{novel_imgt_count}:  The number of times the sequence \code{novel_imgt}
        is found in the input data. Considers the subsequence of \code{novel_imgt}
        in the \code{pos_range}.
  \item \code{novel_imgt_unique_j}: Number of distinct J calls associated to \code{novel_imgt}
        in the input data. Considers the subsequence of \code{novel_imgt} in the \code{pos_range}.
  \item \code{novel_imgt_unique_cdr3}: Number of distinct CDR3 sequences associated
        with \code{novel_imgt} in the input data. Considers the subsequence of \code{novel_imgt}
        in the \code{pos_range}.
  \item \code{perfect_match_count}: Final number of sequences retained to call the new
        allele. These are unique sequences that have V segments that perfectly match
        the predicted germline in the \code{pos_range}.
  \item \code{perfect_match_freq}: \code{perfect_match_count / germline_call_count}
  \item \code{germline_call_count}: The number of sequences with the \code{germline_call}
        in the input data that were initially considered for the analysis.
  \item \code{germline_call_freq}: The fraction of sequences with the \code{germline_call}
        in the input data initially considered for the analysis.
  \item \code{germline_imgt}: Germline sequence for \code{germline_call}.
  \item \code{germline_imgt_count}: The number of times the \code{germline_imgt}
        sequence is found in the input data.
  \item \code{mut_min}: Minimum mutation considered by the algorithm.
  \item \code{mut_max}: Maximum mutation considered by the algorithm.
  \item \code{mut_pass_count}: Number of sequences in the mutation range.
  \item \code{pos_min}: First position of the sequence considered by the algorithm (IMGT numbering).
  \item \code{pos_max}: Last position of the sequence considered by the algorithm (IMGT numbering).
  \item \code{y_intercept}: The y-intercept above which positions were considered
        potentially polymorphic.
  \item \code{y_intercept_pass}: Number of positions that pass the \code{y_intercept} threshold.
  \item \code{snp_pass}: Number of sequences that pass the \code{y_intercept} threshold and are
        within the desired nucleotide range (\code{min_seqs}).
  \item \code{unmutated_count}: Number of unmutated sequences.
  \item \code{unmutated_freq}: Number of unmutated sequences over \code{germline_imgt_count}.
  \item \code{unmutated_snp_j_gene_length_count}: Number of distinct combinations
        of SNP, J gene, and junction length.
  \item \code{snp_min_seqs_j_max_pass}: Number of SNPs that pass both the \code{min_seqs}
        and \code{j_max} thresholds.
  \item \code{alpha}: Significance threshold to be used when constructing the
        confidence interval for the y-intercept.
  \item \code{min_seqs}: Input \code{min_seqs}. The minimum number of total sequences
        (within the desired mutational range and nucleotide range) required
        for the samples to be considered.
  \item \code{j_max}: Input \code{j_max}. The maximum fraction of sequences perfectly
        aligning to a potential novel allele that are allowed to utilize to a particular
        combination of junction length and J gene.
  \item \code{min_frac}: Input \code{min_frac}. The minimum fraction of sequences that must
        have usable nucleotides in a given position for that position to be considered.
}

The following comments can appear in the \code{note} column:

\itemize{
  \item \emph{Novel allele found}: A novel allele was detected.
  \item \emph{Same as:}: The same novel allele sequence
              has been identified multiple times. If this happens, the function
              will also throw the message 'Duplicated polymorphism(s) found'.
  \item \emph{Plurality sequence too rare}: No sequence is frequent enough to pass
        the J test (\code{j_max}).
  \item \emph{A J-junction combination is too prevalent}: Not enough J diversity (\code{j_max}).
  \item \emph{No positions pass y-intercept test}: No positions above \code{y_intercept}.
  \item \emph{Insufficient sequences in desired mutational range}:
        \code{mut_range} and \code{pos_range}.
  \item \emph{Not enough sequences}: Not enough sequences in the desired mutational
        range and nucleotide range (\code{min_seqs}).
  \item \emph{No unmutated versions of novel allele found}: All observed variants of the
        allele are mutated.
}
}
\description{
\code{findNovelAlleles} analyzes mutation patterns in sequences thought to
align to each germline allele in order to determine which positions
might be polymorphic.
}
\details{
The TIgGER allele-finding algorithm, briefly, works as follows:
Mutations are determined through comparison to the provided germline.
Mutation frequency at each *position* is determined as a function of
*sequence-wide* mutation counts. Polymorphic positions exhibit a high
mutation frequency despite sequence-wide mutation count. False positive of
potential novel alleles resulting from clonally-related sequences are guarded
against by ensuring that sequences perfectly matching the potential novel
allele utilize a wide range of combinations of J gene and junction length.
}
\examples{
\donttest{
# Note: In this example, with SampleGermlineIGHV,
# which contains reference germlines retrieved on August 2014,
# TIgGER finds the allele IGHV1-8*02_G234T. This allele
# was added to IMGT as IGHV1-8*03 on March 28, 2018.

# Find novel alleles and return relevant data
novel <- findNovelAlleles(AIRRDb, SampleGermlineIGHV)
selectNovel(novel)
}

}
\seealso{
\link{selectNovel} to filter the results to show only novel alleles.
\link{plotNovel} to visualize the data supporting any
novel alleles hypothesized to be present in the data and
\link{inferGenotype} and \link{inferGenotypeBayesian} to determine if the novel alleles are frequent
enough to be included in the subject's genotype.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{SampleNovel}
\alias{SampleNovel}
\title{Example novel allele detection results}
\format{
A \code{data.frame} where rows correspond to alleles checked for
polymorphisms and columns give results as well as parameters used to run
the test.
}
\description{
A \code{data.frame} of novel allele detection results from \link{findNovelAlleles}.
Source data was a collection of V(D)J immunoglobulin sequences derived from a single
individual (PGP1), sequenced on the Roche 454 platform, and assigned by
IMGT/HighV-QUEST to IGHV1 family alleles.
}
\references{
\enumerate{
  \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
        sequencing data reveals a high frequency of novel immunoglobulin V gene
        segment alleles. PNAS. 112(8):E862-70.
}
}
\seealso{
See \link{findNovelAlleles} for detailed column descriptions.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{SampleGenotype}
\alias{SampleGenotype}
\title{Example genotype inference results}
\format{
A \code{data.frame} where rows correspond to genes carried by an
individual and columns lists the alleles of those genes and their counts.
}
\description{
A \code{data.frame} of genotype inference results from \link{inferGenotype}
after novel allele detection via \link{findNovelAlleles}.
Source data was a collection of V(D)J immunoglobulin sequences derived from a single
individual (PGP1), sequenced on the Roche 454 platform, and assigned by
IMGT/HighV-QUEST to IGHV1 family alleles.
}
\references{
\enumerate{
  \item Gadala-Maria, et al. (2015) Automated analysis of high-throughput B cell
        sequencing data reveals a high frequency of novel immunoglobulin V gene
        segment alleles. PNAS. 112(8):E862-70.
}
}
\seealso{
See \link{inferGenotype} for detailed column descriptions.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{SampleGermlineIGHV}
\alias{SampleGermlineIGHV}
\title{Example Human IGHV germlines}
\format{
Values correspond to IMGT-gaped nucleotide sequences (with
nucleotides capitalized and gaps represented by ".") while names correspond
to stripped-down IMGT allele names (e.g. "IGHV1-18*01").
}
\description{
A \code{character} vector of all 344 human IGHV germline gene segment alleles
in IMGT/GENE-DB release 2014-08-4.
}
\references{
\enumerate{
  \item Xochelli, et al. (2014) Immunoglobulin heavy variable (IGHV) genes and
        alleles: new entities, new names and implications for research and
        prognostication in chronic lymphocytic leukaemia. Immunogenetics. 67(1):61-6.
}
}
\keyword{data}
